#!/bin/sh
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# 2001 QAcafe http://qacafe.com
#
# $Id: rip-routes.tcl,v 1.1 2001/06/01 01:16:26 joe Exp $
#
# This is an example of a pktsrc tool for RIP. It sends a directed RIP request
# to a RIPv1/v2 router for its routing table. It waits 5 seconds for all
# responses and displays the RIP routing table.
#
# The next line execs tclsh \
exec tclsh "$0" "$@"

# -- include pktsrc
package require pktsrc

# -- help/usage
proc help { name } {
  puts "$name \[option \[value]] ..."

  puts "-trace         Display protocol decodes"  
  puts "-packets       Display hex packet dumps"
  puts "-timestamp     Show timestamps in logging messages"
  puts "-myip          Specify my IP address" 
  puts "-mymask        Specify my network mask (default 255.255.255.0)"
  puts "-dut           Specify DUT's IP address"
  puts "-version       1 | 2  set the RIP version for the request"
  puts "-interface     Set the physical interface to use, default is eth0"
  puts "-help          Display this help"
  puts ""

}

#
# -- parse command line options
#

for { set i 0 } { $i < $argc } { incr i } {

  set arg [lindex $argv $i]

    switch -- $arg \
       "-trace" {
           set options(trace) on
           continue

       } "-packets" {
           set options(packets) 1
           continue

       } "-timestamp" {
           set options(timestamp) 1
           continue

       } "-myip" {
           incr i
           set options(myip) "[lindex $argv $i]"
           continue

       } "-mymask" {
           incr i
           set options(mymask) "[lindex $argv $i]"
           continue

       } "-version" {
           incr i
           set options(version) "[lindex $argv $i]"
           continue

       } "-dut" {
           incr i
           set options(dut) "[lindex $argv $i]"
           continue

       } "-interface" {
           incr i
           set options(interface) "[lindex $argv $i]"
           continue

       } "-help" {
           help $argv0 
           exit

       }  default {
           puts "Unknown switch \"$arg\""
           exit
       }

    break
}


# -- check for -myip
if [ info exists options(myip) ] {
   set myip $options(myip)
} else {
   puts "Error: you must specify the IP address with -myip"
   exit
}

# -- check for -dut
if [ info exists options(dut) ] {
   set dut $options(dut)
} else {
   puts "Error: you must specify the IP of target router with -dut"
   exit
}

if [ info exists options(mymask) ] {
   # -- set the mask if specified
   set mask $options(mymask)
} else {
   set mask 255.255.255.0
}


# -- open a packet source interface
set interface eth0
if [ info exists options(interface) ] {
   set interface $options(interface)
}
set port1 [ pktsrc open $interface ]

# -- place filters on each packet source, I only care about ARP,
# my IP address and OSPF
$port1 filter "arp or host $myip or 224.0.0.9"


# -- create stacks 
Stack_init stack1 $port1 $myip $mask RIP

 
# -- turn on protocol decoding
if [ info exists options(trace) ] {
   set stack1(print) on
}

# -- turn on message timestamps
if [ info exists options(timestamp) ] {
   set stack1(timestamp) on
}

# -- turn on raw packet dumps
if [ info exists options(packets) ] {
   $port1 dump on
}  

# -- arp for DUT
ARP_get stack1 [ iptohex $dut ]
Event_wait 1000

# -- allow RIP requests to be received on a different port
set srcPort [ RIP_client stack1 ]

# -- the version to RIPv2
if [ info exists options(version) ] {
    if { $options(version) == 2 } {
      RIP_set_version stack1 2
    }
}

# -- set up a callback to handle RIP reponses on this port
Stack_add_callback stack1 ripType 02 rip_response

# -- send a RIP request
RIP_send_request stack1 $dut $srcPort

set myroutes(total) 0

#
# -- here is the RIP response handler
#

proc rip_response { packet stack options } {
   global srcPort myroutes
   upvar $stack s

   # -- get the contents of the IP packet from the stack
   set ipState $s(current,IPv4)
   upvar #$s(current,IPv4,level) $ipState ip

   # -- get the contents of the UDP packet from the stack
   set udpState $s(current,UDP)
   upvar #$s(current,UDP,level) $udpState udp
 
   # puts "UDP dest port is [expr 0x$udp(dstPort)], mySrc = $srcPort"

   # -- get the port number is decimal
   set port [expr 0x$udp(dstPort)]

   # -- check that this is the matching udp destination port
   if { "$srcPort" == "$port" } {

      # -- get the contents of the RIP packet from the stack
      set ripState $s(current,RIP)
      upvar #$s(current,RIP,level) $ripState rip 
  
      puts "Got RIP response from [hextoip $ip(src)] with $rip(total) routes"

      # -- process routes ... check unique, etc
      for { set i 1 } { $i <= $rip(total) } { incr i } { 
        if [ info exists rip(route,$i) ] {

           # -- check that this route is unique. I need to handle overlapping
           # prefixes ... but I'm not right now.

           if [ info exists myroutes($rip(route,$i),nexthop) ] {
             continue
           }

           # -- decode this route
           RIP_decode_route $rip(route,$i) out

           # -- build my own route table to display later
           set key $out(address)
           lappend myroutes(routes) $key
           incr myroutes(total)

           # -- remember some fields
           set myroutes($key,nexthop)  $out(nexthop)
           set myroutes($key,mask)     $out(mask)
           set myroutes($key,metric)   $out(metric)
        }
      }
   } else {
      # -- Ignore, probably a normal update
   }
}

# -- wait 5 seconds for all repsonses
Event_wait 5000


# -- print out my route table in cisco-like fashion

puts "\n$myroutes(total) RIP routes from $dut\n"

if { $myroutes(total) > 0 } {
   foreach r $myroutes(routes) {
      puts "R   [ format "%s \[%s] via %-16s" \
        "[hextoip $r]/[ maskbits [hextoip $myroutes($r,mask)]]" \
         [expr 0x$myroutes($r,metric)] \
         [hextoip $myroutes($r,nexthop)]]"
   }
}

puts ""


