#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# 2002 QAcafe http://qacafe.com
#
# $Id: dns.tcl,v 1.5 2002/11/15 13:23:54 joe Exp $
#
#
# pktsrc DNS
#
# decode based on rfc 1035
#
# ----------------------------------------------------------------------------

# -- this is part of the pktsrc package
package provide pktsrc 1.3

proc DNS_input { packet stack { options none }} {
   upvar $stack s

   if { $options == "print" } {
      if [ catch { DNS_decode $packet dns } reason ] {
         pktsrcWarn s "Unable to fully parse DNS packet"
      } else {
         DNS_print dns
      }
      return
   }

   # -- remember this packet in the 'dns' variable. This allows other
   # callbacks to get access to the decodes at any Tcl stack level

   eval [ Stack_save_packet s DNS dns ]
   
   # -- now decode the packet into the 'dns' variable
   if [ catch { DNS_decode $packet dns } reason ] {
      pktsrcWarn s "Unable to fully parse DNS packet"
      return
   }

   # -- check for callbacks 

   if [ info exists s(dnsType,$dns(opcode)) ] {

      # -- drive any input callbacks
      foreach c $s(dnsType,$dns(opcode),callbacks) {
          # -- send the whole packet
          $c $packet s $options
      }

   }
}

#
#
# DNS_init - set up any deaults for DNS
#

proc DNS_init { stack } {
   upvar $stack s

   # -- attempt to select a unique DNS id
   set s(DNS,id) [ expr [clock clicks] % 65535 ]
}


proc DNS_decode { packet dns } {
   upvar $dns out

#
# from RFC 1035
#
# 4.1.1. Header section format
#
# The header contains the following fields:
#
#                                    1  1  1  1  1  1
#      0  1  2  3  4  5  6  7  8  9  0  1  2  3  4  5
#    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
#    |                      ID                       |
#    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
#    |QR|   Opcode  |AA|TC|RD|RA|   Z    |   RCODE   |
#    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
#    |                    QDCOUNT                    |
#    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
#    |                    ANCOUNT                    |
#    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
#    |                    NSCOUNT                    |
#    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
#    |                    ARCOUNT                    |
#    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
#

   # -- parse out all the fields we find interesting ...

   set out(id)             [ string range $packet 0 3 ]

   set word                [ string range $packet 4 7 ]
   set out(qr)             [ expr 0x$word >> 15 ]
   set out(opcode)         [ expr (0x$word & 0x7800) >> 11 ]

   set out(aa)             [ expr (0x$word & 0x0400) >> 10 ]
   set out(tc)             [ expr (0x$word & 0x0200) >> 9 ]
   set out(rd)             [ expr (0x$word & 0x0100) >> 8 ]
   set out(ra)             [ expr (0x$word & 0x0080) >> 7 ]
   set out(z)              [ expr (0x$word & 0x0070) >> 4 ]
   set out(rcode)          [ expr (0x$word & 0x000f) ]

   set out(qdcount)        [ string range $packet 8 11 ]
   set out(ancount)        [ string range $packet 12 15 ]
   set out(nscount)        [ string range $packet 16 19 ]
   set out(arcount)        [ string range $packet 20 23 ]

   
   DNS_decode_questions $packet out 24 [ expr 0x$out(qdcount) ]

}

proc DNS_decode_questions { packet dns startp count } {
   upvar $dns d

   # -- parse out questions

   for { set i 0 } { $i < $count } { incr i } {

      set name ""
      set endp [ expr $startp + 1 ]

      # -- read the length
      set length [ string range $packet $startp $endp ]

      incr startp 2

      while { "$length" != "00" } {


        set endp [ expr $startp + (0x$length * 2 ) -1 ]
        set tuple [ string range $packet $startp $endp ]
	if { "$name" == "" } {
          set name [ pktsrc ascii2text $tuple ]
        } else {
          set name "${name}.[ pktsrc ascii2text $tuple ]"
        }
  
        incr startp [ expr 0x$length * 2 ]
        set endp [ expr $startp + 1 ]
        set length [ string range $packet $startp $endp ]
        incr startp 2

      }

      set endp [ expr $startp + 3 ]
      set type [ string range $packet $startp $endp ]
      incr startp 4
      incr endp 4
      set class [ string range $packet $startp $endp ]

      incr startp 4

      # puts "### $name, $type, $class"
      set d(question,1,name) $name
      set d(question,1,type) $type
      set d(question,1,class) $class
 
   }

}
 
proc DNS_print { dns } {
   upvar $dns out

   puts "DNS: ---- DNS Packet ----"
   puts "DNS: "
   puts "DNS: ID                         = 0x$out(id)"
   puts "DNS: Type                       = [DNS_type $out(qr)]"
   puts "DNS: Opcode                     = [DNS_opcode_type $out(opcode)] ($out(opcode))"

   puts "DNS: Questions (qdcount)        = [expr 0x$out(qdcount)]"
   puts "DNS: Answers (ancount)          = [expr 0x$out(ancount)]"
   puts "DNS: Authority RR (nscount)     = [expr 0x$out(nscount)]"
   puts "DNS: Additional RR (arcount)    = [expr 0x$out(arcount)]"

 
   # -- display questions
  
   for { set i 1 } { $i <= [ expr 0x$out(qdcount)] } { incr i } {

      puts "DNS: Question $i"
      puts "DNS:   Name                     = $out(question,$i,name)"
      puts "DNS:   Type                     = $out(question,$i,type)"
      puts "DNS:   Class                    = $out(question,$i,class)"
   }

   # -- display answers + other (TBD)

   puts "DNS:"
   puts "DNS: ---- End DNS Packet ----"

}

proc DNS_type { qr } {
   if { "$qr" == "1" } {
      return "Response"
   } elseif { "$qr" == "0" } {
      return "Query"
   } else {
      error "DNS_type called with ->$qr<-, only 0 or 1 supported"
   }
}

proc DNS_opcode_type { opcode } {
   if { "$opcode" == "0" } {
      return "Standard Query"
   } elseif { "$opcode" == "1" } {
      return "Inverse Query"
   } elseif { "$opcode" == "2" } {
      return "Server Status Request"
   } else {
      return "unknown type"
   }
}

#
# DNS_dummy_server
#
# Start a dummy DNS server on the specified port
#
#
 
proc DNS_dummy_server { stack { port 53 } } {
   upvar $stack s

   set s(DNS,dummyServer) 1

   set srcPort [ format "%4.4x" $port ]

   pktsrcInfo s "Starting up dummy dns server on port $port"

   UDP_server s $port DNS_server_input

}
  
proc DNS_dummy_server_shutdown { stack { port 53 }} {
   upvar $stack s

   set s(DNS,dummyServer) 0

   set srcPort [ format "%4.4x" $port ]

   pktsrcInfo s "Shuting down dummy dns server on port $port"

   UDP_server_shutdown s $port DNS_server_input

}


#
# DNS_server_input
#
# This is the input routine for an dns session
# Don't bother parsing the input request, just return the dummy answer
#

proc DNS_server_input { packet stack stack {options none}} {
   upvar $stack s

   if { "$options" == "print" } {
       return
   }

   set dnsState $s(current,DNS)
   upvar #$s(current,DNS,level) $dnsState dns

   # -- dummy server is running, process request

   # -- get the current IP contents
   set ipState $s(current,IPv4)
   upvar #$s(current,IPv4,level) $ipState ip

   # -- get the current udp contents
   set udpState $s(current,UDP)
   upvar #$s(current,UDP,level) $udpState udp

   # -- get the current dns contents
   set dnsState $s(current,DNS)
   upvar #$s(current,DNS,level) $dnsState dns

   # -- just send a dump response back
   pktsrcInfo s "Responding to DNS query (0x$dns(id)) from [hextoip $ip(src)]:[expr 0x$udp(srcPort)]"

   set response "${dns(id)}85800001000200020002037777770671616361666503636f6d0000010001c00c000500010000a8c00002c010c010000100010000a8c00004ccaedf80c01000020001000151800010036e7331096e65746e6174696f6ec017c01000020001000151800006036e7332c04ec04a0001000100000e100004ccaedf01c0660001000100000e100004ccaedf1f"

   UDP_send s $response $ip(src) [ expr 0x$udp(srcPort) ] 53 

}

proc DNS_who_is { stack server } {
   upvar $stack s

   set port [ UDP_get_client_port s ]
   set srcPort [ format "%4.4x" $port ]

   # -- generate a unique id for this request
   set id [ DNS_next_id s ]
   pktsrcInfo s "Initiate DNS query (0x$id) to $server:53 using src port $port"

   # -- send request
   set request "${id}01000001000000000000037777770671616361666503636f6d0000010001"

   UDP_send s "$request" [iptohex $server] 53 $port 

   set result [ UDP_get_packet s $srcPort udp 10 ] 

   if { "$result" != "-1" } {
      pktsrcInfo s "got DNS response from $server"
      return "okay"
   } else {
      pktsrcWarn s "no response from $server after 10 seconds"
      return ""
   }
}

proc DNS_who_is_async { stack server session } {
   upvar $stack s
   upvar $session sess

   set port [ UDP_open s sess $server 53 ]

   # -- generate a unique id for this request
   set id [ DNS_next_id s ]
   pktsrcInfo s "Initiate DNS query (0x$id) to $server:53 using src port $port"

   # -- send request
   set request "${id}01000001000000000000037777770671616361666503636f6d0000010001"

   # -- should be some kind of session send ?
   UDP_send s "$request" [iptohex $server] 53 $port 

}

#
# DNS_read
#
# read any queued data from the udp session
# store any data in ascii hex format
# return the number of bytes read
#

proc DNS_read { stack udpSession data } {
   upvar $stack s
   upvar $udpSession session
   upvar $data dataText

   set bytes 0
   set dataBuf ""
   set dataText ""

   #
   # While the session is open, keep reading data
   # No Parsing is implemented

   set dataIn [ UDP_get_data s session ]
   set dataBuf "$dataBuf$dataIn"

   # -- run the event loop (blocking call)
   Event_wait 10

   # -- don't convert data to text since its not a text stream
   # return the hex stream

   # set dataText [ pktsrc ascii2text $dataBuf ]
   set dataText $dataBuf

   set bytes [ expr [ string length $dataText ] / 2 ]

   # -- return the number of bytes
   return $bytes
}

#
# DNS_next_id
#
# return the next available DNS id
#

proc DNS_next_id { stack } {
   upvar $stack s

   set id $s(DNS,id)

   if { "$id" == "65535" } {
     set s(DNS,id) 1
   } else {
     incr s(DNS,id)
   }

   # -- return the id in hex format
   return [ format "%4.4x" $id ]
}






