#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# 2002 QAcafe http://qacafe.com
#
# $Id: ftp.tcl,v 1.3 2002/11/15 13:23:54 joe Exp $
#
#
# Poor man's Ftp
#
#
# ----------------------------------------------------------------------------

# -- this is part of the pktsrc package
package provide pktsrc 1.3

proc FTP_input { packet stack { options none }} {
   upvar $stack s

   if { $options == "print" } {

      FTP_decode $packet ftp
      FTP_print ftp
      return

   } else {

      # -- remember this packet in the 'ftp' variable. This allows other
      # callbacks to get access to the decodes at any Tcl stack level

      eval [ Stack_save_packet s FTP ftp ]
   
      # -- now decode the packet into the 'ftp' variable
      FTP_decode $packet ftp

   }

}


#
#
# FTP_init - set up any deaults for FTP
#

proc FTP_init { stack } {
   upvar $stack s

}

proc FTP_decode { packet ftp } {
   upvar $ftp out

   # -- just treat everything as one text string for now

   set out(data)        [ string range $packet 0 end ]

}

proc FTP_print { ftp } {
   upvar $ftp out

   if { $out(data) != "" } {

     puts "FTP: ---- FTP Packet ----"
     puts "FTP: "

     # -- display everything as one text string
     puts "[pktsrc ascii2text $out(data) ]"

     puts "FTP: "
     puts "FTP: ---- End of FTP ----"

   }
}

proc FTP_send { stack tcpSession data } {
   upvar $stack s
   upvar $tcpSession session

   set hex [ pktsrc text2ascii $data ]

   # -- add carriage return and line feeds at the end of the data
   return [ TCP_write s session "${hex}0d0a" ]

}

#
# FTP_read
#
# read data from an http connection
# timeout after 'timeOut' if no data is received
#

proc FTP_read { stack tcpSession data { timeOut 5000 } } {
   upvar $stack s
   upvar $tcpSession session
   upvar $data dataText

   set bytes 0
   set dataBuf ""
   set dataText ""

   #
   # While the session is open, keep reading data
   # No Parsing is implemented

   set elapsed 0

   while { $elapsed < $timeOut } {

      set dataIn [ TCP_get_data s session ]
      set dataBuf "$dataBuf$dataIn"

      # -- check the state
      set state [ TCP_state s session ]

      # -- keep reading while we are establsihed
      if { "$state" != "established" } {
         break
      }
 
      # -- run the event loop (blocking call)
      Event_wait 100
      incr elapsed 100

      # -- break if we have some data to return

      # NOTE: we should really be reading until we find a carriage return
      # line feed that ends the ftp command.

      if { "$dataBuf" != "" } {
         break
      }

   }

   # -- convert data to text
   set dataText [ pktsrc ascii2text $dataBuf ]
   set bytes [ string length $dataText ]

   if { $elapsed >= $timeOut } {
     pktsrcWarn s "timeout waiting for data in FTP_read, $bytes bytes read"
   }

   # -- return the number of bytes
   return $bytes
}


#
# FTP_dummy_server
#
# Start a dummy FTP server on the specified port
#
# 

proc FTP_dummy_server { stack { port 21 }} {
   upvar $stack s

   set s(FTP,dummyServer) 1
   set s(FTP,lastPortCmd) ""

   set srcPort [ format "%4.4x" $port ]

   pktsrcInfo s "Starting up dummy ftp server on port $port"

   TCP_server s $port FTP_server_input

}

proc FTP_dummy_server_shutdown { stack { port 21 }} {
   upvar $stack s

   set s(FTP,dummyServer) 0

   set srcPort [ format "%4.4x" $port ]

   pktsrcInfo s "Shuting down dummy ftp server on port $port"

   TCP_server_shutdown s $port FTP_server_input

}

#
# FTP_server_input
#
# This is the input routine for an ftp session
#

proc FTP_server_input { packet stack stack {options none}} {
   upvar $stack s

   if { "$options" == "print" } {
       return
   }

   set ftpState $s(current,FTP)
   upvar #$s(current,FTP,level) $ftpState ftp

   # -- get the current IP contents
   set ipState $s(current,IPv4)
   upvar #$s(current,IPv4,level) $ipState ip

   # -- get the current tcp contents
   set tcpState $s(current,TCP)
   upvar #$s(current,TCP,level) $tcpState tcp

   if { $ftp(data) != "" } {

      # -- dummy server is running, process request

      # -- check that this is a valid session
      set sid \
        [ TCP_get_session s session $ip(src) $tcp(srcPort) $tcp(dstPort) ]

      if { "$sid" != "0" } {

        # -- process ftp command
        set command [ string trim [ pktsrc ascii2text $ftp(data) ] \r\n ]

        pktsrcInfo s "ftp server received command '$command'"

        set cmd [ string range $command 0 3 ]
	if { "$cmd" == "PORT" || "$cmd" == "port" } {
           set s(FTP,lastPortCmd) [ string range $command 5 end ]
           pktsrcInfo s "saving results of last port command '$s(FTP,lastPortCmd)'"

        }
        
        # -- just return a success code for each command 
        FTP_send s session "200 command successful."

      }
   } else {

      # -- new session was created
      set sid \
        [ TCP_get_session s session $ip(src) $tcp(srcPort) $tcp(dstPort) ]

      
      if { "$sid" != "0" } {
         pktsrcInfo s "new ftp client connection from [hextoip $ip(src)]:[expr 0x$tcp(srcPort)]"
         FTP_send s session "220 qacafe FTP server (Version 1.0) ready."
      }
   }
}

#
# FTP_build_port_command
#
# build the ip,port string used in the ftp port commands
#

proc FTP_build_port_command { ip port } {

   set iphex [ iptohex $ip ]
   set porthex [ format "%4.4x" $port ]

   set b1 [ string range $iphex 0 1 ]
   set b2 [ string range $iphex 2 3 ]
   set b3 [ string range $iphex 4 5 ]
   set b4 [ string range $iphex 6 7 ]

   set p1 [ string range $porthex 0 1 ]
   set p2 [ string range $porthex 2 3 ]

   set str [ format "%d,%d,%d,%d,%d,%d" 0x$b1 0x$b2 0x$b3 0x$b4 0x$p1 0x$p2 ]

   return $str

}

#
# FTP_find_ip_from_PORT
#
# find the ip address specified in the ftp PORT command
#

proc FTP_find_ip_from_PORT { command } {

   set ip ""

   # -- try to parse the port command
   set result \
       [ regexp {(.+),(.+),(.+),(.+),(.+),(.+)} $command all b1 b2 b3 b4 p1 p2 ]

   if { "$result" == "1" } {
       set ip "${b1}.${b2}.${b3}.${b4}"
   }

   return $ip
}

#
# FTP_find_port_from_PORT
#
# find the port specified in the ftp PORT command
#

proc FTP_find_port_from_PORT { command } {
   set port ""

   # -- try to parse the port command
   set result \
      [ regexp {(.+),(.+),(.+),(.+),(.+),(.+)} $command all b1 b2 b3 b4 p1 p2 ]

   if { "$result" == "1" } {
       set port [ expr $p1 * 256 + $p2 ]
   }

   return $port
}



