#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# 2001 QAcafe http://qacafe.com
#
# $Id: icmp.tcl,v 1.10 2002/11/19 22:33:23 joe Exp $
#
#
# Poor man's ICMP
#
# This version currently supports ICMP Echo
#
# ----------------------------------------------------------------------------

# -- this is part of the pktsrc package
package provide pktsrc 1.3

proc ICMP_input { packet stack { options none }} {
   upvar $stack s

   set port $s(port)
   ICMP_decode $packet icmp

   if { $options == "print" } {
      ICMP_print icmp
      return
   } else {
      # -- I should hide this in a save proto
      set s(current,ICMP) icmp
      set s(current,ICMP,level) [info level]
   }

   # -- check for callbacks at the ICMP level
   if [ info exists s(icmpType,$icmp(type)) ] {

      # -- drive any input callbacks
      foreach c $s(icmpType,$icmp(type),callbacks) {
          $c $icmp(data) s $options
      }

   } else {
      # no callback
   }
}

#
#
# ICMP_init - set up any deaults for ICMP
#

proc ICMP_init { stack } {
   upvar $stack s

   # -- setup a ping client by default
   set s(icmpType,08)             enabled
   set s(icmpType,08,callbacks)   ICMP_echo_input

   # -- turn responses on
   set s(ICMP,pingResponse,mode) on
}

proc ICMP_decode { packet icmp } {
   upvar $icmp out

   # -- parse out all the fields we find interesting ...

   set out(type)        [ string range $packet 0 1 ]
   set out(code)        [ string range $packet 2 3 ]
   set out(checksum)    [ string range $packet 4 7 ]

   if { $out(type) == "08" || $out(type) == "00" } {

      set out(id)          [ string range $packet 8 11 ]
      set out(sequence)    [ string range $packet 12 15 ]
      set out(data)        [ string range $packet 16 end ]
   }

}

proc ICMP_print { icmp } {
   upvar $icmp out

   puts "ICMP: ---- ICMP Packet ----"
   puts "ICMP: "
   puts "ICMP: Type                      = [ICMP_type $out(type)] (0x$out(type))"
   puts "ICMP: Code                      = $out(code)"
   puts "ICMP: Checksum                  = 0x$out(checksum)"

   # -- Echo 

   if { $out(type) == "00" || $out(type) == "08" } {
       puts "ICMP: ID                        = [ expr 0x$out(id)] (0x$out(id))"
       puts "ICMP: Sequence                  = [ expr 0x$out(sequence) ]"
   }

   puts "ICMP:"
   puts "ICMP: ---- End ICMP Packet ----"
}

proc ICMP_type { type } {

   if { $type == "00" } {
      set str "Echo Reply"
   } elseif { $type == "03" } {
      set str "Destination Unreachable"
   } elseif { $type == "04" } {
      set str "Source Quench"
   } elseif { $type == "05" } {
      set str "Redirect"
   } elseif { $type == "08" } {
      set str "Echo"
   } elseif { $type == "09" } {
      set str "Router Advertisement"
   } elseif { $type == "0a" } {
      set str "Router Solicitation"
   } elseif { $type == "0b" } {
      set str "Time Exceeded"
   } elseif { $type == "0c" } {
      set str "Parameter Problem"
   } elseif { $type == "0d" } {
      set str "Timestamp"
   } elseif { $type == "0e" } {
      set str "Timestamp Reply"
   } elseif { $type == "20" } {
      set str "Mobile Host Redirect"
   } elseif { $type == "23" } {
      set str "Mobile Registration Request"
   } elseif { $type == "24" } {
      set str "Mobile Registration Reply"
   } else {
      set str "Unknown ICMP Type ($type)"
   }

   return $str

}

proc ICMP_ping_client { stack } {

   upvar $stack s

   # -- turn this stack into a ping client
   
   # -- add callback for ICMP Echo
   set s(icmpType,08)             enabled
   set s(icmpType,08,callbacks)   ICMP_echo_input
  
}

proc ICMP_build_echo_response { stack } {
   upvar $stack s

   set icmpState $s(current,ICMP)
   upvar #$s(current,ICMP,level) $icmpState icmp 

   set p [ format "%s%s%s%s%s%s" "00" "00" "0000" \
       $icmp(id) $icmp(sequence) $icmp(data) ]


   set checksum [ pktsrc checksum $p ]

   set packet [ format "%s%s%s%s%s%s" "00" "00" $checksum $icmp(id) $icmp(sequence) $icmp(data) ]

   return $packet 

}

proc ICMP_build_echo_request { stack id seq { data "" } } {

   upvar $stack s

   # if not data was supplied build some
   if { "$data" == "" } {
      set data [ string repeat deadbe 12 ]
   }

   set p [ format "%s%s%s%4.4x%4.4x%s" "08" "00" "0000" $id $seq $data ]

   # -- compute the checksum
   set checksum [ pktsrc checksum $p ]
   set packet [ format "%s%s%s%4.4x%4.4x%s" "08" "00" $checksum $id $seq $data ]

   return $packet 

}

proc ICMP_build_port_unreachable { stack data } {
   upvar $stack s

   # -- build packet and comput checksum
   set p [ format "%s%s%s%s%s" "03" "03" "0000" "00000000" $data ]

   set checksum [ pktsrc checksum $p ]
   set packet [ format "%s%s%s%s%s" "03" "03" $checksum "00000000" $data ]

   # -- return new packet
   return $packet 

}


proc ICMP_echo_input { packet stack { options none }} {
   upvar $stack s

   # -- check if ping responses are enabled
   if { "$s(ICMP,pingResponse,mode)" != "on" } {
      return
   }

   # -- I should hide this
   set ipState $s(current,IPv4)
   upvar #$s(current,IPv4,level) $ipState ip

   set icmpState $s(current,ICMP)
   upvar #$s(current,ICMP,level) $icmpState icmp 


   # -- get the source address from IP
  
   # -- ignore if this is not for me
   if { "$ip(dst)" != "$s(ip)" } {
      return
   }

   pktsrcInfo s "ICMP Echo Request from [hextoip $ip(src)], icmp_seq=[expr 0x$icmp(sequence)]"

   # -- send a response
   set response [ ICMP_build_echo_response s ]


   # -- send the packet on to IP
   IPv4_send $response s $ip(src) 01

}

proc ICMP_port_unreachable { stack } {
   upvar $stack s

   # -- I should hide this
   set ipState $s(current,IPv4)
   upvar #$s(current,IPv4,level) $ipState ip

   set udpState $s(current,UDP)
   upvar #$s(current,UDP,level) $udpState udp

   set ethState $s(current,Ethernet)
   upvar #$s(current,Ethernet,level) $ethState eth

   pktsrcInfo s "Sending ICMP Port Unreachable to [hextoip $ip(src)] for Port [expr 0x$udp(dstPort)]"

   set ret [ string range $eth(data) 0 55 ]
   # -- send a response
   set response [ ICMP_build_port_unreachable s $ret ]

   # -- send the packet on to IP
   IPv4_send $response s $ip(src) 01

}
 
#
# -- send an ICMP packet

proc ICMP_send { stack packet dest } {
   upvar $stack s
   IPv4_send $packet s [iptohex $dest ] 01

}

proc ICMP_ping_response { packet stack { options none } } {
   upvar $stack s

   if { "$options" != "print" } {

     # -- get the contents of the IP packet from the stack
     set ipState $s(current,IPv4)
     upvar #$s(current,IPv4,level) $ipState ip

     # -- get the contents of the ICMP packet from the stack
     set icmpState $s(current,ICMP)
     upvar #$s(current,ICMP,level) $icmpState icmp 

     # -- get the length of the IP packet
     set len [expr 0x$ip(length)]

     # -- display the result

     # -- now display the usual
     pktsrcInfo s "$len bytes from [hextoip $ip(src)]: icmp_seq=[expr 0x$icmp(sequence)]"

     # -- trigger the packet event
     Event_got_packet

   }
}

#
# -- perform a ping
proc ICMP_ping { stack dest { count 5 } { mode wait-for-response } } {
   upvar $stack s

   Stack_add_callback s icmpType 00 ICMP_ping_response
  
   set to 0
   set response 0

   for { set i 1 } { $i <= $count } { incr i } {

      # -- build an ICMP echo request
      set packet [ ICMP_build_echo_request s 01 $i ]
 
      # -- send the packet
      ICMP_send s $packet $dest

      # -- ping can just send the packets and not wait for a response
      if { "$mode" != "wait-for-response" } {
         continue
      }

      # -- wait 3 seconds
      set result [ Event_wait_packet 3000 ]

      if { "$result" == "1" } {
         pktsrcWarn s "Time out waiting for ping response from $dest"
         incr to
      } else {
         incr response
      }
   }

   Stack_remove_callback s icmpType 00 ICMP_ping_response

   return $response
}


proc ICMP_ping_disable { stack } {
   upvar $stack s

   pktsrcInfo s "Disable ICMP ping responses for this host"
   set s(ICMP,pingResponse,mode) off

}

proc ICMP_ping_enable { stack } {
   upvar $stack s

   pktsrcInfo s "Enable ICMP ping responses for this host"
   set s(ICMP,pingResponse,mode) on

}
