#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# 2001 QAcafe http://qacafe.com
#
# $Id: ip-arp.tcl,v 1.12 2002/11/15 13:23:54 joe Exp $
#
#
# Basic IPv4 Arp implementation
#
# ----------------------------------------------------------------------------

# -- this is part of the pktsrc package
package provide pktsrc 1.3

proc ARP_input { packet stack {options none }} {
   upvar $stack s

   set port $s(port)
   ARP_decode $packet arp

   if { $options == "print" } {
       ARP_print arp
     
       # that's it for printing
       return
   }

   if { $arp(opcode) == "0001" } {
      pktsrcInfo s "ARP request, who is [ hextoip $arp(protoDst) ], tell [ hextoip $arp(protoSrc)]"
   }

   # -- put the new entry in the cache
   ARP_add_entry s $arp(protoSrc) $arp(hwSrc)  

   # -- only respond if arp is enabled
   if { "$s(ARP,status)" == "on" } {

      if { $arp(opcode) == "0001" && $arp(protoDst) == $s(ip) } {

         # -- this is for me
         ARP_send_reply_self s $port $arp(protoSrc) $arp(hwSrc)

      } else {

         # -- search my proxy list too!
         if [ info exists s(proxyarp,$arp(protoDst)) ] {
            ARP_send_reply_other s $port $arp(protoSrc) $arp(hwSrc) $arp(protoDst)
         }
      }
   }

   # -- check for other callbacks
   if [ info exists s(arpType,$arp(opcode)) ] {

      # -- drive any input callbacks
      foreach c $s(arpType,$arp(opcode),callbacks) {
          # -- send the whole packet
          $c $packet s $options
      }
   }   
}

proc ARP_init { stack } {
   upvar $stack s

   # -- set an entry for myself
   set s(arpcache,$s(ip)) $s(mac)

   # -- add an entry for the broadcast address
   set s(arpcache,ffffffff) ffffffffffff

   # -- enable arp by default
   set s(ARP,status) on
}

proc ARP_decode { packet arp } {
   upvar $arp out

   # -- parse out all the fields we find interesting ...

   set out(hwType)   [ string range $packet 0 3 ]
   set out(proto)    [ string range $packet 4 7 ]
   set out(hwLen)    [ string range $packet 8 9 ]
   set out(protoLen) [ string range $packet 10 11 ]

   # -- opcode
   set out(opcode)   [ string range $packet 12 15 ]

   set out(hwSrc)    [ string range $packet 16 27 ]
   set out(protoSrc) [ string range $packet 28 35 ]
   set out(hwDst)    [ string range $packet 36 47 ]
   set out(protoDst) [ string range $packet 48 55 ]
   
   # -- no data/payload to set for ARP
    
}

proc ARP_print { arp } {
   upvar $arp out

   puts "ARP: ---- ARP Packet ----"
   puts "ARP: "
   puts "ARP: Hardware Type              = [ARP_hwType $out(hwType)]"
   puts "ARP: Protocol Type              = [ARP_protoType $out(proto)]"
   puts "ARP: Hardware Address Length    = $out(hwLen)"
   puts "ARP: Protocol Address Length    = $out(protoLen)"
   puts "ARP: Opcode                     = [ ARP_opcode $out(opcode) ] (0x$out(opcode))"
   puts "ARP: Src Hardware Address       = [hextomac $out(hwSrc)]"
   puts "ARP: Src Protocol Address       = [hextoip $out(protoSrc)]"
   puts "ARP: Dest Hardware Address      = [hextomac $out(hwDst)]"
   puts "ARP: Dest Protocol Address      = [hextoip $out(protoDst)]"
   puts "ARP:"
   puts "ARP: ---- End ARP Packet ----"

}

proc ARP_opcode { opcode } {
  
   if { $opcode == "0001" } {
      set str "ARP Request"
   } elseif { $opcode == "0002" } {
      set str "ARP Reply"
   } else {
      set str "Unknown Type ($opcode)"
   }  
   return $str
}

proc ARP_hwType { type } {
   if { $type == "0001" } {
      set str "Ethernet"
   } else {
      set str "Unknown Type ($type)"
   }
   return $str
}

proc ARP_protoType { type } {
   if { $type == "0800" } {
      set str "IPv4"
   } else {
      set str "Unknown Protocol ($type)"
   }
   return $str
}

proc ARP_static_entry { stack ip mac } {
   upvar $stack s
   ARP_add_entry s [ iptohex $ip ] $mac
}

proc ARP_add_entry { stack ip mac } {

   upvar $stack s

   # -- check for dups here
   set s(arpcache,$ip) $mac

   pktsrcInfo s "Adding ARP entry for [hextoip $ip] at [hextomac $mac]"

   # -- if there is a packet waiting to be sent, send it now
   if [ info exists s(arpBuffer) ] {

       # -- check for a match on the dest ip address
       if { "$ip" == "$s(arpBuffer,dest)" } { 

          set epack [ format "%s%s%s%s" $mac $s(mac) "0800" $s(arpBuffer) ]
          $s(port) send $epack
          Ethernet_log_send $epack s

          # -- clear the arp buffer
          unset s(arpBuffer)
          unset s(arpBuffer,dest)
       }
   }
}

#
# -- add an IP address for proxy apr

proc ARP_add_proxy { stack ip } {

   upvar $stack s

   # -- check for dups here
   set s(proxyarp,$ip) $ip
   pktsrcInfo s "Adding proxy ARP entry for [hextoip $ip]"
}


#
# -- send an ARP reply

proc ARP_send_reply_self { stack port ip mac } {
   upvar $stack s

   set eth [ format "%s%s%s" $mac $s(mac) "0806" ]

   set arp [ format "%s%s%s%s%s%s" $eth \
      "0001080006040002" $s(mac) $s(ip) \
      $mac $ip ]

   pktsrcInfo s "Sending ARP Reply [hextoip $s(ip)] is at [hextomac $s(mac)]"


   ARP_send $arp s
}

proc ARP_annouce { stack } {
   upvar $stack s

   set port $s(port)

   # -- send a request for myself
   ARP_send_request_self s $port $s(ip)

   # -- and all my proxies
   set n [ array names s proxyarp,* ]

   foreach ip $n {
     set proxyip $s($ip)
     ARP_send_request_self s $port $proxyip
   }
}

proc ARP_send_reply_other { stack port ip mac otherip } {
   upvar $stack s

   set eth [ format "%s%s%s" $mac $s(mac) "0806" ]

   set arp [ format "%s%s%s%s%s%s" $eth \
      "0001080006040002" $s(mac) $otherip \
      $mac $ip ]

   pktsrcInfo s "Sending ARP Reply [hextoip $otherip] is at [hextomac $s(mac)]"


   ARP_send $arp s
}


#
# -- send an ARP request

proc ARP_send_request_self { stack port ip } {
   upvar $stack s

   set eth [ format "ffffffffffff%s%s" $s(mac) "0806" ]

   set arp [ format "%s%s%s%s%s%s" $eth \
      "0001080006040001" $s(mac) $s(ip) \
      "000000000000" $ip ]

   pktsrcInfo s "Sending ARP Request for [hextoip $ip]"

   ARP_send $arp s

}

#
# -- get an entry from the cache or initiate an ARP request

proc ARP_get { stack iphex } {
   upvar $stack s

   set port $s(port)
   set entry ""

   if [ info exist s(arpcache,$iphex) ] {
     set entry $s(arpcache,$iphex)
   } else {
     ARP_send_request_self s $port $iphex
   } 

   return $entry

}

#
# ARP_get_wait
#
# Blocking arp call
#  - delete any existing entry for this address
#  - send out arp request for this entry
#  - return response or timeout
#

proc ARP_get_wait { stack ip { timeOut 5000 }} {
   upvar $stack s

   set iphex [ iptohex $ip ]
   set port $s(port)
   set entry ""

   if [ info exist s(arpcache,$iphex) ] {
     unset s(arpcache,$iphex)
   }

   # -- add callback for arp responses
   Stack_add_callback s arpType 0002 ARP_generic_input

   ARP_send_request_self s $port $iphex

   set result [ Event_wait_packet $timeOut ]

   # -- should check response and see if it matches
   
   if { "$result" == "1" } {
     pktsrcInfo s "timeout waiting for ARP response from $ip"
   }

   # -- return arp entry if it now exists
   if [ info exist s(arpcache,$iphex) ] {
      set entry $s(arpcache,$iphex)
   }

   # -- remove callback for arp reponses
   Stack_remove_callback s arpType 0002 ARP_generic_input

   return $entry

}

proc ARP_generic_input { packet stack { options none }} {

   if { "$options" != "print" } {
      Event_got_packet
   }
}

proc ARP_send { packet stack } {
   upvar $stack s
 
   set port $s(port)

   # -- handle vlans
   if [ info exists s(Ethernet,802.1q) ] {

      set header [ string range $packet 0 23 ]
      set rest [ string range $packet 24 end ]
      set packet [ format "%s%s%s" $header $s(Ethernet,802.1q) $rest ]

   }

   # -- actually send the packet
   $port send $packet

   # -- log the send
   Ethernet_log_send $packet s

}

proc ARP_disable { stack } {
   upvar $stack s
   set s(ARP,status) off
}

proc ARP_enable { stack } {
   upvar $stack s
   set s(ARP,status) on
}






