#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# 2001 QAcafe http://qacafe.com
#
# $Id: ip.tcl,v 1.23 2002/12/10 22:28:33 joe Exp $
#
#
# Basic IPv4 Support
#
# ----------------------------------------------------------------------------

# -- this is part of the pktsrc package
package provide pktsrc 1.3

proc IPv4_input { packet stack { options none }} {
   upvar $stack s



   if { $options == "print" } {

      IPv4_decode $packet ip
      IPv4_print ip

   } else {

      # -- remember this packet in the 'ip' variable. This allows other
      # callbacks to get access to the decodes at any Tcl stack level

      eval [ Stack_save_packet s IPv4 ip ]

      IPv4_decode $packet ip

   }

   # -- check for callbacks at the IP level

   # -- first make sure this IP packet is for me
   # TBD

   # -- see if this packet is for a vnode
   if { $options != "print" } {
     if [ info exists s(vnode,$ip(dst),stack) ] {

       # -- send to vnode
       set stackName $s(vnode,$ip(dst),stackname)
       set stackVar $s(vnode,$ip(dst),stack)

       eval $stackVar
       # pktsrcInfo s "Route packet to vnode -> $stackName"

       # -- drive the IP input for the vnode stack
       IPv4_input $packet $stackName $options

       return
     }
   }

   if [ info exists s(ipProtocol,$ip(protocol)) ] {

      # -- drive any input callbacks
      foreach c $s(ipProtocol,$ip(protocol),callbacks) {
          # $s(ipProtocol,$ip(protocol),input) $ip(data) s $options
          $c $ip(data) s $options
      }
   } else {
      # no callback
   }

   if { "$options" != "print" } {

     # -- now see about reassembling
     set fullIp [ IPv4_reassemble s ]
 
     if { "$fullIp" != "" } {

        if [ info exists s(ipReAssembled,$ip(protocol)) ] {

           # -- drive any input callbacks
           foreach c $s(ipReAssembled,$ip(protocol),callbacks) {
               $c $fullIp s $options
           }
        }
     }
   } 

   # -- hmmm could add stacks like ...
   # incr s(IPv4,incount);

}

proc IPv4_init { stack } {

   upvar $stack s

   # -- set up default options
   set s(IPv4,tos) 00
   set s(IPv4,ttl) ff
   set s(IPv4,id)  1
   set s(IPv4,mtu) 1500

   set s(IPv4,dataMtu) [ expr $s(IPv4,mtu) - 20 ]\

   # -- default gateway is not set
   set s(IPv4,defaultGateway)  00000000

   # -- turn on checksums
   set s(IPv4,checksum) on
}

proc IPv4_decode { packet ip } {
   upvar $ip out

   # -- parse out all the fields we find interesting ...

   set v              [ string range $packet 0 1 ]
   set out(version)   [ expr (0xf0 & 0x$v) >> 4 ]
   set out(hdLen)     [ expr 0x0f & 0x$v ]

   set out(tos)       [ string range $packet 2 3 ]
   set out(length)    [ string range $packet 4 7 ]
   set out(id)        [ string range $packet 8 11 ]
   set out(flags)     [ string range $packet 12 15 ]
   set out(ttl)       [ string range $packet 16 17 ]
   set out(protocol)  [ string range $packet 18 19 ]
   set out(checksum)  [ string range $packet 20 23 ]
   set out(src)       [ string range $packet 24 31 ]
   set out(dst)       [ string range $packet 32 39 ]

   # -- the payload, use IP length to determine end of payload
   # some hosts may include trailing data that is not part of IP payload

   set pkend [ expr 0x$out(length) * 2 -1 ]
   set out(data)      [ string range $packet 40 $pkend ]
    
}

proc IPv4_print { ip } {
   upvar $ip out

   puts "IPv4: ---- IPv4 Header ----"
   puts "IPv4: "
   puts "IPv4: Version                   = $out(version)"
   puts "IPv4: Header Length             = $out(hdLen)"
   puts "IPv4: TOS byte                  = 0x$out(tos)"
   puts "IPv4: Total Length              = [expr 0x$out(length)]"
   puts "IPv4: ID                        = 0x$out(id)"
   puts "IPv4: Flags/Offset              = 0x$out(flags)"

   if { [ expr 0x4000 & 0x$out(flags) ] } {
      set df 1
   } else {
      set df 0
   }

   if { [ expr 0x2000 & 0x$out(flags) ] } {
      set mf 1
   } else {
      set mf 0
   }

   puts "IPv4:   Don't Fragment          = $df"
   puts "IPv4:   More Fragments          = $mf"

   puts "IPv4:   Offset                  = [ expr 0x1fff & 0x$out(flags)]"
   puts "IPv4: TTL                       = [expr 0x$out(ttl)]"
   puts "IPv4: Protocol                  = [IPv4_proto $out(protocol)] (0x$out(protocol))"
   puts "IPv4: Checksum                  = 0x$out(checksum)"
   puts "IPv4: Src                       = [hextoip $out(src)]"
   puts "IPv4: Dest                      = [hextoip $out(dst)]"
   puts "IPv4:"


}

proc IPv4_proto { type } {

   if { $type == "01" } {
      set str "ICMP"
   } elseif { $type == "02" } {
      set str "IGMP"
   } elseif { $type == "04" } {
      set str "IP-IP"
   } elseif { $type == "06" } {
      set str "TCP"
   } elseif { $type == "11" } {
      set str "UDP"
   } elseif { $type == "32" } {
      set str "IPSEC"
   } elseif { $type == "59" } {
      set str "OSPF"
   } elseif { $type == "70" } {
      set str "VRRP"
   } else {
      set str "Unknown Protocol [expr 0x$type]"
   }

   return $str
}

proc IPv4_reassemble { stack } {
   upvar $stack s

   # -- get the current IP contents
   set ipState $s(current,IPv4)
   upvar #$s(current,IPv4,level) $ipState ip

   set offset [ expr (0x1fff & 0x$ip(flags)) * 8]
   set mf [ expr 0x2000 & 0x$ip(flags) ]

   if { "$offset" == "0"  && $mf == "0" } {
      return $ip(data)
   } else {
 
      if ![ info exists s(IPv4,reassemble,$ip(id),chunks) ] {
         set s(IPv4,reassemble,$ip(id),chunks) 1
      } else {
         incr s(IPv4,reassemble,$ip(id),chunks)
      }

      set chunk $s(IPv4,reassemble,$ip(id),chunks)
      pktsrcInfo s "IPv4 reassemble of id $ip(id), adding chunk $chunk, offset $offset, flags 0x$ip(flags)"

      set s(IPv4,reassemble,$ip(id),$chunk) $ip(data)
      set s(IPv4,reassemble,$ip(id),$chunk,$mf) $mf

      # -- reassemble packet

      # assumptions
      #
      # packets are in order
      # no packet loss
      #

      if { "$mf" == "0" } {

          pktsrcInfo s "last packet for id $ip(id), reassemble $chunk chunks"
          set packet ""
	  for { set i 1 } { $i <= $chunk } { incr i } {
             set packet "$packet$s(IPv4,reassemble,$ip(id),$i)"
             unset s(IPv4,reassemble,$ip(id),$i)
          }
          unset s(IPv4,reassemble,$ip(id),chunks)
          return $packet
      }
   }
}

proc IPv4_reconfigure { stack ip mask } {
   upvar $stack s

   set s(ip) [iptohex $ip ] 
   set s(netmask) [ iptohex $mask ]
   set s(network) [ format "%8.8x" [ expr 0x$s(netmask) & 0x$s(ip) ]]

   # -- gratuitious arp
   ARP_annouce s
}

proc IPv4_get { stack field } {
   upvar $stack s

   if { "$field" == "address" } {
      return [ hextoip $s(ip) ]
   }

   error "Unsupported field in IPv4_get '$field'"

}

proc IPv4_set { stack field value } {
   upvar $stack s

   # -- set MTU and check
   if { "$field" == "mtu" } {

       pktsrcInfo s "setting IPv4 MTU to $value"
       set s(IPv4,mtu) $value

       set s(IPv4,dataMtu) [ expr $s(IPv4,mtu) - 20 ]
       if { $s(IPv4,dataMtu) <= 0 } {
          error "bad mtu size, must be greater than 20"
       }

       if { [ expr $s(IPv4,dataMtu) % 8 ] != 0 } {
           error "Sorry, data MTU must be multiple of 8 bytes ((mtu-20) % 8 = 0)"
       }

       return $value

   } elseif { "$field" == "default" || "$field" == "default-gateway" } {

      pktsrcInfo s "setting default gateway to $value"
      set network [iptohex [ ip_network_address $value [hextoip $s(netmask)] ]]

      if { "$network" != "$s(network)" } {

          pktsrcWarn s "Can not set default-gateway for [hextoip $s(ip)]"
          pktsrcWarn s \
              "default-gateway $value is not on network [hextoip $s(network)]"
          return 0

      } else {

         # -- validate default is on network
         set s(IPv4,defaultGateway) [iptohex $value]

         # -- arp for gateway
         ARP_get s $s(IPv4,defaultGateway)
         return $value
      }

   } else {
     error "Unknown IPv4 field '$field' in IPv4_set"
   }
}

proc IPv4_send { packet stack dest proto } {
   upvar $stack s

   incr s(IPv4,id)

   # -- handle id wrapping
   if { $s(IPv4,id) > 65535 } {
      set s(IPv4,id) 1
   }

   # -- now build the ID in hex
   set idhex [ format "%4.4x" $s(IPv4,id) ]
   
   set lengthDec [expr [ string length $packet ] / 2]

   set offset 0
   set flags 0000

   if { $lengthDec > $s(IPv4,dataMtu) } {
    pktsrcInfo s "Fragmenting IP with length $lengthDec (MTU is $s(IPv4,mtu))"

      while { $offset < $lengthDec } {

        pktsrcInfo s "Building Fragment with offset $offset"
        set fragStart [ expr ($offset * 2) ]
        set fragEnd [ expr ($offset + $s(IPv4,dataMtu)) * 2 - 1 ]
        set fragData [ string range $packet $fragStart $fragEnd ]
        set fragLen [ expr [ string length $fragData ] /2 ]

        set flags [ expr $offset / 8 ]


        incr offset $s(IPv4,dataMtu)


        if { $offset >= $lengthDec } {
          set flags [ format "%4.4x" [ expr $flags | 0x0000 ]]
        } else {
          set flags [ format "%4.4x" [ expr $flags | 0x2000 ]]
        }

        pktsrcInfo s "Sending IPv4 Fragment length $fragLen, flags 0x$flags"
        IPv4_send_process  $fragData s $dest $proto $idhex $flags

        # -- let the event loop run while we are sending
        Event_wait 1

     }
   } else {
      IPv4_send_process  $packet s $dest $proto $idhex 0000
   }

}

proc IPv4_send_process { packet stack dest proto idhex flags } {
   upvar $stack s

   # -- compute length of IPv4 packet
   set length [ format "%4.4x" [expr [ string length $packet ] / 2 + 20 ] ]

   # -- rebuild packet with length and tos
   set ip "45${s(IPv4,tos)}${length}${idhex}${flags}"

   set ipheader [ format "%s%s%s0000%s%s" \
       $ip $s(IPv4,ttl) $proto $s(ip) $dest ]

   # -- now compute checksum
   set checksum [ pktsrc checksum $ipheader ]

   # -- check for other checksum options
   if { "$s(IPv4,checksum)" == "bad" } {

      # -- use a bad checksum, increment the checksum by 1
      set checksum [ incrhex16 $checksum ]

   } elseif { "$s(IPv4,checksum)" == "off" } {
  
      # -- don't set the checksum
      set checksum 0000
   }
 
   set ippacket [ format "%s%s%s%s%s%s%s" \
              $ip $s(IPv4,ttl) $proto $checksum $s(ip) $dest $packet ]

   # -- send this to Ethernet (this uses ip-send)

   set network [format "%8.8x" [ expr 0x$dest & 0x$s(netmask) ]]

   # -- send directly if ....

   # desintation is on network
   # multicast address

   # there are many tcl issues trying to do unsigned comparisons, so
   # just convert the multicast bits to a hex string in order to
   # check for a multicast packet

   set mcastbits [ format "%8.8x" [ expr 0x$dest & 0xe0000000 ]]

   # pppoe hack. For now, if we have a ppp session, send all IP data
   # back out PPPoE

   if { "$s(NCP,state)" == "up" } {

      # -- send PPP packet
      PPP_send s $ippacket 0021

   } elseif [ info exists s(vnode,gateway,stack) ] {

      # -- this is a vnode, packet should go back to vnode gateway
      set stackName $s(vnode,gateway,stackname)
      set stackVar $s(vnode,gateway,stack)

      # -- get the context of the other stack
      eval $stackVar

      # pktsrcInfo s "route packet back to gateway stack $stackName"

      # -- now I have the other stack
      IPv4_route_vnode s $stackName $ippacket $dest
 
   } elseif { "$network" == "$s(network)" || "$mcastbits" == "e0000000" } {
      Ethernet_IPv4_send $ippacket s $dest
   } else {
      if { $s(IPv4,defaultGateway) == 00000000 } {
        pktsrcWarn s "Droping packet to [hextoip $dest], no gateway set"
      } else {
        Ethernet_IPv4_send $ippacket s $s(IPv4,defaultGateway)
      }
   }
}

proc IPv4_route_vnode { stack pstack packet dest } {
   upvar $stack s
   upvar $pstack p

   set network [format "%8.8x" [ expr 0x$dest & 0x$p(netmask) ]]
   set mcastbits [ format "%8.8x" [ expr 0x$dest & 0xe0000000 ]]

   # -- check for PPPoE
   if { "$p(NCP,state)" == "up" } {

      # -- send PPP packet
      PPP_send p $packet 0021

   } elseif { "$network" == "$p(network)" || "$mcastbits" == "e0000000" } {
      Ethernet_IPv4_send $packet p $dest
   } else {
      if { $p(IPv4,defaultGateway) == 00000000 } {
        pktsrcWarn p "Droping packet to [hextoip $dest], no gateway set"
      } else {
        Ethernet_IPv4_send $packet p $p(IPv4,defaultGateway)
      }
   }
}

#
# IPv4 checksum control

proc IPv4_checksum_bad { stack } {
   upvar $stack s
   set s(IPv4,checksum) bad
}

proc IPv4_checksum_on { stack } {
   upvar $stack s
   set s(IPv4,checksum) on
}

proc IPv4_checksum_off { stack } {
   upvar $stack s
   set s(IPv4,checksum) off
}

#
# IPv4_get_header 
#
#
# return the IP header for the current packet on the stack
#

proc IPv4_get_header { stack } {
   upvar $stack s

   # -- look at the ethernet header
   set ethState $s(current,Ethernet)
   upvar #$s(current,Ethernet,level) $ethState eth

   if { "$eth(type)" == "8864" } {
      set header [ string range $eth(data) 16 55 ]
   } else { 
      set header [ string range $eth(data) 0 39 ]
   }

   return $header
}


#
# IPv4_router_neighbor
#
# Send all IP packets using this neighbor
#

proc IPv4_route_neighbor { stack neighborStack } {
   upvar $stack s
   upvar neighborStack neighbor

   set s(IPv4,neighborStack) $neighbor

}

#
# IPv4_server_input
#
# input process for IPv4 packets
#
#  - no data buffer is done here like tcp
#  - call any session callbacks for udp


proc IPv4_server_input { packet stack {options none}} {
   upvar $stack s

   if { $options == "print" } {
      return
   }

   # -- get the current IP contents
   set ipState $s(current,IPv4)
   upvar #$s(current,IPv4,level) $ipState ip

   pktsrcInfo s "[hextoip $ip(dst)] received proto [ expr 0x$ip(protocol) ] from [hextoip $ip(src)]"

   if [ info exists s(ipv4SessionProtocol,$ip(protocol)) ] {

      # -- drive any input callbacks
      foreach c $s(ipv4SessionProtocol,$ip(protocol),callbacks) {
          $c $packet s $options
      }
   }
}

#
# IPv4_server
#
# Listen for IPv4 on specific protocol
#

proc IPv4_server { stack protocol { sessionCallback "" }} {
   upvar $stack s

   # -- get the protocol in hex
   set proto [ format "%2.2x" $protocol ]

   # -- add a callback for this server
   Stack_add_callback s ipReAssembled $proto IPv4_server_input

   if { "$sessionCallback" != "" } {
     Stack_add_callback s ipv4SessionProtocol $proto $sessionCallback
   }

   pktsrcInfo s "Accepting IPv4 protocol $protocol packets"
}

#
# IPv4_server_shutdown
#
# Shutdown down IPv4 server on a protocol
#

proc IPv4_server_shutdown { stack protocol { sessionCallback "" }} {

   upvar $stack s

   set proto [ format "%2.2x" $protocol ]

   # -- add a callback for this server
   Stack_remove_callback s ipReAssembled $proto IPv4_server_input

   if { "$sessionCallback" != "" } {
     Stack_remove_callback s ipv4SessionProtocol $proto $sessionCallback
   }

   pktsrcInfo s "IPv4 server on protocol $protocol has shutdown"
}


proc IPv4_echo_server { stack protocol } {
   upvar $stack s

   pktsrcInfo s "Starting up IP echo server for protocol $protocol"

   IPv4_server s $protocol IPv4_echo_server_input

}

proc IPv4_echo_server_shutdown { stack protocol } {
   upvar $stack s

   pktsrcInfo s "Shutting down IPv4 echo server for protocol $protocol"

   IPv4_server_shutdown s $protocol IPv4_echo_server_input

}


#
# IPv4_echo_server_input
#
#

proc IPv4_echo_server_input { packet stack {options none}} {
   upvar $stack s

   if { "$options" == "print" } {
       return
   }

   # -- get the current IP contents
   set ipState $s(current,IPv4)
   upvar #$s(current,IPv4,level) $ipState ip

   # -- just send a dump response back
   pktsrcInfo s "Responding to IPv4 protocol [ expr 0x$ip(protocol) ] from [hextoip $ip(src)]"

   set response "$packet"

   IPv4_send $response s $ip(src) $ip(protocol)  

}

proc IPv4_echo_protocol { stack server protocol { bytes 1500 } } {
   upvar $stack s

   set proto [ format "%2.2x" $protocol ]
   set length [ expr $bytes - 20 ]

   pktsrcInfo s "Sending IP protocol $protocol to $server ($bytes bytes)"

   # -- send request
   set request [ string repeat a7 $length ]

   IPv4_send "$request" s [iptohex $server] $proto 

   set result [ IPv4_get_reassembled_packet s $proto ip 10 ] 

   if { "$result" != "-1" } {
      pktsrcInfo s "got IPv4 proto $protocol response from $server"

      # -- check length
      set inl [ expr [ string length $s(IPv4,reassembled) ] / 2 ]

      if { "$inl" == "$length" } {
          pktsrcInfo s "sent $length data bytes, received $inl"
          return "okay"
      } else {
          pktsrcWarn s "sent $length data bytes but only received $inl"
          return ""
      }
   } else {
      pktsrcWarn s "no IPv4 proto $protocol response from $server after 10 seconds"
      return ""
   }
}


#
# IPv4_generic_input
#
# generic callback for catching IPv4 packets
#

proc IPv4_generic_input { packet stack { options none }} {
   upvar $stack s

   if { "$options" != "print" } {

      # -- remember the reassembled $packet
      set s(IPv4,reassembled) $packet

      # -- trigger the packet event
      Event_got_packet
   }
}

#
# IPv4_get_packet
#
# blocking call for receiving IPv4 packets
# this call blocks until the packet is received or a timeout occurs
#

proc IPv4_get_packet { stack protocol packet timeout } {
   upvar $stack s

   set ret -1

   # add a callback for IPv4 packets for this protocol
   Stack_add_callback s ipProtocol $protocol IPv4_generic_input

   pktsrcInfo s \
     "Waiting up to $timeout seconds for IPv4 protocol [expr 0x$protocol]"

   # -- compute timeout in milli seconds
   set msec [ expr $timeout * 1000 ]

   # -- now wait for the packet or timeout
   set result [ Event_wait_packet $msec  ]

   #
   # result = 2 -> got packet
   #

   if { "$result" == "2" } {

      # -- get the contents of the IPv4 packet from the stack
      set ipState $s(current,IPv4)
      upvar #$s(current,IPv4,level) $ipState ip
      upvar $packet p

      # -- copy the packet to pass back to the user
      pktsrc_copy_packet ip p

      set ret 0


   #
   # result = 1 -> timeout
   #

   } elseif { "$result" == "1" } {

      
      pktsrcInfo s \
         "timeout waiting for IPv4 packet at protocol [expr 0x$protocol]"

      set ret -1

   } else {

      set ret -1
   }

   # -- removing the callback
   Stack_remove_callback s ipProtocol $protocol IPv4_generic_input

   return $ret
}

proc IPv4_get_reassembled_packet { stack protocol packet timeout } {
   upvar $stack s

   set ret -1

   # add a callback for IPv4 packets for this protocol
   Stack_add_callback s ipReAssembled $protocol IPv4_generic_input

   pktsrcInfo s \
     "Waiting up to $timeout seconds for IPv4 protocol [expr 0x$protocol]"

   # -- compute timeout in milli seconds
   set msec [ expr $timeout * 1000 ]

   # -- now wait for the packet or timeout
   set result [ Event_wait_packet $msec  ]

   #
   # result = 2 -> got packet
   #

   if { "$result" == "2" } {

      # -- get the contents of the IPv4 packet from the stack
      set ipState $s(current,IPv4)
      upvar #$s(current,IPv4,level) $ipState ip
      upvar $packet p

      # -- copy the packet to pass back to the user
      pktsrc_copy_packet ip p

      set ret 0


   #
   # result = 1 -> timeout
   #

   } elseif { "$result" == "1" } {

      
      pktsrcInfo s \
         "timeout waiting for IPv4 packet at protocol [expr 0x$protocol]"

      set ret -1

   } else {

      set ret -1
   }

   # -- removing the callback
   Stack_remove_callback s ipReAssembled $protocol IPv4_generic_input

   return $ret
}







