#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# 2002 QAcafe http://qacafe.com
#
# $Id: ipfwd.tcl,v 1.12 2002/12/10 15:03:21 joe Exp $
#
#
# Supporting utilities for IP forwarding related functions
#
#
# ----------------------------------------------------------------------------

# -- this is part of the pktsrc package
package provide pktsrc 1.3

#
# IPv4_FWD_udp_test
#
# forward udp packet from stack1 to stack
#
# if connection is outbound through NAT, allow the expected NAT address
#
#

proc IPv4_FWD_udp_test { stack1 stack2 destPort length { natAddress 0.0.0.0 } { timeOut 5000 } } {
   upvar $stack1 s1
   upvar $stack2 s2
   set ret 0
 
   # -- build the data string
   set packet [ string repeat a7 $length ] 

   # -- compute the expected udp length
   set sendLen [ expr $length + 8 ]

   # -- setup a handler on the dest port
   set portHex [ format "%4.4x" $destPort ]
   Stack_add_callback s2 udpPort $portHex IPv4_FWD_udp_test_callback

   # -- send a udp packet
   UDP_send s1 $packet $s2(ip) $destPort 1201   

   # -- now wait for the packet or timeout
   set result [ Event_wait_packet $timeOut ]

   #
   # result = 2 -> got packet
   #

   if { "$result" == "2" } {

      # -- get the contents of the UDP packet from the stack
      set udpState $s2(current,UDP)
      set ipState $s2(current,IPv4)
      upvar #$s2(current,UDP,level) $udpState udp
      upvar #$s2(current,IPv4,level) $ipState ip
      set ethState $s2(current,Ethernet)
      upvar #$s2(current,Ethernet,level) $ethState eth

      set inLen [ expr 0x$udp(length) ]
      pktsrcInfo s2 "Received udp packet with length $inLen from [hextoip $ip(src)]"

      # -- perform checks

      #
      # -- Check 1, ip src address should match expected source address

      if { "$natAddress" == "0.0.0.0" } {
         set expectedIp $s1(ip)
      } elseif { "$natAddress" != "any" } {
         set expectedIp [ iptohex $natAddress ]
      }


      if { "$natAddress" == "any" } {
        pktsrcInfo s2 "Received IP packet with source address [hextoip $ip(src)]"

      } elseif { "$ip(src)" != "$expectedIp" } {

        pktsrcWarn s2 "found unexpected IP source address [ hextoip $ip(src) ]"
        pktsrcWarn s2 "expecting IP source address [ hextoip $expectedIp ]"
        set ret -4
      }

      #
      # -- Check 2, verify length
     
      if { "$inLen" != "$sendLen" } {
        pktsrcWarn s2 "Received udp length $inLen but sent $sendLen"
        set ret -3
      }

      #
      # -- Verify ip check sum

      set header [ IPv4_get_header s2 ]
      set checksum [ pktsrc checksum $header ]

      if { [ string compare "$checksum" "0000" ] } {
         pktsrcWarn s2 "Received invalid IPv4 checksum from [ hextoip $ip(src) ]"
         set ret -10
      }

      #
      # -- Verify udp checksum
      set checksum [ UDP_compute_checksum s2 ]
      if { [ string compare "$checksum" "0000" ] } {
         pktsrcWarn s2 "Received invalid UDP checksum from [ hextoip $ip(src) ]"
         set ret -11
      }

   #
   # result = 1 -> timeout
   #

   } elseif { "$result" == "1" } {

      pktsrcInfo s2 "$timeOut msec timeout waiting for udp packet"
      set ret -1

   } else {

      set ret -1
   }

   # -- removing the callback
   Stack_remove_callback s2 udpPort $portHex IPv4_FWD_udp_test_callback

   return $ret
}


#
# IPv4_FWD_udp_nat_test
#
# forward udp packet from stack1 to stack2
#
#
#

proc IPv4_FWD_udp_nat_test { stack1 stack2 srcPort natDestPort natDestIp destPort length { timeOut 5000 } } {
   upvar $stack1 s1
   upvar $stack2 s2
   set ret 0
 
   # -- build the data string
   set packet [ string repeat a7 $length ] 

   # -- compute the expected udp length
   set sendLen [ expr $length + 8 ]

   # -- setup a handler on the dest port
   set portHex [ format "%4.4x" $destPort ]
   Stack_add_callback s2 udpPort $portHex IPv4_FWD_udp_test_callback

   # -- send a udp packet
   UDP_send s1 $packet [iptohex $natDestIp] $natDestPort $srcPort   

   # -- now wait for the packet or timeout
   set result [ Event_wait_packet $timeOut ]

   #
   # result = 2 -> got packet
   #

   if { "$result" == "2" } {

      # -- get the contents of the UDP packet from the stack
      set udpState $s2(current,UDP)
      set ipState $s2(current,IPv4)
      upvar #$s2(current,UDP,level) $udpState udp
      upvar #$s2(current,IPv4,level) $ipState ip
      set ethState $s2(current,Ethernet)
      upvar #$s2(current,Ethernet,level) $ethState eth

      set inLen [ expr 0x$udp(length) ]
      pktsrcInfo s2 "Received udp packet with length $inLen from [hextoip $ip(src)]"

      # -- perform checks

      #
      # -- Check 1, ip src address should match expected source address

      set expectedIp $s1(ip)
      if { "$ip(src)" != "$expectedIp" } {
        pktsrcWarn s2 "found unexpected IP source address [ hextoip $ip(src) ]"
        pktsrcWarn s2 "expecting IP source address [ hextoip $expectedIp ]"
        set ret -4
      }

      #
      # -- Check 2, verify length
     
      if { "$inLen" != "$sendLen" } {
         pktsrcWarn s2 "Received udp length $inLen but sent $sendLen"
         set ret -3
      }



      #
      # -- Verify ip check sum

      set header [ string range $eth(data) 0 39 ]
      set checksum [ pktsrc checksum $header ]

      if { [ string compare "$checksum" "0000" ] } {
         pktsrcWarn s2 "Received invalid IPv4 checksum from [ hextoip $ip(src) ]"
         set ret -11
      }

      #
      # -- Verify udp checksum
      set checksum [ UDP_compute_checksum s2 ]
      if { [ string compare "$checksum" "0000" ] } {
         pktsrcWarn s2 "Received invalid UDP checksum from [ hextoip $ip(src) ]"
         set ret -11
      }

   #
   # result = 1 -> timeout
   #

   } elseif { "$result" == "1" } {

      pktsrcInfo s2 "$timeOut msec timeout waiting for udp packet"
      set ret -1

   } else {

      set ret -1
   }

   # -- removing the callback
   Stack_remove_callback s2 udpPort $portHex IPv4_FWD_udp_test_callback

   return $ret
}

proc IPv4_FWD_udp_test_callback { packet stack { options none } } {
   
   if { "$options" != "print" } {
       # -- trigger the packet event
       Event_got_packet
   }
}


#
# IPv4_FWD_tcp_scan_test
#
# forward scan packet from stack1 to stack2
# look for natted packet or unnatted packet
#
#

proc IPv4_FWD_tcp_scan_test { stack1 stack2 srcPort natDestPort natDestIp { timeOut 1000 } } {
   upvar $stack1 s1
   upvar $stack2 s2
   set ret 0
 

   # -- setup a handler on the dest port
   set portHex [ format "%4.4x" $natDestPort ]

   # -- we attach at the TCP protocol layer so that we will get all TCP
   # packets. This handles the case where port forwarding is sending
   # packets to ports we are not expecting.

   Stack_add_callback s2 ipProtocol 06 IPv4_FWD_udp_test_callback

   TCP_init_session s1 sess

   set sess(srcPort) [ format "%4.4x" $srcPort ]
   set sess(dstPort) [ format "%4.4x" $natDestPort ]
   set sess(dstIp) [ iptohex $natDestIp ]

   # -- send a tcp-syn packet
   TCP_send_state s1 sess 02 syn-sent


   while { 1 } {

      # -- now wait for the packet or timeout
      set result [ Event_wait_packet $timeOut ]

      #
      # result = 2 -> got packet
      #

      if { "$result" == "2" } {

         # -- get the contents of the TCP packet from the stack
         set tcpState $s2(current,TCP)
         set ipState $s2(current,IPv4)
         upvar #$s2(current,TCP,level) $tcpState tcp
         upvar #$s2(current,IPv4,level) $ipState ip


         pktsrcInfo s2 "Received tcp packet from [hextoip $ip(src)] at port [ expr 0x$tcp(dstPort)]"
	 if { "$ip(src)" == "$s1(ip)" } {

            # -- check IP checksum
            set header [ IPv4_get_header s2 ]
            set checksum [ pktsrc checksum $header ]

            if { [ string compare "$checksum" "0000" ] } {
               pktsrcWarn s2 \
                  "Received invalid IPv4 checksum from [ hextoip $ip(src) ]"
               set ret -11
               break
            }

            # -- check TCP checksum
            set checksum [ TCP_compute_checksum s2 ]
            if { [ string compare "$checksum" "0000" ] } {
               pktsrcWarn s2 \
                  "Received invalid TCP checksum from [ hextoip $ip(src) ]"
               set ret -10
               break
            } 

            break
         } else {
            continue
         }


      #
      # result = 1 -> timeout
      #

      } elseif { "$result" == "1" } {

         pktsrcInfo s2 "$timeOut msec timeout waiting for tcp packet"
         set ret -1
         break
      } else {
         break
         set ret -1
      }
   }

   # -- removing the callback
   Stack_remove_callback s2 ipProtocol 06 IPv4_FWD_udp_test_callback

   return $ret
}

proc IPv4_FWD_udp_scan_test { stack1 stack2 srcPort natDestPort natDestIp { timeOut 1000 } } {
   upvar $stack1 s1
   upvar $stack2 s2
   set ret 0
 
   # -- setup a handler on the dest port
   set portHex [ format "%4.4x" $natDestPort ]

   # -- we attach at the UDP protocol layer so that we will get all TCP
   # packets. This handles the case where port forwarding is sending
   # packets to ports we are not expecting.

   Stack_add_callback s2 ipProtocol 11 IPv4_FWD_udp_test_callback

   # -- build the data string
   set packet [ string repeat a7 64 ] 

   # -- send a udp packet
   UDP_send s1 $packet [iptohex $natDestIp] $natDestPort $srcPort

   while { 1 } {

      # -- now wait for the packet or timeout
      set result [ Event_wait_packet $timeOut ]

      #
      # result = 2 -> got packet
      #

      if { "$result" == "2" } {

        # -- get the contents of the IP packet from the stack
        set ipState $s2(current,IPv4)
        upvar #$s2(current,IPv4,level) $ipState ip

        set udpState $s2(current,UDP)
        upvar #$s2(current,UDP,level) $udpState udp

        pktsrcInfo s2 "Received udp packet from [hextoip $ip(src)] at port [ expr 0x$udp(dstPort)]"




	if { "$ip(src)" == "$s1(ip)" } {


           #
           # -- Verify ip check sum

           set header [ IPv4_get_header s2 ]
           set checksum [ pktsrc checksum $header ]

           if { [ string compare "$checksum" "0000" ] } {
              pktsrcWarn s2 \
                 "Received invalid IPv4 checksum from [ hextoip $ip(src) ]"
              set ret -10
           }

           #
           # -- Verify udp checksum
           set checksum [ UDP_compute_checksum s2 ]
           if { [ string compare "$checksum" "0000" ] } {
              pktsrcWarn s2 "Received invalid UDP checksum from [ hextoip $ip(src) ]"
              set ret -11
           }

          break
        } else {
          continue
        }

      #
      # result = 1 -> timeout
      #

      } elseif { "$result" == "1" } {

         pktsrcInfo s2 "$timeOut msec timeout waiting for udp packet from [hextoip $s1(ip)]"
         set ret -1
         break
      } else {
         set ret -1
         break
      }
   }

   # -- removing the callback
   Stack_remove_callback s2 ipProtocol 11 IPv4_FWD_udp_test_callback

   return $ret
}



proc IPv4_FWD_udp_test_router { stack1 stack2 destIP destPort length { natAddress 0.0.0.0 } { timeOut 5000 } } {
   upvar $stack1 s1
   upvar $stack2 s2
   set ret 0

   # -- build the data string
   set packet [ string repeat a7 $length ] 

   # -- compute the expected udp length
   set sendLen [ expr $length + 8 ]

   # -- setup a handler on the dest port
   set portHex [ format "%4.4x" $destPort ]
   Stack_add_callback s2 udpPort $portHex IPv4_FWD_udp_test_callback

   # -- send a udp packet
   UDP_send s1 $packet [ iptohex $destIP ] $destPort 1201   

   # -- now wait for the packet or timeout
   set result [ Event_wait_packet $timeOut ]

   #
   # result = 2 -> got packet
   #

   if { "$result" == "2" } {

      # -- get the contents of the UDP packet from the stack
      set udpState $s2(current,UDP)
      set ipState $s2(current,IPv4)
      upvar #$s2(current,UDP,level) $udpState udp
      upvar #$s2(current,IPv4,level) $ipState ip
      set ethState $s2(current,Ethernet)
      upvar #$s2(current,Ethernet,level) $ethState eth

      set inLen [ expr 0x$udp(length) ]
      pktsrcInfo s2 "Received udp packet with length $inLen from [hextoip $ip(src)]"

      # -- perform checks

      #
      # -- Check 1, ip src address should match expected source address

      if { "$natAddress" == "0.0.0.0" } {
         set expectedIp $s1(ip)
      } elseif { "$natAddress" != "any" } {
         set expectedIp [ iptohex $natAddress ]
      }

      if { "$natAddress" == "any" } {
        pktsrcInfo s2 "Received IP packet with source address [hextoip $ip(src)]"
      } elseif { "$ip(src)" != "$expectedIp" } {
        pktsrcWarn s2 "found unexpected IP source address [ hextoip $ip(src) ]"
        pktsrcWarn s2 "expecting IP source address [ hextoip $expectedIp ]"
        set ret -4
      }

      #
      # -- Check 2, verify length
     
      if { "$inLen" != "$sendLen" } {
        pktsrcWarn s2 "Received udp length $inLen but sent $sendLen"
        set ret -3
      }


      #
      # -- Verify ip check sum

      set header [ IPv4_get_header s2 ]
      set checksum [ pktsrc checksum $header ]

      if { [ string compare "$checksum" "0000" ] } {
         pktsrcWarn s2 "Received invalid IPv4 checksum from [ hextoip $ip(src) ]"
         set ret -10
      }

      #
      # -- Verify udp checksum
      set checksum [ UDP_compute_checksum s2 ]
      if { [ string compare "$checksum" "0000" ] } {
         pktsrcWarn s2 "Received invalid UDP checksum from [ hextoip $ip(src) ]"
         set ret -11
      }

      set ret 0

   #
   # result = 1 -> timeout
   #

   } elseif { "$result" == "1" } {

      pktsrcInfo s2 "$timeOut msec timeout waiting for udp packet"
      set ret -1

   } else {

      set ret -1
   }

   # -- removing the callback
   Stack_remove_callback s2 udpPort $portHex IPv4_FWD_udp_test_callback

   return $ret
}













