#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# 2001 QAcafe http://qacafe.com
#
# $Id: lit-utils.tcl,v 1.7 2002/11/15 13:23:54 joe Exp $
#
# Helper procs for lit
#
# ----------------------------------------------------------------------------

# -- this is part of the pktsrc package
package provide pktsrc 1.3

#
# -- Initialize Lit. This is mainly used by tools that generate LSAs

proc Lit_init { ip mask } {

   global createcounters


   # -- counters for all the LSA types we are creating
   set createcounters(type1) 0
   set createcounters(type2) 0
   set createcounters(type3) 0
   set createcounters(type4) 0
   set createcounters(type5) 0

   # -- the free address range pool
   set createcounters(nextunusednet)     $ip
   set createcounters(nextunusednetmask) $mask

   # -- initialize the nets
   getnextnet

}


#
#
# -- build a LSA database converting all Tcl based LSAs into a packed
# form used by Lit and pktsrc

proc Lit_build_database { { database "" }} {
   global LSA lsaData

   if { "$database" == "" } {
      set database LSA
   }

   upvar $database data

   set count 0

   puts -nonewline "Lit: LSA Count [format "%6.6d" $count]"
   flush stdout

   # -- get a list of all LSAs to send
   set lsaList [ array names data "*,name" ]
   foreach key $lsaList {
      set lsa $data($key)
      set packed [ Lit_load_lsa $lsa data ]
      set lsaData(lsa,$lsa) $packed
      incr count

      if { [ expr $count % 5 ] == 0 } {
         puts -nonewline "\b\b\b\b\b\b[format "%6.6d" $count]"
         flush stdout
      }
   }

   set lsaData(count) $count
   puts -nonewline "\b\b\b\b\b\b[format "%6.6d" $count]"
   puts "\nLit: Loaded $lsaData(count) LSAs into Lit OSPF Database"
}


#
# -- find an LSA entry in the Tcl database
# build a single lsa array for this entry

proc Lit_get_lsa { name database lsa } {
   upvar $lsa l
   upvar $database data

   # -- pull out all the LSA values by this name

   set offset [ string length "$name," ]

   foreach n [array names data "$name,*" ] {
      set key [ string range $n $offset end ]
      set l($key) "$data($n)"
   }

   if { $l(type) == 1 } {
      Lit_convert_type_1 l
   } elseif { $l(type) == 2 } {
      Lit_convert_type_2 l
   } elseif { $l(type) == 3 } {
      Lit_convert_type_3 l
   } elseif { $l(type) == 4 } {
      Lit_convert_type_4 l
   } elseif { $l(type) == 5 } {
      Lit_convert_type_5 l
   } else {
      error "Error: unknown LSA type '$l(type)'"   
   }
}

# -- lit uses strings to store LSA information. This makes it easier to
# edit by hand. Here we want to convert to the same internal format that
# the rest of OSPF pktsrc is using.

proc Lit_convert_lsa_header { lsa } {
   upvar $lsa l

   set l(age)    [ format "%4.4x" $l(age) ]
   set l(type)   [ format "%2.2x" $l(type) ]
   set l(lsID)   [ iptohex $l(lsID) ]
   set l(router) [ iptohex $l(router) ]

   # -- just set the Ebit for now
   set l(options) 02
}

proc Lit_convert_type_1 { lsa } {
    upvar $lsa l

    # -- convert the header
    Lit_convert_lsa_header l

    # -- calc the length
    set l(length)   [ expr 24 + (12 * $l(links))]

    # -- build the options
    set l(routerOptions) 00

    if [ info exists l(bitE) ] {
       set l(routerOptions) \
          [ format "%2.2x" [ expr 0x$l(routerOptions) | 0x02 ]]
    }

    if [ info exists l(bitB) ] {
       set l(routerOptions) \
          [ format "%2.2x" [ expr 0x$l(routerOptions) | 0x01 ]]
    }

   
    # -- convert the links list
    for { set i 1 } { $i <= $l(links) } { incr i } {
      set l(link,$i,id)     [ iptohex $l(link,$i,id) ]   
      set l(link,$i,data)   [ iptohex $l(link,$i,data) ] 
      set l(link,$i,type)   [ format "%2.2x" $l(link,$i,type) ]  
      set l(link,$i,metric) [ format "%4.4x" $l(link,$i,type) ] 
   }

   # --okay
}

proc Lit_convert_type_2 { lsa } {
    upvar $lsa l

    # -- convert the header
    Lit_convert_lsa_header l

    set rlist ""
    set l(length) 24
    set l(netmask) [iptohex $l(netmask)]

    foreach r $l(routers) {
        incr l(length) 4
        lappend rlist [iptohex $r ]
    }

    # -- update with packed list
    set l(routers) $rlist
}

proc Lit_convert_type_3 { lsa } {
    upvar $lsa l

    # -- convert the header
    Lit_convert_lsa_header l
    set l(length) 28
    set l(netmask) [iptohex $l(netmask)]

}
proc Lit_convert_type_4 { lsa } {
    upvar $lsa l

    # -- convert the header
    Lit_convert_lsa_header l
    set l(length) 28
    set l(netmask) [iptohex $l(netmask)]
}

proc Lit_convert_type_5 { lsa } {
    upvar $lsa l

    # -- convert the header
    Lit_convert_lsa_header l

    set l(length) 36
    set l(netmask) [iptohex $l(netmask)]
    set l(forward) [iptohex $l(forward)]
    set l(tag)     [iptohex $l(tag)]

    set l(ebit)    00
    if [ info exists l(Ebit) ] {
       set l(ebit)    80
    }
}

proc Lit_load_lsa { name database } {
   upvar $database d

   set hexlsa ""

   # -- check that this LSA exists

   if [ info exists d($name,name) ] {
      # -- found it
      Lit_get_lsa $name d lsa

       if { $lsa(type) == "01" } {
          set hexlsa [ OSPF_build_router_lsa lsa ]
       } elseif { $lsa(type) == "02" } {
          set hexlsa [ OSPF_build_network_lsa lsa ]
       } elseif { $lsa(type) == "03" } {
          set hexlsa [ OSPF_build_summary_lsa lsa ]
       } elseif { $lsa(type) == "04" } {
          set hexlsa [ OSPF_build_summary_lsa lsa ]
       } elseif { $lsa(type) == "05" } {
          set hexlsa [ OSPF_build_external_lsa lsa ]
       } else {
          error "Unknown LSA type '$lsa(type)'"
       }
   } else {
      error "LSA '$name' that does in exist in database '$database'"
   }
 
   return $hexlsa
}

proc Lit_send_lsa { stack name } {
   global lsaData

   upvar $stack s

   # -- check that this LSA exists

   if [ info exists lsaData(lsa,$name) ] {
       puts "LIT: Flooding LSA '$name'"
       OSPF_send_lsa_update s $lsaData(lsa,$name)
   } else {
       error "LIT: Can not find LSA '$name'"
   }
}

proc Lit_send_all { stack { waitTime 10 } } {
   global LSA
   upvar $stack s

   # -- get a list of all LSAs to send
   set lsaList [ array names LSA "*,name" ]
   foreach key $lsaList {
      set lsa $LSA($key)
      Lit_send_lsa s $lsa
      Event_wait $waitTime
   }
}  

#
# -- print out a LSA by name

proc Lit_print_lsa { name } {
   global lsaData

   puts "Lit: LSA '$name'"
   if [ info exists lsaData(lsa,$name) ] {
       OSPF_print_lsa $lsaData(lsa,$name)
   } else {
      error "LIT: can't find LSA named '$name'"
   }       
}

proc Lit_next_sequence { name } {
   global lsaData

   if [ info exists lsaData(lsa,$name) ] {
       set lsaData(lsa,$name) [ OSPF_bump_seq $lsaData(lsa,$name) ]
   } else {
      error "LIT: can't find LSA named '$name'"
   }       
}

proc Lit_max_age { stack name } {
   global lsaData

   upvar $stack s

   if [ info exists lsaData(lsa,$name) ] {

       # -- send out an aged copy of the LSA
       set aged [ OSPF_max_age $lsaData(lsa,$name) ]
       OSPF_send_lsa_update s $aged
   } else {
      error "LIT: can't find LSA named '$name'"
   }       
}

proc Lit_max_age_all { stack } {
   global LSA
   upvar $stack s

   # -- get a list of all LSAs to age
   set lsaList [ array names LSA "*,name" ]
   foreach key $lsaList {
      set lsa $LSA($key)
      Lit_max_age s $lsa
      Event_wait 10
   }
}
  

