#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# 2001 QAcafe http://qacafe.com
#
# $Id: ospf-lsa.tcl,v 1.7 2002/11/15 13:23:54 joe Exp $
#
# LSA specific procs for OSPF
#
# ----------------------------------------------------------------------------

# -- this is part of the pktsrc package
package provide pktsrc 1.3


proc OSPF_parse_lsa { data } {

  # -- decode the lsa to get the length, note this could be optimized
  # to just pull out the length field rather than a full decode

  OSPF_decode_lsa_header $data lsa

  # -- compute the end of the lsa
  set end [ expr 0x$lsa(length) * 2 ]

  # -- return the lsa data
  set lsaData [ string range $data 0 $end ] 

}


proc OSPF_create_router_lsa { stack lsa } {
   upvar $stack s
   upvar $lsa out
  
   set out(age) 0001
   set out(options) 00
   set out(type) 01
   set out(lsID) $s(OSPF,routerID)
   set out(router) $s(OSPF,routerID)
   set out(seq) 80000001
   set out(checksum) 00000
   set out(length) 24
   set out(routerOptions) 00
   set out(links) 0
}

proc OSPF_router_lsa_add_link { lsa id data type metric } {
   upvar $lsa out
   
   # -- bump the link count
   incr out(links)

   # -- add the new link
   set out(link,$out(links),id) [iptohex $id ]
   set out(link,$out(links),data) [iptohex $data ]
   set out(link,$out(links),type) [format "%2.2x" $type ]
   set out(link,$out(links),metric) [ format "%4.4x" $metric ]

   # -- compute length in decimal for now
   incr out(length) 12
}



proc OSPF_build_lsa_header { lsa } {
   upvar $lsa out

   # -- build the header

   set header [ format "%s%s%s%s%s%s0000%s" \
     $out(age) $out(options) $out(type) $out(lsID) $out(router) $out(seq) \
     [ format "%4.4x" $out(length) ] ]


   return $header

}



proc OSPF_build_router_lsa { lsa } {
   upvar $lsa out

   # -- build the header
   set header [ OSPF_build_lsa_header out ]
   set r [format "%s%s00%4.4x"  $header $out(routerOptions) $out(links) ]

   # -- now loop through each link
   for { set i 1 } { $i <= $out(links) } {  incr i } {
      set link [ format "%s%s%s00%s" $out(link,$i,id) \
        $out(link,$i,data) $out(link,$i,type) $out(link,$i,metric) ]

       set r "$r$link"
   }

   # -- okay now for some checksum action
   set p [ string range $r 4 end ]

   # -- compute the fletcher checksum
   set checksum [ pktsrc fletcher $p 15 ]

   # -- now place the checksum back in in the packet
   set packet [ string replace $r 32 35 $checksum ]

   return $packet
}

proc OSPF_build_network_lsa { lsa } {
   upvar $lsa out

   # -- build the header

   set r [ OSPF_build_lsa_header out ]
   
   set rlist ""

   # -- build the router list
   foreach router $out(routers) {
      set rlist "$rlist$router"
   }

   set r "$r$out(netmask)$rlist"

   # -- okay now for some checksum action
   set p [ string range $r 4 end ]

   # -- compute the fletcher checksum
   set checksum [ pktsrc fletcher $p 15 ]

   # -- now place the checksum back in in the packet
   set packet [ string replace $r 32 35 $checksum ]

   return $packet
}

proc OSPF_build_summary_lsa { lsa } {
   upvar $lsa out

   # -- build the header
   set header [ OSPF_build_lsa_header out ]
   set r [format "%s%s00%6.6x"  $header $out(netmask) $out(metric) ]

   # -- okay now for some checksum action
   set p [ string range $r 4 end ]

   # -- compute the fletcher checksum
   set checksum [ pktsrc fletcher $p 15 ]

   # -- now place the checksum back in in the packet
   set packet [ string replace $r 32 35 $checksum ]

   return $packet
}

proc OSPF_build_external_lsa { lsa } {
   upvar $lsa out

   # -- build the header
   set header [ OSPF_build_lsa_header out ]
   set r [format "%s%s%s%6.6x%s%s"  $header $out(netmask) $out(ebit) \
     $out(metric) $out(forward) $out(tag) ]

   # -- okay now for some checksum action
   set p [ string range $r 4 end ]

   # -- compute the fletcher checksum
   set checksum [ pktsrc fletcher $p 15 ]

   # -- now place the checksum back in in the packet
   set packet [ string replace $r 32 35 $checksum ]

   return $packet
}

#
# -- dump the sequence of an LSA
proc OSPF_bump_seq { data { seq "" } } {

   if { "$seq" == "" } {
       set seq [ string range $data 24 31 ]
   }

   # +++joe check that incrhex handles wrapping correctly
   set newseq [ incrhex $seq ]
   

   set newlsa [ string replace $data 24 31 $newseq ]


   set packet [ string replace $newlsa 32 35 0000 ]
   set p [ string range $packet 4 end ]

   # -- compute the fletcher checksum
   set checksum [ pktsrc fletcher $p 15 ]

   # -- now place the checksum back in in the packet
   set newlsa [ string replace $packet 32 35 $checksum ]

   # -- need new check
   return $newlsa

}

#
# -- max age an LSA
proc OSPF_max_age { data } {

   set age [ format "%4.4x" 3600 ]

   # no new checksum required
   set newlsa [ string replace $data 0 3 $age ]
   return $newlsa

}

#
# -- convert from dotted quad string to integer
proc str_to_ipv4 { ip } {

  if { [ scan $ip "%d.%d.%d.%d" one two three four ] != 4 } {
    puts "Can not convert IP address $ip into 32 bit value"
    exit
  }

  set ipNum [ expr $one * 256 * 256 * 256 + $two * 256 * 256 + \
      $three * 256 + $four ]

  return $ipNum
}

#
# -- convert from integer to string. Tcl treats the integer as
# signed, so special case negative numbers

proc ipv4_to_str { ip } {

  # special case the high byte
  if { $ip < 0 } {
     set four [ expr $ip - 0x80000000 ]
     set highbit 1
  } else {
     set highbit 0
  }

  set four [ expr ($ip & 0xFF000000)>>24 ]
  set three [ expr ($ip & 0x00FF0000)>>16 ]
  set two [ expr ($ip & 0x0000FF00)>>8 ]
  set one [ expr $ip & 0x000000FF ]

  if { $highbit == 1 } {
     # puts "four before = [format "%x" $four]"
     set four [ expr $four ^ 0xFFFFFF80 ]
     set four [ expr $four + 0x80 ]
  }

  return [ format "%d.%d.%d.%d" $four $three $two $one ]
}

proc print_type_5 { lsa } {
  upvar $lsa l

  puts "\n#"
  puts "# $l(name) - Type 5 LSA" 
  OSPF_lsa_to_tcl l
  return 

}

proc print_type_4 { lsa } {
  upvar $lsa l

  puts "\n#"
  puts "# $l(name) - Type 4 LSA" 
  OSPF_lsa_to_tcl l

}

proc print_type_3 { lsa } {
  upvar $lsa l

  puts "\n#"
  puts "# $l(name) - Type 3 LSA" 
  OSPF_lsa_to_tcl l

}

proc print_type_2 { lsa } {
  upvar $lsa l

  puts "\n#"
  puts "# $l(name) - Type 2 LSA" 
  OSPF_lsa_to_tcl l

}

proc print_type_1 { lsa } {
  upvar $lsa l

  puts "\n#"
  puts "# $l(name) - Type 1 LSA" 
  OSPF_lsa_to_tcl l

}

proc OSPF_lsa_to_tcl { lsa } {
    upvar $lsa l
    set pattern *

    set maxl 0
    foreach name [array names l $pattern] {
	if {[string length $name] > $maxl} {
	    set maxl [string length $name]
	}
    }

    set maxl [expr {$maxl + [string length $lsa] + 15}]
    foreach name [lsort [array names l $pattern]] {
	set nameString [format LSA(%s,%s) $l(name) $name]
	puts stdout [format "set %-*s \"%s\"" $maxl $nameString $l($name)]
    }   
}


proc create_type_5 { l route mask ar { metric 10 } } {
   global createcounters 
   upvar $l lsa

   incr createcounters(type5)
   set lsa(name) "e$createcounters(type5)" 
   set lsa(type) 5
   set lsa(age) 0 
   set lsa(lsID) $route 
   set lsa(router) $ar 
   set lsa(seq) 80000002 
   set lsa(Ebit) "Ebit"
   set lsa(netmask) $mask 
   set lsa(metric) $metric 
   set lsa(forward) "0.0.0.0"
   set lsa(tag) 4.3.2.1 
}

proc create_type_4 { l route ar { metric 10 } } {
   global createcounters 
   upvar $l lsa

   incr createcounters(type4)
   set lsa(name) "asbr$createcounters(type4)" 
   set lsa(type) 4 
   set lsa(age) 0
   set lsa(lsID) $route 
   set lsa(router) $ar 
   set lsa(seq) 80000002 
   set lsa(Ebit) "Ebit"
   set lsa(netmask) 0.0.0.0
   set lsa(metric) $metric 
}

proc create_type_3 { l route mask ar { metric 10 } } {
   global createcounters 
   upvar $l lsa

   incr createcounters(type3)
   set lsa(name) "sum$createcounters(type3)" 
   set lsa(type) 3
   set lsa(age) 0 
   set lsa(lsID) $route 
   set lsa(router) $ar 
   set lsa(seq) 80000002 
   set lsa(Ebit) "Ebit"
   set lsa(netmask) $mask 
   set lsa(metric) $metric 
}

proc create_type_2 { l route mask ar rtr_list } {
   global createcounters 
   upvar $l lsa

   incr createcounters(type2)
   set lsa(name) "net$createcounters(type2)" 
   set lsa(type) 2 
   set lsa(age)      0
   set lsa(lsID) $route 
   set lsa(router) $ar 
   set lsa(seq) 80000002 
   set lsa(Ebit) "Ebit"
   set lsa(netmask) $mask 
   set lsa(routers) $rtr_list
}

proc create_type_1 { l id ar bitB bitE if_list } {
   global createcounters 
   upvar $l lsa
   upvar $if_list list 

   incr createcounters(type1)
   set lsa(name)     "rtr$createcounters(type1)" 
   set lsa(type)     1 
   set lsa(age)      0
   set lsa(lsID)     $id 
   set lsa(router)   $ar 
   set lsa(seq)      80000002 

   # header lsa bits
   set lsa(Ebit)     1

   # router LSA bits
   set lsa(bitE)     $bitE
   set lsa(bitB)     $bitB
   set lsa(links)    $list(count)

   for { set i 1 } { $i <= $list(count) } { incr i } {
      set lsa(link,$i,id)     $list($i,id)
      set lsa(link,$i,data)   $list($i,data)
      set lsa(link,$i,type)   $list($i,type)
      set lsa(link,$i,metric) $list($i,metric)
   }
}


proc getnextnet { } {
  global createcounters

  set returnnet $createcounters(nextunusednet)

  # compute net network
  set net [ str_to_ipv4 $createcounters(nextunusednet) ]
  set mask [ str_to_ipv4 $createcounters(nextunusednetmask) ]
  set newnet [ expr $net + (~$mask + 1) ]

  # -- put in string format
  set newnetstr [ ipv4_to_str $newnet ]

  # -- remember the next free net
  set createcounters(nextunusednet) $newnetstr
  set createcounters(nextunusednet_num) $newnet

  # set up the next unused address
  set createcounters(nextunusednetaddress) $newnetstr

  return $newnetstr
}

proc getnextnetaddress { } {
  global createcounters

  # -- get the address
  set address [ str_to_ipv4 $createcounters(nextunusednetaddress) ]

  # -- bump
  incr address

  # -- remeber 
  set createcounters(nextunusednetaddress) [ ipv4_to_str $address ]
  set createcounters(nextunusednetaddress_num) $address
   
  # -- return the address
  return $createcounters(nextunusednetaddress)
}


