#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# 2001 QAcafe http://qacafe.com
#
# $Id: ospf.tcl,v 1.12 2002/11/15 13:23:54 joe Exp $
#
# Poor man's OSPF
#
# ----------------------------------------------------------------------------

# -- this is part of the pktsrc package
package provide pktsrc 1.3

proc OSPF_input { packet stack { options none }} {
   upvar $stack s

   if { $options == "print" } {

      OSPF_decode $packet ospf
      OSPF_print ospf
      return

   } else {

      # -- remember this packet in the 'ospf' variable. This allows other
      # callbacks to get access to the decodes at any Tcl stack level

      eval [ Stack_save_packet s OSPF ospf ]
   
      # -- now decode the packet into the 'ospf' variable
      OSPF_decode $packet ospf

   }

   if [ info exists s(OSPF,running) ] {
       if { $s(OSPF,running) == 1 } {
         # -- okay OSPF is running
       } else {
         return
       }
   } else {
     # -- just ignore this packet since OSPF is not running
     return
   }


   # -- check the checksum before any callbacks
 
   # -- authenticate packet
   if ![ OSPF_authenticate_packet s ospf ] {
      return
   }

   # -- check for callbacks at the ICMP level
   if [ info exists s(ospfType,$ospf(type)) ] {

      # -- drive any input callbacks
      foreach c $s(ospfType,$ospf(type),callbacks) {
          # -- send the whole packet
          $c $packet s $options
      }
   }
}

#
#
# OSPF_init - set up any deaults for OSPF
#

proc OSPF_init { stack } {
   upvar $stack s

   # -- mark this as running
   if [ info exists s(OSPF,running) ] {
       # -- init was called again, restart

       # -- delete all neighbors
       OSPF_delete_neighbors s

       # -- cancel any existing hello timers
       if [ info exists s(OSPF,helloId) ] {
          after cancel $s(OSPF,helloId)
          unset s(OSPF,helloId)
       }
   }
   set s(OSPF,running) 1

   pktsrcInfo s "Setting up OSPF stack on [hextoip $s(ip)]"

   # Add MAC entries for all DR, and all Routers
   ARP_add_entry s [iptohex 224.0.0.5] 01005e000005
   ARP_add_entry s [iptohex 224.0.0.6] 01005e000006


   set s(OSPF,options) 0x02
   set s(OSPF,routerID) $s(ip)
 
   # -- default to the backbone area
   set s(OSPF,areaID) 00000000

   # -- set the dead interval to 40 seconds
   set s(OSPF,deadInt)  40
   set s(OSPF,helloInt) 10
   set s(OSPF,rxInt)     5

   # -- default to null auth
   set s(OSPF,auth) 0000
   set s(OSPF,authData) 0000000000000000

   # -- set my router priority so that I'll never be the DR
   set s(OSPF,priority) 0

   set s(OSPF,options) 02

   # -- default dr/bdr
   set s(OSPF,dr) 00000000
   set s(OSPF,bdr) 00000000


   # -- start with an empty neighbor list
   set s(OSPF,neighborList) ""

   # -- any counters
   set s(OSPF,badAuth) 0
   
}

#
# OSPF_set_auth
#
# support types
#
#   null
#   simple 
#   md5
#


proc OSPF_set_auth { stack type { key "" } { keyId "" } } {
   upvar $stack s

   if { "$type" == "null" } {

       # -- default to null auth
       set s(OSPF,auth) 0000
       set s(OSPF,authData) 0000000000000000

       pktsrcInfo s "Set authentication type to null"

   } elseif { "$type" == "simple" } {
        
        set length [ string length $key ] 
        if { $length > 8 } {
          error "OSPF_set_auth: auth data too big for simple auth '$key'"
        }

       set diff [ expr 8 - $length ]
       if { $diff > 0 } {
          set pad [ string repeat "00" $diff ]
       }

       set s(OSPF,auth) 0001
       set s(OSPF,authData) "[ pktsrc text2ascii $key ]$pad"

       pktsrcInfo s "Set authentication type to simple, password = '$key'"



   } elseif { "$type" == "md5" } {

       set s(OSPF,auth) 0002
       set s(OSPF,authKey) [ format "%2.2x" $keyId ]
       set s(OSPF,authSeq) 00000001
       set s(OSPF,authData) \
          [ format "0000%s10%s" $s(OSPF,authKey) $s(OSPF,authSeq) ]
       set s(OSPF,authSecret) $key
       pktsrcInfo s "Set authentication type to md5, message-digest-id = '$s(OSPF,authKey)', 'key = '$key'"

   } else {
       error "OSPF_set_auth: unknown auth type $type"
   }    

}

proc OSPF_decode { packet ospf } {
   upvar $ospf out

   # -- parse out all the fields we find interesting ...

   set out(version)     [ string range $packet 0 1 ]
   set out(type)        [ string range $packet 2 3 ]
   set out(length)      [ string range $packet 4 7 ]
   set out(routerID)    [ string range $packet 8 15 ]
   set out(areaID)      [ string range $packet 16 23 ]
   set out(checksum)    [ string range $packet 24 27 ]
   set out(auType)      [ string range $packet 28 31 ]
   set out(auth)        [ string range $packet 32 47 ]

   if { $out(auType) == "0002" } {
      set out(authKey)    [ string range $out(auth) 4 5 ]
      set out(authKeyLen) [ string range $out(auth) 6 7 ]
      set out(authSeq)    [ string range $out(auth) 8 15 ]

      # -- last 16 bytes should be md5
      set endmd5   [ expr [ string length $packet ] - 1 ]
      set startmd5 [ expr $endmd5 - 31 ]
      set out(md5)    [ string range $packet $startmd5 $endmd5 ]
   }

   # -- nothing to transport
   set out(data)       ""

   if { $out(type) == "01" } {
      OSPF_decode_hello $packet out
   } elseif { $out(type) == "02" } {
      OSPF_decode_dd $packet out
   } elseif { $out(type) == "03" } {
      OSPF_decode_lsa_request $packet out
   } elseif { $out(type) == "04" } {
      OSPF_decode_lsa_update $packet out
   } elseif { $out(type) == "05" } {
      OSPF_decode_lsack $packet out
   }   

}

proc OSPF_decode_hello { packet ospf } {
   upvar $ospf out

   # -- parse out all the fields we find interesting ...

   set out(netmask)     [ string range $packet 48 55 ]
   set out(helloInt)    [ string range $packet 56 59 ]
   set out(options)     [ string range $packet 60 61 ]
   set out(rtrPri)      [ string range $packet 62 63 ]
   set out(deadInt)     [ string range $packet 64 71 ]
   set out(dr)          [ string range $packet 72 79 ]
   set out(bdr)         [ string range $packet 80 87 ]

   set nbrLen [ expr 0x$out(length) - 44 ]
   set p 88
   set out(nbr) ""

   while { $nbrLen >= 4 } {  
      lappend out(nbr) [ string range $packet $p [ expr $p + 7 ]]
      incr p 8
      incr nbrLen -4
   }
}

proc OSPF_decode_dd { packet ospf } {
   upvar $ospf out

   # -- parse out all the fields we find interesting ...

   set out(mtu)        [ string range $packet 48 51 ]
   set out(options)    [ string range $packet 52 53 ]
   set out(bits)       [ string range $packet 54 55 ]

   if { [ expr 0x$out(bits) & 0x01 ] } {
       set out(MS_bit) 1
   } else {
       set out(MS_bit) 0
   }

   if { [ expr 0x$out(bits) & 0x02 ] } {
       set out(M_bit) 1
   } else {
       set out(M_bit) 0
   }

   if { [ expr 0x$out(bits) & 0x04 ] } {
       set out(I_bit) 1
   } else {
       set out(I_bit) 0
   }
      

   set out(ddSeq)      [ string range $packet 56 63 ]

   set lsaLen [ expr 0x$out(length) - 32 ]
   set p 64
   set out(lsa) ""

   while { $lsaLen >= 20 } {  
      lappend out(lsa) [ string range $packet $p [ expr $p + 39 ]]
      incr p 40
      incr lsaLen -20
   }
}



proc OSPF_decode_lsa_update { packet ospf } {
   upvar $ospf out

   # -- parse out all the fields we find interesting ...

   set out(lsaCount)        [ string range $packet 48 55 ]

   set lsaLen [ expr 0x$out(length) - 28 ]
   set p 56

   set out(lsa) ""
   set count 0

   # -- parse through all LSAs 

   while { $lsaLen >= 20 } {  
      incr count
      set data [ string range $packet $p end ]
      set out(lsa,$count) [ OSPF_parse_lsa $data ]
      set length [expr [ string length $out(lsa,$count)]/2 ]
      incr p [ expr $length * 2 ]
      incr lsaLen -$length

   }

   # +++joe add consistency check -> count = lsaCount
}

proc OSPF_decode_lsa_request { packet ospf } {
   upvar $ospf out

   # -- parse out all the fields we find interesting ...

   set lsaLen [ expr 0x$out(length) - 24 ]
   set p 48

   set out(lsa) ""
   set count 0

   # -- parse through all LSAs 

   while { $lsaLen >= 12 } {  
      incr count
      
      set data [ string range $packet $p end ]
 
      set out(lsa,$count,type)   [string range $data 0 7 ]
      set out(lsa,$count,lsID)   [string range $data 8 15 ]
      set out(lsa,$count,router) [string range $data 16 23 ]

      incr p 24
      incr lsaLen -12
   }

   set out(count) $count
}

proc OSPF_decode_lsack { packet ospf } {
   upvar $ospf out

   set lsaLen [ expr 0x$out(length) - 24 ]
   set p 48
   set out(lsa) ""

   while { $lsaLen >= 20 } {  
      lappend out(lsa) [ string range $packet $p [ expr $p + 39 ]]
      incr p 40
      incr lsaLen -20
   }
}


proc OSPF_print { ospf } {
   upvar $ospf out

   puts "OSPF: ---- OSPF Packet ----"
   puts "OSPF: "
   puts "OSPF: Version                   = [expr 0x$out(version)]"
   puts "OSPF: Type                      = [OSPF_type $out(type)] (0x$out(type))"
   puts "OSPF: Length                    = [expr 0x$out(length)]"
   puts "OSPF: Router ID                 = [hextoip $out(routerID)]"
   puts "OSPF: Area ID                   = [hextoip $out(areaID)]"
   puts "OSPF: Checksum                  = 0x$out(checksum)"
   puts "OSPF: Auth Type                 = [OSPF_auth_type $out(auType)] (0x$out(auType))"
   puts "OSPF: Auth                      = '[pktsrc ascii2text $out(auth)]' (0x$out(auth))"

   if { $out(auType) == "0002" } {
      puts "OSPF:   Key ID:                 = 0x$out(authKey)"
      puts "OSPF:   Auth Len:               = [ expr 0x$out(authKeyLen)]"
      puts "OSPF:   Cyrpto Seq:             = 0x$out(authSeq)"
   }

   if { $out(type) == "01" } {
      OSPF_print_hello out
   } elseif { $out(type) == "02" } {
      OSPF_print_dd out
   } elseif { $out(type) == "03" } {
      OSPF_print_lsa_request out
   } elseif { $out(type) == "04" } {
      OSPF_print_lsa_update out
   } elseif { $out(type) == "05" } {
      OSPF_print_lsack out
   } else {
      puts "OSPF: Unknown type '0x$out(type)'"
   }

   puts "OSPF:"
   if { $out(auType) == "0002" } {

       puts "OSPF: Md5                       = 0x$out(md5)"
   }
   puts "OSPF: ---- End OSPF Packet ----"
}

proc OSPF_print_hello { ospf } {
   upvar $ospf out

   puts "OSPF: ----  Hello Information ----"
   puts "OSPF: Netmask                   = [hextoip $out(netmask)]"
   puts "OSPF: Hello Interval            = [expr 0x$out(helloInt)]"
   puts -nonewline "OSPF: Options                   = 0x$out(options)"
   if { [ expr 0x$out(options) & 0x02 ] } {
      puts " (Ebit set)"                
   } else {
      puts " (Ebit not set)"                
   }
   puts "OSPF: Router Priority           = [expr 0x$out(rtrPri)]"
   puts "OSPF: Dead Interval             = [expr 0x$out(deadInt)]"
   puts "OSPF: Designated Router         = [hextoip $out(dr)]"
   puts "OSPF: Backup Designated Router  = [hextoip $out(bdr)]"

   foreach n $out(nbr) {
      puts "OSPF:   Neighbor                = [hextoip $n]"
   }
}

proc OSPF_print_dd { ospf } {
   upvar $ospf out

   puts "OSPF: ----  Database Description Information ----"
   puts "OSPF: Interface MTU             = [expr 0x$out(mtu)]"
   puts -nonewline "OSPF: Options                   = 0x$out(options)"
   if { [ expr 0x$out(options) & 0x02 ] } {
      puts " (Ebit set)"                
   } else {
      puts " (Ebit not set)"                
   }
   puts -nonewline "OSPF: DD Bits                   = 0x$out(bits)"
   puts " I:$out(I_bit) M:$out(M_bit) MS:$out(MS_bit)"

   puts "OSPF: DD Sequence               = 0x$out(ddSeq)"

   foreach lsa $out(lsa) {
      OSPF_print_lsa_header $lsa
   }
}

proc OSPF_print_lsa_request { ospf } {
   upvar $ospf out

   puts "OSPF: ----  LSA Request Information ----"

   for { set i 1 } { $i <= $out(count) } { incr i } { 

        puts "OSPF: LS Type                    = [OSPF_lsa_type $out(lsa,$i,type)]"
        puts "OSPF: Link State ID              = [ hextoip $out(lsa,$i,lsID)]"
        puts "OSPF: Advertising Router         = [ hextoip $out(lsa,$i,router)]"

   }
}

proc OSPF_print_lsa_update { ospf } {
   upvar $ospf out

   puts "OSPF: ----  LSA Information ----"
   puts "OSPF: LSA Count                 = [expr 0x$out(lsaCount)]"

   # -- get the count in decimel
   set lsaCount [ expr 0x$out(lsaCount) ]

   for { set i 1 } { $i <= $lsaCount } { incr i } { 
        # -- just print the headers for now
        if [ info exists out(lsa,$i) ] {
            # OSPF_print_lsa_header $out(lsa,$i)
            OSPF_print_lsa $out(lsa,$i)
        } else {
            break
        }
   }
}

proc OSPF_print_lsack { ospf } {
   upvar $ospf out

   puts "OSPF: ----  LS Ack Information ----"

   # -- display all the LSAs
   foreach lsa $out(lsa) {
      OSPF_print_lsa_header $lsa
   }
}


proc OSPF_decode_lsa_header { lsa output } {
   upvar $output out

   set out(age)       [ string range $lsa 0 3 ]
   set out(options)   [ string range $lsa 4 5 ]
   set out(type)      [ string range $lsa 6 7 ]
   set out(lsID)      [ string range $lsa 8 15 ]
   set out(router)    [ string range $lsa 16 23 ]
   set out(seq)       [ string range $lsa 24 31 ]
   set out(checksum)  [ string range $lsa 32 35 ]
   set out(length)    [ string range $lsa 36 39 ]
}

proc OSPF_decode_type_1 { lsa output } {
   upvar $output out

   set out(routerOptions)       [ string range $lsa 0 1 ]
   set out(links)               [ string range $lsa 4 7 ]
   
   set lsaLen [ expr [ string length $lsa ] / 2 ]
   set p 8

   set i 1
   while { $lsaLen >= 12 } {  
      lappend out(link,$i,lsID) [ string range $lsa $p [ expr $p + 7 ]]
      incr p 8
      lappend out(link,$i,data) [ string range $lsa $p [ expr $p + 7 ]]
      incr p 8
      lappend out(link,$i,type) [ string range $lsa $p [ expr $p + 1 ]]
      incr p 4
      lappend out(link,$i,metric) [ string range $lsa $p [ expr $p + 3 ]]
      incr p 4
      incr lsaLen -12
      incr i
   }
   # -- check that i equals links for consistency ?
}

proc OSPF_decode_type_2 { lsa output } {
   upvar $output out

   set out(netmask)               [ string range $lsa 0 7 ]
   
   set lsaLen [ expr ([ string length $lsa ] / 2) - 4 ]
   set p 8

   while { $lsaLen >= 4 } {  
      lappend out(routers) [ string range $lsa $p [ expr $p + 7 ]]
      incr p 8
      incr lsaLen -4
   }
   # -- check that i equals links for consistency ?
}

proc OSPF_decode_type_3 { lsa output } {
   upvar $output out

   set out(netmask)               [ string range $lsa 0 7 ]
   set out(metric)                [ string range $lsa 10 15 ]
   
   # -- check that i equals links for consistency ?
}

proc OSPF_decode_type_5 { lsa output } {
   upvar $output out

   set out(netmask)               [ string range $lsa 0 7 ]
   set out(ebit)                  [ string range $lsa 8 9 ]
   set out(metric)                [ string range $lsa 10 15 ]
   set out(forward)               [ string range $lsa 16 23 ]
   set out(tag)                   [ string range $lsa 24 31 ]
}

proc OSPF_decode_type_7 { lsa output } {
   upvar $output out

   set out(netmask)               [ string range $lsa 0 7 ]
   set out(ebit)                  [ string range $lsa 8 9 ]
   set out(metric)                [ string range $lsa 10 15 ]
   set out(forward)               [ string range $lsa 16 23 ]
   set out(tag)                   [ string range $lsa 24 31 ]
}

proc OSPF_print_type_1 { lsa } {
   OSPF_decode_type_1 $lsa out
  
   set links [ expr 0x$out(links) ]

   puts "OSPF: ---- Router LSA (Type 1) ----"
   puts "OSPF: Links                     = $links"

   for { set i 1 } { $i <= $links } { incr i } {
      puts "OSPF: -- Link $i ----"   
      puts "OSPF:   ID                      = [hextoip $out(link,$i,lsID)]"
      puts "OSPF:   Data                    = [hextoip $out(link,$i,data)]"
      puts "OSPF:   Type                    = [ OSPF_router_link_type $out(link,$i,type)] (0x$out(link,$i,type))"
      puts "OSPF:   Metric                  = [expr 0x$out(link,$i,metric)] (0x$out(link,$i,metric))"
   }
}

proc OSPF_print_type_2 { lsa } {

   OSPF_decode_type_2 $lsa out
  
   puts "OSPF: ---- Network LSA (Type 2) ----"
   puts "OSPF: Netmask                   = [hextoip $out(netmask)]"

   foreach r $out(routers) {
      puts "OSPF:   Attached Router         = [hextoip $r]"
   }
}

proc OSPF_print_type_3 { lsa } {

   OSPF_decode_type_3 $lsa out
  
   puts "OSPF: ---- Summary LSA (Type 3) ----"
   puts "OSPF: Netmask                   = [hextoip $out(netmask)]"
   puts "OSPF: Metric                    = [expr 0x$out(metric)] (0x$out(metric))"

}
proc OSPF_print_type_4 { lsa } {

   # -- type 3 and type 4 decodes are the same!!
   OSPF_decode_type_3 $lsa out
  
   puts "OSPF: ---- Summary LSA (Type 4) ----"
   puts "OSPF: Netmask                   = [hextoip $out(netmask)]"
   puts "OSPF: Metric                    = [expr 0x$out(metric)] (0x$out(metric))"

}

proc OSPF_print_type_5 { lsa } {

   OSPF_decode_type_5 $lsa out

   puts "OSPF: ---- AS-external LSA (Type 5) ----"
   puts "OSPF: Netmask                   = [hextoip $out(netmask)]"
   puts "OSPF: Metric                    = [expr 0x$out(metric)] (0x$out(metric))"

   if { $out(ebit) == "80" } {
     puts "OSPF:   Type 2 (0x$out(ebit))"
   } else {
     puts "OSPF:   Type 1 (0x$out(ebit))"
   }
   puts "OSPF: Forwarding Address        = [hextoip $out(forward)]"
   puts "OSPF: Route Tag                 = [hextoip $out(tag)]"

}

proc OSPF_print_type_7 { lsa } {

   OSPF_decode_type_7 $lsa out

   puts "OSPF: ---- NSSA-External LSA (Type 7) ----"
   puts "OSPF: Netmask                   = [hextoip $out(netmask)]"
   puts "OSPF: Metric                    = [expr 0x$out(metric)] (0x$out(metric)) "

   if { [ expr 0x$out(ebit) & 0x08 ] } {
     puts "OSPF:   N/P bit is set (0x$out(ebit))"
   } else {
     puts "OSPF:   N/P bit is not set (0x$out(ebit))"
   }

   puts "OSPF: Forwarding Address        = [hextoip $out(forward)]"
   puts "OSPF: Route Tag                 = [hextoip $out(tag)]"

}


proc OSPF_print_lsa { lsa } {

   OSPF_decode_lsa_header $lsa out
   OSPF_print_lsa_header $lsa

   set lsaBody [ string range $lsa 40 end ]  

   if { $out(type) == "01" } {
       OSPF_print_type_1 $lsaBody
   } elseif { $out(type) == "02" } {
       OSPF_print_type_2 $lsaBody
   } elseif { $out(type) == "03" } {
       OSPF_print_type_3 $lsaBody
   } elseif { $out(type) == "04" } {
       OSPF_print_type_4 $lsaBody
   } elseif { $out(type) == "05" } {
       OSPF_print_type_5 $lsaBody
   } elseif { $out(type) == "07" } {
       OSPF_print_type_7 $lsaBody
   } else {
       puts "OSPF: Unkown LSA type '0x$out(type)'"
   }
}

proc OSPF_print_lsa_header { lsa } {

   OSPF_decode_lsa_header $lsa out

   puts "OSPF: ----  LSA Header ----"
   puts "OSPF: LS Age                    = [expr 0x$out(age) ] (0x$out(age))"
   puts "OSPF: Options                   = 0x$out(options)"
   puts "OSPF: LS Type                   = [OSPF_lsa_type $out(type) ] (0x$out(type))"
   puts "OSPF: LS ID                     = [hextoip $out(lsID)]"
   puts "OSPF: Advertising Router        = [hextoip $out(router)]"
   puts "OSPF: LS Sequence Number        = 0x$out(seq)"
   puts "OSPF: LS Checksum               = 0x$out(checksum)"
   puts "OSPF: Length                    = [ expr 0x$out(length) ]"

}

proc OSPF_type { type } {

   if { $type == "01" } {
      set str "Hello"
   } elseif { $type == "02" } {
      set str "Database Description"
   } elseif { $type == "03" } {
      set str "LS Request"
   } elseif { $type == "04" } {
      set str "LS Update"
   } elseif { $type == "05" } {
      set str "LS Ack"
   } else {
      set str "Unknown OSPF Type (0x$type)"
   }

   return $str

}

proc OSPF_auth_type { type } {

   if { $type == "0000" } {
      set str "Null"
   } elseif { $type == "0001" } {
      set str "Simple Password"
   } elseif { $type == "0002" } {
      set str "MD5"
   } else {
      set str "Unknown OSPF AuthType (0x$type)"
   }

   return $str

}

proc OSPF_lsa_type { type } {

   if { $type == "01" } {
      set str "Router-LSA"
   } elseif { $type == "02" } {
      set str "Network-LSA"
   } elseif { $type == "03" } {
      set str "Type 3 Summary LSA"
   } elseif { $type == "04" } {
      set str "Type 4 Summary LSA"
   } elseif { $type == "05" } {
      set str "AS-External LSA"
   } elseif { $type == "07" } {
      set str "NSSA-External LSA"
   } else {
      set str "Unknown OSPF LSA Type (0x$type)"
   }

   return $str
}

proc OSPF_router_link_type { type } {

   if { $type == "01" } {
      set str "Point-to-Point"
   } elseif { $type == "02" } {
      set str "Transit Network"
   } elseif { $type == "03" } {
      set str "Stub Network"
   } elseif { $type == "04" } {
      set str "Virtual Link"
   } else {
      set str "Unknown Router Link Type"
   }

   return $str

}



#
# -- initialize the hello protocol for this stack

proc OSPF_init_hello { stack } {
   upvar $stack mystack

   OSPF_init mystack

   # -- add a callback for hello
    if { ![info exists mystack(OSPF,handlers) ] } {
      Stack_add_callback mystack ospfType 01 OSPF_hello_input
      Stack_add_callback mystack ospfType 02 OSPF_dd_input
      Stack_add_callback mystack ospfType 03 OSPF_request_input
      Stack_add_callback mystack ospfType 04 OSPF_update_input
      Stack_add_callback mystack ospfType 05 OSPF_ack_input
      set mystack(OSPF,handlers) 1       
   }

   # -- set up a hello interval timer
   OSPF_hello_timer mystack

}

proc OSPF_authenticate_packet { stack ospf } {
   upvar $stack s
   upvar $ospf o

   # -- make sure the authentication types match
   if { "$o(auType)" != "$s(OSPF,auth)" } {
       incr s(OSPF,badAuth)
       return 0
   }

   # -- null authentication
   if { "$o(auType)" == "0000" } {
       return 1

   # -- simple authentication
   } elseif { "$o(auType)" == "0001" } {
       if { "$o(auth)" == "$s(OSPF,authData)" } {
          incr s(OSPF,badAuth)
          return 1
       } else {
          return 0
       }

   # -- md5 auth
   } elseif { "$o(auType)" == "0002" } {

       # -- need to check crypto sequence number

       # -- compute the md5 for the last ospf packet on this stack
       set md5 [ OSPF_md5 s ]

       # -- check for a match
       if { "$md5" == "$o(md5)" } {
	   return 1 
       } else {
          pktsrcInfo s "MD5 authentication failed, computed '$md5'"
          incr s(OSPF,badAuth)
          return 0
       }

   # -- unknown auth type
   } else {
       pktsrcInfo s "Received unknown OSPF authentication type '$o(auType)'"
       return 0
   }
}

proc OSPF_hello_timer { stack } {
   upvar $stack s

   # -- check my neighbor states
   set current [ clock seconds ]
   foreach n $s(OSPF,neighborList) {
     set interval [ expr $current - $s(OSPF,neighbor,$n,last) ]
     if { $interval >= $s(OSPF,deadInt) } {
         pktsrcInfo s "Neighbor [hextoip $n] is dead, dropping"
         OSPF_delete_neighbor s $n
     }
   }

   pktsrcInfo s "Sending OSPF Hello pri:$s(OSPF,priority) int:$s(OSPF,helloInt) dead:$s(OSPF,deadInt)"
   OSPF_send_hello s

   # -- next timer
   set s(OSPF,helloId) [ after [ expr $s(OSPF,helloInt) * 1000 ]  "
       $s(getStack)
       OSPF_hello_timer $s(getStackName) 
   "]
}

proc OSPF_hello_input { packet stack { options none }} {
   upvar $stack s

   # -- I should hide this
   set ipState $s(current,IPv4)
   upvar #$s(current,IPv4,level) $ipState ip

   # -- I should hide this
   set ospfState $s(current,OSPF)
   upvar #$s(current,OSPF,level) $ospfState ospf

   set port $s(port)

   set seconds [ clock seconds ]

   # -- make a few checks
  
   # -- check netmask 
   # -- check options bits
   # -- intervals, etc

   # -- get this neighbors router ID
      
   pktsrcInfo s "Received Hello Packet from [hextoip $ospf(routerID)]" 

   if [ info exists s(OSPF,neighbor,$ospf(routerID)) ] {
     # -- update the dead interval counter
     set s(OSPF,neighbor,$ospf(routerID),last) $seconds

     # -- check for 2 way
     if {  $s(OSPF,neighbor,$ospf(routerID),state) == "Init" } {
         foreach n $ospf(nbr) {
	    if { $n == $s(OSPF,routerID) } {
               set s(OSPF,neighbor,$ospf(routerID),state) "2-Way"
            }
         }
     }

     # -- existing neighbor

     # -- check for neighbor state change
     set found 0
     foreach n $ospf(nbr) {
	 if { $n == $s(OSPF,routerID) } {
             set found 1
         }
     }

     if { $found == 0 } {
        pktsrcWarn s "Lost Adjacency with OSPF neighbor [hextoip $ospf(routerID)]"
     } 

   } else {
      pktsrcInfo s "Adding new OSPF neighbor [hextoip $ospf(routerID)]"     
      set s(OSPF,neighbor,$ospf(routerID)) $ospf(rtrPri)
      set s(OSPF,neighbor,$ospf(routerID),last) $seconds
      set s(OSPF,neighbor,$ospf(routerID),srcIP) $ip(src)

      # -- move this to new proc
      set state Init      
      foreach n $ospf(nbr) {
	  if { $n == $s(OSPF,routerID) } {
             set state 2-Way
          }
      }

      set s(OSPF,neighbor,$ospf(routerID),state) $state
      pktsrcInfo s "Neighbor [hextoip $ospf(routerID)] state: $s(OSPF,neighbor,$ospf(routerID),state)"
  
      lappend s(OSPF,neighborList) $ospf(routerID) 
   }


   # -- check if dr/bdr is elected

   # -- This is a really crude DR election
   if { $ospf(dr) != 00000000 && $s(OSPF,dr) == 00000000 } {
         set s(OSPF,dr) $ospf(dr)
         pktsrcInfo s "Selecting [hextoip $ospf(dr)] as the DR" 
 
   }
   if { $ospf(bdr) != 00000000 && $s(OSPF,bdr) == 00000000 } {
         set s(OSPF,bdr) $ospf(bdr) 
         pktsrcInfo s "Selecting [hextoip $ospf(bdr)] as the BDR" 
   }

   # -- check for Exchange

   # -- if this is the DR or BDR enter ExStart with this neighbor

   if {  $s(OSPF,neighbor,$ospf(routerID),state) == "2-Way"  &&
            ( "$s(OSPF,dr)" == "$ip(src)" ||
              "$s(OSPF,bdr)" == "$ip(src)" ) } {

       # -- Enter the ExStart state

       set s(OSPF,neighbor,$ospf(routerID),state) ExStart
       set s(OSPF,neighbor,$ospf(routerID),seq) [ format "%8.8x" [ clock clicks ]]
       pktsrcInfo s "Start neighbor state ExStart with DR/BDR neighbor"
       OSPF_Exstart s $ospf(routerID)
   }
 
}


proc OSPF_dd_input { packet stack { options none }} {
   upvar $stack s

   # -- I should hide this
   set ipState $s(current,IPv4)
   upvar #$s(current,IPv4,level) $ipState ip

   # -- I should hide this
   set ospfState $s(current,OSPF)
   upvar #$s(current,OSPF,level) $ospfState ospf

   set port $s(port)

   #
   # -- check the neighbor state
   if [ info exists s(OSPF,neighbor,$ospf(routerID)) ] {


         # -- get the state for this neighbor
         set state $s(OSPF,neighbor,$ospf(routerID),state)

         pktsrcInfo  s \
            "Received DD from neighbor [hextoip $ospf(routerID)] state: $state I:$ospf(I_bit) M:$ospf(M_bit) MS:$ospf(MS_bit) Seq 0x$ospf(ddSeq)" 

         OSPF_retransmission_list s ospf

         # -- EXSTART

         if { $state == "ExStart" } {
           pktsrcInfo s "checking neighbor:[hextoip $ospf(routerID)] for MS/M/I bits"
	   if { $ospf(I_bit) == 1 && $ospf(M_bit) == 1 && $ospf(MS_bit) == 1 } {
	       if { [ expr 0x$ospf(routerID) > 0x$s(OSPF,routerID) ] } {
                   # -- I am now a slave
                   set s(OSPF,neighbor,$ospf(routerID),MS) 0
                   set s(OSPF,neighbor,$ospf(routerID),state) ExChange
                   pktsrcInfo s "Entering state ExChange with neighbor $ospf(routerID)"
                   OSPF_Exchange s $ospf(routerID)
	       } else {
                   pktsrcInfo s "neighbor:[hextoip $ospf(routerID)] has lesser router ID [hextoip $s(OSPF,routerID)]"
                   pktsrcInfo s "Ignoring this DD packet since I should become DD master"
                   # -- remember the sequence number!!

               }
           } elseif {  $ospf(I_bit) == 0 && $ospf(MS_bit) == 0 } {

               pktsrcInfo s "neighbor [hextoip $ospf(routerID)] looks like slave"

	       if { $ospf(ddSeq) == $s(OSPF,neighbor,$ospf(routerID),seq) } {
                  pktsrcInfo s "DD sequence numbers match, I am the master"
                   # -- I am now master
                   set s(OSPF,neighbor,$ospf(routerID),MS) 1
                   # -- go to exchange
                   pktsrcInfo s "Entering state ExChange with neighbor [hextoip $ospf(routerID)]"
                   set s(OSPF,neighbor,$ospf(routerID),state) ExChange

                   OSPF_Exchange s $ospf(routerID)

               } else {
                   pktsrcInfo s "DD sequence mismatch (expect $ospf(ddSeq) got $s(OSPF,neighbor,$ospf(routerID),seq))"
               }
           }

       #
       # -- EXCHANGE or FULL

       }  elseif { $state == "ExChange" || $state == "Full" } {

           # if they are the Master, get the dd sequence number

            if { $ospf(MS_bit) } {
               set s(OSPF,neighbor,$ospf(routerID),seq) $ospf(ddSeq)
            }

            # -- see if they have any more to send
            if { $ospf(M_bit) } {
               pktsrcInfo s "Neighbor indicated more DD records"
               OSPF_Exchange s $ospf(routerID)
            } else {
               pktsrcInfo s "Neighbor indicated no more DD records"

               # -- If I'm the slave I still need to acknowledge the DD
               if { $s(OSPF,neighbor,$ospf(routerID),MS) == 0 } {
                  OSPF_Exchange s $ospf(routerID)
               }
            }

       }  elseif { $state == "2-Way" } {
             pktsrcInfo s \
               "Ignorning DD packet from neighbor [hextoip $ospf(routerID)]" 
             pktsrcInfo s \
               "Still in state 2-Way with [hextoip $ospf(routerID)], no DR elected"

       } else {

             # -- ALL OTHER STATES
             pktsrcInfo s \
               "Ignorning DD packet from neighbor [hextoip $ospf(routerID)]" 
       }

   } else {
     # -- ignore this packet
     pktsrcInfo  s \
       "Ignoring DD packet from unknown neighbor [hextoip $ospf(routerID)]"  
   }
 
}

proc OSPF_retransmission_list { stack ospf } {
   upvar $stack s
   upvar $ospf o

   foreach lsa $o(lsa) {

      OSPF_decode_lsa_header $lsa out

      if { $out(lsID) == $s(OSPF,routerID) && $out(router) == $s(OSPF,routerID) } {
        pktsrcInfo s \
            "DD contains my old router LSA with seq 0x$out(seq), bumping seq"
        set s(OSPF,routerLSA) [ OSPF_bump_seq $s(OSPF,routerLSA) $out(seq) ]
      }
   }
}

proc OSPF_update_input { packet stack { options none }} {
   upvar $stack s

   # -- get IP state
   set ipState $s(current,IPv4)
   upvar #$s(current,IPv4,level) $ipState ip

   # -- get the OSPF packet info
   set ospfState $s(current,OSPF)
   upvar #$s(current,OSPF,level) $ospfState ospf

   pktsrcInfo s "LSA Update from [hextoip $ospf(routerID)]"

   # -- The default behavior is to send an LSAck for everything
   # -- it would be nice to check if I originated the LSA ...

   set update ""

   # -- parse through all LSAs, grabbing the headers

   for { set i 1 } { $i <= $ospf(lsaCount) } { incr i } {  
      set lsaH [ string range $ospf(lsa,$i) 0 39 ]
      set update "$update$lsaH"
   }

   # -- build a LSA ACK header
   set header [ OSPF_build_header s "05" ]

   pktsrcInfo s "Sending LSA Ack to [hextoip $ip(src)]"

   # -- send the packet

   # +++joem I should not assum the routerID is the same as the interface
   # address - better to grab the IP src
   OSPF_send s "$header$update" $ip(src)
 
}

proc OSPF_ack_input { packet stack { options none }} {
   upvar $stack s

   # -- get the OSPF packet info
   set ospfState $s(current,OSPF)
   upvar #$s(current,OSPF,level) $ospfState ospf

   pktsrcInfo s "LSA Ack from [hextoip $ospf(routerID)]"
 
}

proc OSPF_request_input { packet stack { options none }} {
   upvar $stack s

   # -- get the OSPF packet info
   set ospfState $s(current,OSPF)
   upvar #$s(current,OSPF,level) $ospfState ospf

   pktsrcInfo s "LSA Request from [hextoip $ospf(routerID)]"

   # -- Assume this is for my router LSA, which is the only
   # thing I advertise during DD exchange

   # -- send an update for my router LSA
   OSPF_send_lsa_update s $s(OSPF,routerLSA)
 
}


proc OSPF_send_hello { stack } {
   upvar $stack s

   # -- send a hello packet
   set hello [ OSPF_build_hello s ]

   OSPF_send s $hello

}

proc OSPF_send { stack packet { dest e0000005 } } {
   upvar $stack s

   # -- set the correct length
   set length [ format "%4.4x" [expr [ string length $packet ] /2 ]]
   set h [ string replace $packet 4 7 $length ]
 
   # -- set the checksum or md5

   if { "$s(OSPF,auth)" == "0002" } {

     # -- compute the md5
     set md5 [ OSPF_md5_packet $h $s(OSPF,authSecret) ]

     # -- append to ospf packet
     set h2 "$h$md5"

   } else {

     # -- compute the checksum

     # -- 0 out the authdata
     set p2 [ string replace $h 32 47 0000000000000000 ]

     # -- compute
     set checksum [ pktsrc checksum $p2 ]

     # -- replace   
     set h2 [ string replace $h 24 27 $checksum ]

   }

   # -- send the packet on to IP
   IPv4_send $h2 s $dest 59

}

proc OSPF_build_header { stack type } {
   upvar $stack s
   set header [ format "02%s0000%s%s0000%s%s" \
        $type $s(OSPF,routerID) $s(OSPF,areaID) $s(OSPF,auth) \
        $s(OSPF,authData) ]

   return $header
}


proc OSPF_build_hello { stack } {
   upvar $stack s

   set header [ OSPF_build_header s "01" ]
   
   set nl ""

   # -- build neighbor list
   foreach n $s(OSPF,neighborList) {
       set nl "$nl$n"
   }

   # -- build body
   set body [ format "%s%s%s%s%s%s%s%s" \
      $s(netmask) [ format "%4.4x" $s(OSPF,helloInt) ] \
      $s(OSPF,options) [ format "%2.2x" $s(OSPF,priority)]  \
      [ format "%8.8x" $s(OSPF,deadInt) ] \
       $s(OSPF,dr) $s(OSPF,bdr) $nl ]

 
   return "$header$body"

}

proc OSPF_build_dd { stack neighbor i m ms } {
   upvar $stack s

   set header [ OSPF_build_header s "02" ]
   
   set bits [format "%2.2x" [ expr ($i << 2)  | ($m << 1) | ($ms) ]]

   # -- build body
   set body [ format "000002%s%s" $bits $s(OSPF,neighbor,$neighbor,seq) ]
 
   pktsrcInfo s "Building DD with sequence number 0x$s(OSPF,neighbor,$neighbor,seq)"
   return "$header$body"
}


proc OSPF_Exstart { stack neighbor } {

   # -- send initial DD packet
   upvar $stack s

   # -- If I'm still in ExStart, start the DD process
   if { $s(OSPF,neighbor,$neighbor,state) == "ExStart" } {

      pktsrcInfo s "Building my Router LSA"
      OSPF_build_database s

      # -- send a response
      set dd [ OSPF_build_dd s $neighbor 1 1 1 ]

      pktsrcInfo s "Sending OSPF DD Packet I:1 M:1 MS:1"

      OSPF_send s $dd $s(OSPF,neighbor,$neighbor,srcIP)
     
      # -- start rxTimer ?
      after [ expr $s(OSPF,rxInt) * 1000 ]  "
          $s(getStack)
         OSPF_Exstart $s(getStackName) $neighbor 
      "
   }
}

proc OSPF_Exchange { stack neighbor } {


   # -- get the stack
   upvar $stack s

   # -- get the last ospf packet
   set ospfState $s(current,OSPF)
   upvar #$s(current,OSPF,level) $ospfState ospf

   # -- bump my sequence number if I am the master
   if { $s(OSPF,neighbor,$neighbor,MS) == 1 } {
        set s(OSPF,neighbor,$neighbor,seq) [ incrhex $s(OSPF,neighbor,$neighbor,seq)]
   }

   pktsrcInfo s "Sending OSPF DD Packet with Router LSA Header I:0 M:0 MS:$s(OSPF,neighbor,$neighbor,MS)"

   # -- send a response
   set dd [ OSPF_build_dd s $neighbor 0 0 $s(OSPF,neighbor,$neighbor,MS) ]
   
   # -- add the router LSA header to the end
   set packet "$dd[string range $s(OSPF,routerLSA) 0 39 ]"
   
   OSPF_send s $packet $s(OSPF,neighbor,$neighbor,srcIP)
  
   # -- if no more packets DD packets, I should be ready to enter
   # Full with this neighbor ...

   if { "$ospf(M_bit)" == "0" } {

      # -- Now, I should be in Full
      if { $s(OSPF,neighbor,$neighbor,state) != "Full" } {
         set s(OSPF,neighbor,$neighbor,state) Full
         pktsrcInfo s "Entering state Full with neighbor [hextoip $neighbor]"  
      }

      # -- send an update for my router LSA
      OSPF_send_lsa_update s $s(OSPF,routerLSA)
   }

}

proc OSPF_build_database { stack } {
   upvar $stack s

   # -- create my own LSA database
   OSPF_create_router_lsa s lsa
   OSPF_router_lsa_add_link lsa [hextoip $s(OSPF,dr)] [hextoip $s(ip)] 2 10

   # -- save my router LSA
   set s(OSPF,routerLSA) [ OSPF_build_router_lsa lsa ]

}

proc OSPF_delete_neighbors { stack } {
   upvar $stack s

   # -- unset any neighbors
   if [ info exists s(OSPF,neighborList) ] {
     foreach n $s(OSPF,neighborList) {
        unset s(OSPF,neighbor,$n)
        unset s(OSPF,neighbor,$n,state)
     }
   }

   # -- set the new list
   set s(OSPF,neighborList) ""

   set s(OSPF,dr) 00000000
   set s(OSPF,bdr) 00000000
 
}


proc OSPF_delete_neighbor { stack neighbor } {
   upvar $stack s

   # unset the neighbor
   unset s(OSPF,neighbor,$neighbor)

   # -- leave the other state

   # -- rebuild the neighbor list
   set nlist ""
   foreach n $s(OSPF,neighborList) {
       if { "$n" != "$neighbor" } {
          lappend nlist $n
       }
   }

   # -- set the new list
   set s(OSPF,neighborList) $nlist

   # +++joe I need a real dr election!

   # -- make sure this wan't the dr or bdr
   if { "$s(OSPF,dr)" == "$n" } {
       set s(OSPF,dr) 00000000
   }

   if { "$s(OSPF,bdr)" == "$n" } {
       set s(OSPF,bdr) 00000000
   }

 
}

proc OSPF_send_lsa_update { stack lsa } {
   upvar $stack s

   # -- build a header for an LSA
   set header [ OSPF_build_header s "04" ]

   set type [ string range $lsa 6 7 ]
   # -- set the lsa count to 1 for 1 router lsa

   pktsrcInfo s "Sending LSA Update [OSPF_lsa_type $type]"

   # -- send along
   OSPF_send s "${header}00000001$lsa"

}


# -- wait timeOut msecs
# The evnt loop continues to run

proc OSPF_wait { timeOut } {
   puts "OSPF: Waiting $timeOut msec"
   Event_wait $timeOut
}


#
# -- This is an event loop that waits until an OSPF neighbor reaches
# a certain state.

proc OSPF_wait_state { stack neighbor state timeOut } {
   upvar $stack s
   set run 1
   set elapsed 0
   set status 0

   while { $run } {
      if [ info exists s(OSPF,neighbor,$neighbor,state) ] {

         # -- get the current state for this neighbor
         set current $s(OSPF,neighbor,$neighbor,state)

         # -- if it matches we are done
         if { "$current" == "$state" } {
            set status 1
        
            # -- return a positive status
            return $status
         }
      }

      # -- this is wait 1000 seconds and process events
      Event_wait 1000

      # -- check for a timeout
      incr elapsed 1
      if { $elapsed >= $timeOut } {
        return $status
      }
   }
}

proc OSPF_md5_packet { packet secret } {

   # -- padd out the secret to 16 bytes
   set secretpad [ padhex [ pktsrc text2ascii $secret ] 16 ]
   
   # -- compute the md5 over the packet and secret together
   set md5 [ pktsrc md5 $packet$secretpad ]

   return $md5
}

proc OSPF_md5 { stack } {
   upvar $stack s

   # -- get the contents of the IP packet from the stack
   set ipState $s(current,IPv4)
   upvar #$s(current,IPv4,level) $ipState ip

   # -- get the OSPF packet
   set length [ string length $ip(data) ]

   # -- chop off last 16 bytes
   set ospf [ string range $ip(data) 0 [ expr $length - 33 ]]

   return [ OSPF_md5_packet $ospf $s(OSPF,authSecret) ]
}