#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# 2002 QAcafe http://qacafe.com
#
# $Id: pppoe.tcl,v 1.9 2002/11/26 19:20:20 joe Exp $
#
#
# pktsrc PPPoE
#
# ----------------------------------------------------------------------------

# -- this is part of the pktsrc package
package provide pktsrc 1.3

#
# Set the PPPoE filter
#

proc PPPoE_set_standard_filter { stack } {
   upvar $stack s

   set port $s(port)

   # -- set the standard filter to receive only PPPoE related traffic
   $port filter "$s(PPPoE,standard-filter)"
}

#
# Get the PPPoE filter
#

proc PPPoE_get_standard_filter { stack } {
   upvar $stack s

   return "$s(PPPoE,standard-filter)"
}


#
# PPPoE main input callback
#

proc PPPoE_input { packet stack { options none }} {
   upvar $stack s

   set port $s(port)

   if { $options == "print" } {
      PPPoE_decode $packet pppoe
      PPPoE_print pppoe
      return
   }

   incr s(PPPoE,inpackets)

   # -- remember this packet in the 'pppoe' variable. This allows other
   # callbacks to get access to the decodes at any Tcl stack level

   eval [ Stack_save_packet s PPPoE pppoe ]
   
   # -- now decode the packet into the 'pppoe' variable
   PPPoE_decode $packet pppoe

   # -- check for callbacks 

   if [ info exists s(pppoeType,$pppoe(code)) ] {

      # -- drive any input callbacks
      foreach c $s(pppoeType,$pppoe(code),callbacks) {
          # -- send the whole packet
          $c $packet s $options
      }

   } else {
      # no callback
   }
}

#
# PPPoE input callback for estbalished PPPoE session
#

proc PPPoE_session_input { packet stack { options none }} {
   upvar $stack s

   set port $s(port)
   PPPoE_session_decode $packet pppoe


   if { $options == "print" } {
      PPPoE_session_print pppoe
   } else {

      # -- I should hide this in a save proto
      set s(current,PPPoE) pppoe
      set s(current,PPPoE,level) [info level]
   }

   # -- check for callbacks 

   if [ info exists s(pppoeSType,$pppoe(code)) ] {

      # -- drive any input callbacks
      foreach c $s(pppoeSType,$pppoe(code),callbacks) {

         if [ catch { $c $pppoe(data) s $options } reason ] {
            puts "PKTSRC-ERROR: input callback '$c' failed"
            puts "PKTSRC-ERROR: $reason"
         }
      }
   } else {
      # no callback
   }

}

#
#
# PPPoE_init - set up any deaults for PPPoE
#

proc PPPoE_init { stack } {
   upvar $stack s

   # -- set the default version
   set s(PPPoE,version) 1

   # -- current session id
   set s(PPPoE,session-id) 0
   set s(PPPoE,service) ""
   set s(PPPoE,ac-mac) 0
   set s(PPPoE,ac-name) qacafe-ac
   set s(PPPoE,next-session-id) 0000
   set s(PPPoE,standard-filter)  "ether proto 0x8863 or ether proto 0x8864"

   # -- initial packet counters
   set s(PPPoE,inpackets) 0
}

proc PPPoE_server_next_session { stack } {
  upvar $stack s

  # -- bump the session ids
  set s(PPPoE,next-session-id) [ incrhex16 $s(PPPoE,next-session-id) ]

  # +++joem I really need a session table hear to support multiple
  # clients in server mode. But for now, just remeber the last session
  # id and assume its the clients id

  set s(PPPoE,session-id) [ expr 0x$s(PPPoE,next-session-id) ]

  # -- return the current id
  return $s(PPPoE,next-session-id)

}

proc PPPoE_decode { packet PPPoE } {
   upvar $PPPoE out

   # -- parse out all the fields we find interesting ...

 
#                       1                   2                   3
#    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
#   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#   |  VER  | TYPE  |      CODE     |          SESSION_ID           |
#   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#   |            LENGTH             |           payload             ~
#   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#


   set out(version)        [ string range $packet 0 0 ]
   set out(type)           [ string range $packet 1 1 ]
   set out(code)           [ string range $packet 2 3 ]
   set out(session-id)     [ string range $packet 4 7 ]
   set out(length)         [ string range $packet 8 11 ]

   set out(tags) 0

   # -- check for tags to decode   
   set len [ string length $packet ]

   if { $len > 12 } {
      PPPoE_decode_tags $packet out
   }
}

proc PPPoE_session_decode { packet PPPoE } {
   upvar $PPPoE out

   # -- parse out all the fields we find interesting ...

 
#                       1                   2                   3
#    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
#   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#   |  VER  | TYPE  |      CODE     |          SESSION_ID           |
#   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#   |            LENGTH             |           payload             ~
#   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#

   set out(version)        [ string range $packet 0 0 ]
   set out(type)           [ string range $packet 1 1 ]
   set out(code)           [ string range $packet 2 3 ]
   set out(session-id)     [ string range $packet 4 7 ]
   set out(length)         [ string range $packet 8 11 ]

   set out(tags) 0

   # -- the pppoe payload
   if { [ string compare "$out(length)" "0000" ] == 0 } {
       set out(data) ""
   } else {

      # -- set the pppoe payload based on the specified length, ignore any
      # trailing data after the payload
      set out(data)    [ string range $packet 12 [ expr 0x$out(length) * 2 + 11 ]]
   }
}


proc PPPoE_decode_tags { packet pppoe } {
   upvar $pppoe out

   set p 12

   set out(tags) 0

   set count 0
   set totalLen 0

   while { 1 } {
     
      if { $totalLen >= [ expr 0x$out(length) ] } {
         break
      } 
      set tag [ string range $packet $p [expr $p + 3]]
      # puts "tag is $tag"

      if { $tag == "0000" } {
        # -- end of tags
        break
      }

      incr p 4
      set length [ string range $packet $p [expr $p + 3]]
      # puts "length is $length"

      incr p 4

      set out(option,$count,tag) $tag
      set out(option,$count,length) $length

      if { "$length" == "0000" } {
         set out(option,$count,data) ""
      } else {
         set out(option,$count,data) \
             [ string range $packet $p [expr $p + 2 * 0x$length - 1 ]]
         incr p [ expr 0x$length * 2  ]
      }

      # puts "data is $out(option,$count,data)"
      incr count

      incr totalLen [ expr 4 + 0x$length ]
      # puts "total length is $totalLen"
   } 

   # -- set the total number of tags found
   set out(tags) $count  
}



proc PPPoE_print { PPPoE } {
   upvar $PPPoE out

   puts "PPPoE: ---- PPPoE Packet ----"
   puts "PPPoE: "
   puts "PPPoE: Version                  = [expr 0x$out(version)]"
   puts "PPPoE: Type                     = [expr 0x$out(type)]"
   puts "PPPoE: Code                     = [PPPoE_get_code $out(code) ] (0x$out(code))"
   puts "PPPoE: Session ID               = [expr 0x$out(session-id)]"
   puts "PPPoE: Length                   = [expr 0x$out(length)]"

   puts "PPPoE: ---- Tags ----"
   puts "PPPoE:"

   if { $out(tags) > 0 } {
      PPPoE_print_tags out
   } else {
      puts "PPPoE: No tags found"
   }

   puts "PPPoE:"
   puts "PPPoE: ---- End PPPoE Packet ----"
}

proc PPPoE_session_print { PPPoE } {
   upvar $PPPoE out

   puts "PPPoE: ---- PPPoE Packet ----"
   puts "PPPoE: "
   puts "PPPoE: Version                  = [expr 0x$out(version)]"
   puts "PPPoE: Type                     = [expr 0x$out(type)]"
   puts "PPPoE: Code                     = [PPPoE_get_code $out(code) ] (0x$out(code))"
   puts "PPPoE: Session ID               = [expr 0x$out(session-id)]"
   puts "PPPoE: Length                   = [expr 0x$out(length)]"
   puts "PPPoE: "
}


proc PPPoE_get_code { code } {
  set str "Unknown"

  if { "$code" == "09" } {
    set str "Initiation (PADI)"
  } elseif { "$code" == "07" } {
    set str "Offer (PADO)"
  } elseif { "$code" == "19" } {
    set str "Request (PADR)"
  } elseif { "$code" == "65" } {
    set str "Session-confirmation (PADS)"
  } elseif { "$code" == "a7" } {
    set str "Terminate (PADT)"
  } elseif { "$code" == "00" } {
    set str "PPPoE Session"
  }

  return $str

}

proc PPPoE_print_tags { pppoe } {
   upvar $pppoe out

   for { set i 0 } { $i < $out(tags) } { incr i } {

      # -- get the option tag in decimal

      set tag $out(option,$i,tag)
      set length $out(option,$i,length)
      set data $out(option,$i,data)

      if { "$tag" == "0101" } {

         # -- check for 'Any Service'
	 if { $length == 0 } {
            set name "\"\" (Any Service)"
	 } else {
            set name [ pktsrc ascii2text $data ] 
         }  

         puts "PPPoE: Service Name             = $name"

      } elseif { "$tag" == "0102" } {

         set name [ pktsrc ascii2text $data ] 
         puts "PPPoE: AC-Name                  = $name"

      } elseif { "$tag" == "0103" } {

         puts "PPPoE: Host-Uniq                = $data"

      } elseif { "$tag" == "0104" } {

         puts "PPPoE: AC-Cookie                = $data"

      } elseif { "$tag" == "0105" } {

         puts "PPPoE: Vendor-Specific          = $data"

      } elseif { "$tag" == "0110" } {

         puts "PPPoE: Relay-Session-Id         = $data"

      } elseif { "$tag" == "0201" } {

         set name [ pktsrc ascii2text $data ] 
         puts "PPPoE: Service-Name-Error       = $name"

      } elseif { "$tag" == "0202" } {

         set name [ pktsrc ascii2text $data ] 
         puts "PPPoE: AC-System                = $name"

      } elseif { "$tag" == "0203" } {

         set name [ pktsrc ascii2text $data ] 
         puts "PPPoE: Generic-Error            = $name"

      } else {

         # -- Unknown Tag display
         puts "PPPoE: Tag Type                 = 0x$tag"
         puts "PPPoE: Tag Length               = 0x$length"
         puts "PPPoE: Tag Data                 = 0x$data"
      }
   }
}

proc PPPoE_send_padi { stack service { opts "" } { add end } } {
   upvar $stack s

   if { $opts != "" } {
       upvar $opts o
   } else {
       set o(tags) 0
   }

   # -- build padi header
   set header [ PPPoE_build_header s 09 ]

   # -- add my tags for padi
   set options(tags) 0

   if { "$add" == "end" } {

      # -- add the service tag option first
      PPPoE_add_service_tag options $service

      # -- now copy any other tags
      PPPoE_copy_all_tags o options

   } else {

      # -- copy any other tags first
      PPPoE_copy_all_tags o options

      # -- now add the service tag
      PPPoE_add_service_tag options $service
   }

   # -- build the tags
   set opt [ PPPoE_build_tags options ]

   pktsrcInfo s \
      "Sending PADI for service '$service' to [hextomac ffffffffffff]"

   # -- send this in an PPPoE discovery packet
   Ethernet_send s "${header}${opt}" 8863 ffffffffffff

}

proc PPPoE_send_padr { stack service destmac } {
   upvar $stack s

   # -- build padr header
   set header [ PPPoE_build_header s 19 ]

   # -- add my tags for padi
   set options(tags) 0

   # -- add the service tag
   PPPoE_add_service_tag options $service

   # -- build the tags
   set opt [ PPPoE_build_tags options ]

   pktsrcInfo s "Sending PADR for service '$service' to [ hextomac $destmac ]"

   # -- send this in an PPPoE discovery packet
   Ethernet_send s "${header}${opt}" 8863 $destmac

}

proc PPPoE_send_pado { stack ops destmac } {
   upvar $stack s
   upvar $ops options

   # -- build padr header
   set header [ PPPoE_build_header s 07 ]

   # -- add the AC name
   PPPoE_add_ac_tag options $s(PPPoE,ac-name)

   # -- build the tags
   set opt [ PPPoE_build_tags options ]

   pktsrcInfo s "Sending PADO to [ hextomac $destmac ]"

   # -- send this in an PPPoE discovery packet
   Ethernet_send s "${header}${opt}" 8863 $destmac

}

proc PPPoE_send_pads { stack ops destmac } {
   upvar $stack s
   upvar $ops options

   set sessionId [ PPPoE_server_next_session s ]

   # -- build pads header
   set header [ PPPoE_build_header s 65 $sessionId ]

   # -- add the AC name
   PPPoE_add_ac_tag options $s(PPPoE,ac-name)

   # -- build the tags
   set opt [ PPPoE_build_tags options ]

   pktsrcInfo s \
      "Sending PADS with session id '$sessionId' to [ hextomac $destmac ]"

   # -- send this in an PPPoE discovery packet
   Ethernet_send s "${header}${opt}" 8863 $destmac

}


proc PPPoE_send_padt { stack ops destmac { session 0 }} {
   upvar $stack s
   upvar $ops options

   # -- use the specified session ID
   if { $session != 0 } {
      set sessionId $session
   } else {

      # -- use the default session ID
      set sessionId $s(PPPoE,session-id)
   }

   if { "$sessionId" == "0" } {

      pktsrcInfo s "No active session to terminate"

   } else {

      # -- build padt header
      set header [ PPPoE_build_header s a7 [ format "%4.4x" $sessionId ]]

      # -- no tags required for padt
      # -- should check ops anyways and build tags if provided

      pktsrcInfo s \
         "Sending PADT with session id '$sessionId' to [ hextomac $destmac ]"

      # -- send this in an PPPoE discovery packet
      Ethernet_send s "${header}0000" 8863 $destmac

      # -- mark the current session as dead
      set s(PPPoE,session-id) 0
   }
}

proc PPPoE_terminate_service { stack { service qacafe } } {
   upvar $stack s

   set destmac $s(PPPoE,ac-mac)
   pktsrcInfo s "Terminating PPPoE service with [hextomac $destmac]"

   # -- no tags required
   set ops(tags) 0

   # -- send a padt   
   PPPoE_send_padt s ops $destmac 

   # -- mark the statful PPP layers down
   set s(LCP,state) down
   set s(NCP,state) down
   set s(PAP,state) down
 
   set s(PPPoE,session-id) 0

   # -- reset LCP
   PPP_LCP_reset s
}

proc PPPoE_session_send { stack packet } {
   upvar $stack s

   set session $s(PPPoE,session-id)
   set destmac $s(PPPoE,ac-mac)

   if { "$session" == "0" } {
     pktsrcWarn s "Droping PPP packet, no PPPoE session exists"
     return
   }
 
   set length [ format "%4.4x" [expr [ string length $packet ] / 2 ]]

   # -- build padr header
   set header [ PPPoE_build_header s 00 [ format "%4.4x" $session ]]

   # -- send this in an PPPoE session packet
   Ethernet_send s "${header}${length}${packet}" 8864 $destmac
}

proc PPPoE_build_tags { op } {
   upvar $op o

   set oplist ""
   set totalLen 0

   for { set i 0 } { $i < $o(tags) } { incr i } {

       set tag    $o(option,$i,tag)
       set length $o(option,$i,length)
       set data   $o(option,$i,data)      
       set oplist ${oplist}${tag}${length}${data}      

       # -- use the real data length, not the supplied length

       set realLen [ expr [ string length $data ] / 2 ]
       incr totalLen [ expr 4 + $realLen ]
   }     

   # -- add an end marker -- 0000
   set len [ format "%4.4x" $totalLen ]
   set oplist "${len}${oplist}00000000"

   return $oplist
}

proc PPPoE_find_tag { op tag { from 0 }} {
   upvar $op o

   set oplist ""
   set totalLen 0

   for { set i $from } { $i < $o(tags) } { incr i } {

       if { "$tag" == "$o(option,$i,tag)" } {
           return $i
       }
   }

   return -1
}

proc PPPoE_build_header { stack code { session 0000 }} {

   upvar $stack s

   set version 1
   set type    1

   return ${version}${type}${code}${session}
 
}

proc PPPoE_add_service_tag { options name } {
    upvar $options o

    set len [ string length $name ]
    set nameHex [ pktsrc text2ascii $name ]

    set i $o(tags) 
    set o(option,$i,tag) 0101

    if { "$name" == "" } {
       set o(option,$i,length) 0000
       set o(option,$i,data) ""
    } else {
       set o(option,$i,length) [ format "%4.4x" $len ]
       set o(option,$i,data) $nameHex
    }
    incr o(tags)
}

proc PPPoE_add_hostuniq_tag { options data } {
    upvar $options o

    set i $o(tags) 
    set o(option,$i,tag) 0103

    set o(option,$i,length) [ format "%4.4x" [ expr [ string length $data ] / 2 ]]
    set o(option,$i,data) "$data"

    incr o(tags)
}

proc PPPoE_add_relaysessionid_tag { options data } {
    upvar $options o

    set i $o(tags) 
    set o(option,$i,tag) 0110

    set o(option,$i,length) [ format "%4.4x" [ expr [ string length $data ] / 2 ]]
    set o(option,$i,data) "$data"

    incr o(tags)
}


proc PPPoE_add_ac_tag { options name } {
   upvar $options o

   set len [ string length $name ]
   set nameHex [ pktsrc text2ascii $name ]

   set i $o(tags) 
   set o(option,$i,tag) 0102
   set o(option,$i,length) [ format "%4.4x" $len ]
   set o(option,$i,data) $nameHex

   incr o(tags)
}


proc PPPoE_add_tag { tags tag length data } {

   upvar $tags o

   set i $o(tags) 
   set o(option,$i,tag) [ format "%4.4x" $tag ]
   set o(option,$i,length) [ format "%4.4x" $length ]
   set o(option,$i,data) $data

   incr o(tags)

}

proc PPPoE_keep_tag { intags outtags tag } {

   upvar $intags in
   upvar $outtags out

   set index [ PPPoE_find_tag in $tag ]
   if { "$index" > -1 } {
        PPPoE_copy_tag in $index out
   }

}

proc PPPoE_copy_tag { intags index outtags } {

   upvar $intags in
   upvar $outtags out

   set i $out(tags)
 
   set out(option,$i,tag)    $in(option,$index,tag)
   set out(option,$i,length) $in(option,$index,length)
   set out(option,$i,data)   $in(option,$index,data)

   incr out(tags)

}

proc PPPoE_copy_all_tags { intags outtags } {
   upvar $intags in
   upvar $outtags out

   for { set i 0 } { $i < $in(tags) } { incr i } {
      PPPoE_copy_tag in $i out
   }
}


proc PPPoE_client_input { packet stack { options none }} {
   upvar $stack s

   # -- get the contents of the PPPoE packet from the stack
   set pppoeState $s(current,PPPoE)
   upvar #$s(current,PPPoE,level) $pppoeState pppoe

   # -- we will need the ethernet information as well
   set ethState $s(current,Ethernet)
   upvar #$s(current,Ethernet,level) $ethState eth  

   pktsrcInfo s "Client received [ PPPoE_get_code $pppoe(code) ] from [hextomac $eth(srcMac)]"

   if { $pppoe(code) == "07" } {
      PPPoE_send_padr s $s(PPPoE,service) $eth(srcMac)
   } elseif { $pppoe(code) == "65" } {

      pktsrcInfo s "New session ID [expr 0x$pppoe(session-id)] from [hextomac $eth(srcMac)]"

      # -- remember the session id
      set s(PPPoE,session-id) [expr 0x$pppoe(session-id)]
      set s(PPPoE,ac-mac) $eth(srcMac)

      # -- at this point PPP should be started
      PPP_client s
   }

}

proc PPPoE_server_input { packet stack { options none }} {
   upvar $stack s

   # -- get the contents of the PPPoE packet from the stack
   set pppoeState $s(current,PPPoE)
   upvar #$s(current,PPPoE,level) $pppoeState pppoe

   # -- we will need the ethernet information as well
   set ethState $s(current,Ethernet)
   upvar #$s(current,Ethernet,level) $ethState eth  

   pktsrcInfo s "Server received [ PPPoE_get_code $pppoe(code) ] from [hextomac $eth(srcMac)]"

   if { $pppoe(code) == "09" } {

      # -- look at the request AC name (ifany)

      # -- look at the requested service
      set index [ PPPoE_find_tag pppoe 0101 ]

      if { "$index" > -1 } {
         set service [ pktsrc ascii2text $pppoe(option,$index,data) ]

	 if { "$service" == "" } {
             set service $s(PPPoE,service)
             pktsrcInfo s "Any service requested, using '$service'"
         }

         set ops(tags) 0
         PPPoE_add_service_tag ops $service

         # -- keep the host-uniq tag if specified
         PPPoE_keep_tag pppoe ops 0103

         # -- now send
         PPPoE_send_pado s ops $eth(srcMac)

         # -- remember this guys mac, this should really be a add_client
         # call that does all of this book keeping

         set s(PPPoE,ac-mac) $eth(srcMac)

      } else {
         # ignore
         pktsrcWarn s "Received PADR with no service tag"
      }


   } elseif { $pppoe(code) == "19" } {

      # -- look at the request AC name (ifany)

      # -- look at the requested service
      set index [ PPPoE_find_tag pppoe 0101 ]

      if { "$index" > -1 } {
         set service [ pktsrc ascii2text $pppoe(option,$index,data) ]


         # -- allow a service name of ""
         
	 if { "$service" == "" } {

             if { "$s(PPPoE,service)" == "" } {
                pktsrcInfo s "PPPoE client requested any service"
             } else {
                pktsrcInfo s "No service requested in PADR"
                return
             }
         }

         # -- I should verify that the service name is correct
         if { "$service" != "$s(PPPoE,service)" } {
            pktsrcWarning s "Requested service '$service' does not match configured service '$s(PPPoE,service)'"
            return
         }

         set ops(tags) 0
         PPPoE_add_service_tag ops $service

         # -- keep the host-uniq tag if specified
         PPPoE_keep_tag pppoe ops 0103

         # -- now send
         PPPoE_send_pads s ops $eth(srcMac)

         PPP_server s

      } else {
         # ignore
         pktsrcWarn s "Received PADR with no service tag"
      }

  } elseif { "$pppoe(code)" == "a7" } {
     PPPoE_terminate_service s $s(PPPoE,service)
  }
}

proc PPPoE_client { stack { service qacafe } } {
   upvar $stack s

   pktsrcInfo s "Starting PPPoE client service"

   PPPoE_init s

   set s(PPPoE,service) $service

   if [ info exist s(PPPoE,handler) ] {
      # -- handler already installed
   } else {
      Stack_add_callback s pppoeType 07 PPPoE_client_input
      Stack_add_callback s pppoeType 65 PPPoE_client_input
      Stack_add_callback s pppoeType a7 PPPoE_client_input

      set s(PPPoE,handler) 1
   }

   # -- start things off with a PADI packet
   PPPoE_send_padi s $service

}

proc PPPoE_server { stack { service qacafe } } {
   upvar $stack s

   pktsrcInfo s "Starting PPPoE server for service '$service'"

   PPPoE_init s

   set s(PPPoE,service) $service

   if [ info exist s(PPPoE,shandler) ] {
      # -- handler already installed
   } else {
      Stack_add_callback s pppoeType 09 PPPoE_server_input
      Stack_add_callback s pppoeType 19 PPPoE_server_input
      Stack_add_callback s pppoeType a7 PPPoE_server_input
      set s(PPPoE,shandler) 1
   }

   # -- just wait for PADI

}

proc PPPoE_shutdown { stack } {
   upvar $stack s

   pktsrcInfo s "Shutting down PPPoE service"

   if { $s(PPPoE,service) != "" } {
       PPPoE_terminate_service s $s(PPPoE,service)
   }   

   # -- reinit PPPoE vars, this will reset the session ids as well
   PPPoE_init s

}

#
# PPPoE_Server_has_client
#
# wait for a client
#

proc PPPoE_server_has_client { stack { timeOut 30000 } } {
    upvar $stack s

    set sessId 0
    set waiting 1
    set elapsed 0

    pktsrcInfo s "Waiting up to $timeOut msec for WAN link to establish"
    while { 1 } {

       Event_wait 5000
       incr elapsed 5000
       set waitPPP 0

       pktsrcInfo s "Checking the status of the PPPoE link"
       if { "$s(LCP,state)" == "up" } {
             # pktsrcInfo s "Ok, PPP LCP is operational"
       } else {
             set waitPPP 1 
             pktsrcInfo s "PPP LCP is down"
       }

       if { "$s(NCP,state)" == "up" } {
             # pktsrcInfo s "Ok, PPP NCP/IPCP is operational"
       } else {
             set waitPPP 1 
             pktsrcInfo s "PPP NCP/IPCP is down"
       }

       if { "$s(PAP,mode)" == "on" } {
          if { "$s(PAP,state)" == "up" } {
               # pktsrcInfo s "Ok, PPP PAP succeeded"
          } else {
               set waitPPP 1 
               pktsrcInfo s "PPP PAP has not succeeded"
          }
       }

       if { "$s(CHAP,mode)" == "on" } {
          if { "$s(CHAP,state)" == "up" } {
                # pktsrcInfo s "Ok, PPP CHAP succeeded"
          } else {
               set waitPPP 1 
               pktsrcInfo s "PPP CHAP has not succeeded"
          }
       }

       if { "$waitPPP" == "0" } {
             pktsrcInfo s "Ok, PPPoE link is ready"
             set sessId $s(PPPoE,session-id)
             break
       }

       if { "$elapsed" >= "$timeOut" } {
             break
       }
   }

   if { "$waitPPP" == "1" } {
      pktsrcInfo s "Unable to bring up the PPP link after $timeOut msec"
   } else {
      pktsrcInfo s "Waiting 10 seconds"
      Event_wait 10000
   }

   return $sessId
}

proc PPPoE_generic_input { packet stack { options none }} {
   upvar $stack s

   # -- trigger the packet event
   if { "$options" != "print" } {
      Event_got_packet
   }
}

proc PPPoE_get_packet { stack type packet timeOut } {
   upvar $stack s
   set ret -1

   # -- add a callback for packets
   Stack_add_callback s pppoeType $type PPPoE_generic_input

   pktsrcInfo s "waiting up to $timeOut msecs for [ PPPoE_get_code $type ] packet"
   set result [ Event_wait_packet $timeOut ]

   if { "$result" == "2" } {

      # -- get the contents of the PPPoE packet from the stack
      set pppoeState $s(current,PPPoE)
      upvar #$s(current,PPPoE,level) $pppoeState pppoe
      upvar $packet p

      # -- copy the packet to pass back to the user
      pktsrc_copy_packet pppoe p
      set ret 0

   } elseif { "$result" == "1" } {
      pktsrcInfo s "timeout waiting for [ PPPoE_get_code $type ] packet" 
      set ret -1
   } else {
      set ret -1
   }

   Stack_remove_callback s pppoeType $type PPPoE_generic_input

   return $ret
}



