#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# 2001 QAcafe http://qacafe.com
#
# $Id: stack.tcl,v 1.28 2002/11/27 13:53:19 joe Exp $
#
#
# Basic Stack
#
# ----------------------------------------------------------------------------

# -- this is part of the pktsrc package
package provide pktsrc 1.3

proc bgerror { y } {

    puts "TCL ERROR: $y, level [ info level ]"
    # -- do a stack trace
    for { set x [expr [info level]-1] } { $x > 0 } { incr x -1 } {
       puts "$x: [ info level $x ]"
    }
}

#
# Stack_print
#
# Configure the default print mode
# Stacks initialized after this point, will have the specified mode
#

proc Stack_print { { mode on } } { 
   global _default_stack_print_mode 

   set _default_stack_print_mode $mode
}

#
# Stack_dump
#
# Configure the default dump mode
# Stacks initialized after this point, will have the specified mode
#

proc Stack_dump { { mode on } } { 
   global _default_stack_dump_mode 

   set _default_stack_dump_mode $mode
}

#
# Stack_log
#
# Configure the default log mode
# Stacks initialized after this point, will have the specified mode
#

proc Stack_log { { mode on } } { 
   global _default_stack_log_mode 

   set _default_stack_log_mode $mode
}

#
# Stack_init
#
# This is the default pktsrc stack. It sets intial properties on the
# stack and sets up protocol callbacks for the default set of protocols.
#

proc Stack_init { stack port ip mask { name none } { level -1 } { vnode 0 } } {
   global _default_stack_print_mode
   global _default_stack_dump_mode
   global _default_stack_log_mode

   upvar $stack mystack

   # -- check name for '-' characters since variables may be
   # created based on the name (TBD)

   # -- capture the stack level

   #
   # If the caller did not specify the level, user the callers
   # level. However, if the caller did specify the level, use it.
   # This allows stacks to be declared as global, but them initialize
   # at a different level.

   if { "$level" == "-1" } {

     # -- set the level to the callers level
     set level [ expr [ info level ] - 1 ]
   } 
     
   # -- internal variables used by Stack only!
   set mystack(_level) $level
   set mystack(_varname) $stack
   set mystack(_vnode) $vnode

   if [ info exists mystack(port) ] {
      puts "WARNING: stack '$stack' may already be initialized"
   }

   set mystack(port) $port
   
   # -- set the MAC address default to the real MAC address
   # skip if vnode

   if { "$vnode" != "1" } {
      set mystack(mac)  [ $port hwaddr ]
   } else {
      # -- use a dummy mac address for vnodes
      set mystack(mac) 000000000000
   }

   set mystack(ip)   [iptohex $ip ]

   # -- default netmask
   set mystack(netmask)  [ iptohex $mask ]


   # -- my network
   set mystack(network) [ format "%8.8x" \
        [ expr 0x$mystack(netmask) & 0x$mystack(ip) ]]


   # -- the default stack name is it's IP address
   if { "$name" == "none" } {
      set mystack(name) $ip
   } else {
      set mystack(name) $name
   }


   # -- turn on timestamps for this stack
   set mystack(timestamp) on

   # -- set up the stack's print mode based on any global default, if set
   if [ info exists _default_stack_print_mode ] {
      set mystack(print) $_default_stack_print_mode
   }
   

   # -- set up the default stack dump mode, skip if vnode
   if { "$vnode" != "1" } {
      if [ info exists _default_stack_dump_mode ] {
         set mystack(dump) $_default_stack_dump_mode
         $port dump $_default_stack_dump_mode
      }
   }

   # -- set up the default stack log mode
   if [ info exists _default_stack_log_mode ] {
      set mystack(log) $_default_stack_log_mode
   } else {
      set mystack(log) on
   }


   # NOTE: I really should seperate the user callbacks from the
   # built in stack callbacks. I don't want the users to have to
   # explicitly deal with the print callback.
   #
   # For now, any user callback may actually get called twice, once
   # for printing and once for processing.
   #

   # This stack initialization should be moved to the protcols Init
   # function.

   # I need a dynamic mechanism to add a new protocol rather than
   # list *_Init for everything. Stack_register ?
   #

   # -- add default callbacks for ARP
   set mystack(ethType,0806)           enabled
   set mystack(ethType,0806,decode)    ARP_decode
   set mystack(ethType,0806,print)     ARP_print
   set mystack(ethType,0806,input)     ARP_input
   set mystack(ethType,0806,callbacks) ARP_input

   # -- add default callbacks for IPv4
   set mystack(ethType,0800)           enabled
   set mystack(ethType,0800,decode)    IPv4_decode
   set mystack(ethType,0800,print)     IPv4_print
   set mystack(ethType,0800,input)     IPv4_input
   set mystack(ethType,0800,callbacks) IPv4_input

   # -- add default callbacks for PPPoE

   # -- PPPoE discovery
   set mystack(ethType,8863)           enabled
   set mystack(ethType,8863,decode)    PPPoE_decode
   set mystack(ethType,8863,print)     PPPoE_print
   set mystack(ethType,8863,input)     PPPoE_input
   set mystack(ethType,8863,callbacks) PPPoE_input

   # -- PPPoE session
   set mystack(ethType,8864)           enabled
   set mystack(ethType,8864,decode)    PPPoE_session_decode
   set mystack(ethType,8864,print)     PPPoE_session_print
   set mystack(ethType,8864,input)     PPPoE_session_input
   set mystack(ethType,8864,callbacks) PPPoE_session_input

   # -- pass PPP sessions up to PPP
   set mystack(pppoeSType,00)           enabled
   set mystack(pppoeSType,00,decode)    PPP_decode
   set mystack(pppoeSType,00,print)     PPP_print
   set mystack(pppoeSType,00,input)     PPP_input
   set mystack(pppoeSType,00,callbacks) PPP_input

   # -- lcp handlers
   set mystack(pppType,c021)              enabled
   set mystack(pppType,c021,decode)       PPP_LCP_decode
   set mystack(pppType,c021,print)        PPP_LCP_print
   set mystack(pppType,c021,input)        PPP_LCP_input
   set mystack(pppType,c021,callbacks)    PPP_LCP_input

   # -- pap handlers
   set mystack(pppType,c023)              enabled
   set mystack(pppType,c023,decode)       PPP_PAP_decode
   set mystack(pppType,c023,print)        PPP_PAP_print
   set mystack(pppType,c023,input)        PPP_PAP_input
   set mystack(pppType,c023,callbacks)    PPP_PAP_input

   # -- chap handlers
   set mystack(pppType,c223)              enabled
   set mystack(pppType,c223,decode)       PPP_CHAP_decode
   set mystack(pppType,c223,print)        PPP_CHAP_print
   set mystack(pppType,c223,input)        PPP_CHAP_input
   set mystack(pppType,c223,callbacks)    PPP_CHAP_input

   # -- ncp/ipcp handlers
   set mystack(pppType,8021)              enabled
   set mystack(pppType,8021,decode)       PPP_NCP_decode
   set mystack(pppType,8021,print)        PPP_NCP_print
   set mystack(pppType,8021,input)        PPP_NCP_input
   set mystack(pppType,8021,callbacks)    PPP_NCP_input

   # -- ip/pppoe handlers
   set mystack(pppType,0021)              enabled
   set mystack(pppType,0021,decode)       IPv4_decode
   set mystack(pppType,0021,print)        IPv4_print
   set mystack(pppType,0021,input)        IPv4_input
   set mystack(pppType,0021,callbacks)    IPv4_input


   # -- add default callbacks for ICMP
   set mystack(ipProtocol,01)           enabled
   set mystack(ipProtocol,01,decode)    ICMP_decode
   set mystack(ipProtocol,01,print)     ICMP_print
   set mystack(ipProtocol,01,input)     ICMP_input
   set mystack(ipProtocol,01,callbacks) ICMP_input

   # -- add default callbacks for UDP
   set mystack(ipProtocol,11)           enabled
   set mystack(ipProtocol,11,decode)    UDP_decode
   set mystack(ipProtocol,11,print)     UDP_print
   set mystack(ipProtocol,11,input)     UDP_input
   set mystack(ipProtocol,11,callbacks) UDP_input

   # -- add default callbacks for TCP
   set mystack(ipProtocol,06)           enabled
   set mystack(ipProtocol,06,decode)    TCP_decode
   set mystack(ipProtocol,06,print)     TCP_print
   set mystack(ipProtocol,06,input)     TCP_input
   set mystack(ipProtocol,06,callbacks) TCP_input

   # -- add default callbacks for OSPF
   set mystack(ipProtocol,59)           enabled
   set mystack(ipProtocol,59,decode)    OSPF_decode
   set mystack(ipProtocol,59,print)     OSPF_print
   set mystack(ipProtocol,59,input)     OSPF_input
   set mystack(ipProtocol,59,callbacks) OSPF_input

   # -- add default callbacks for VRRP
   set mystack(ipProtocol,70)           enabled
   set mystack(ipProtocol,70,decode)    VRRP_decode
   set mystack(ipProtocol,70,print)     VRRP_print
   set mystack(ipProtocol,70,input)     VRRP_input
   set mystack(ipProtocol,70,callbacks) VRRP_input


   # -- add default callbacks for SSDP
   set mystack(udpPort,076c)             enabled
   set mystack(udpPort,076c,decode)      SSDP_decode
   set mystack(udpPort,076c,print)       SSDP_print
   set mystack(udpPort,076c,input)       SSDP_input
   set mystack(udpPort,076c,callbacks)   SSDP_input

   # -- add default callbacks for IKE
   set mystack(udpPort,01f4)             enabled
   set mystack(udpPort,01f4,decode)      IKE_decode
   set mystack(udpPort,01f4,print)       IKE_print
   set mystack(udpPort,01f4,input)       IKE_input
   set mystack(udpPort,01f4,callbacks)   IKE_input

   # -- add default callbacks for DHCP port 67 and 68

   set mystack(udpPort,0044)             enabled
   set mystack(udpPort,0044,decode)      DHCP_decode
   set mystack(udpPort,0044,print)       DHCP_print
   set mystack(udpPort,0044,input)       DHCP_input
   set mystack(udpPort,0044,callbacks)   DHCP_input

   set mystack(udpPort,0043)             enabled
   set mystack(udpPort,0043,decode)      DHCP_decode
   set mystack(udpPort,0043,print)       DHCP_print
   set mystack(udpPort,0043,input)       DHCP_input
   set mystack(udpPort,0043,callbacks)   DHCP_input

   # -- add default callbacks for RIP
   set mystack(udpPort,0208)             enabled
   set mystack(udpPort,0208,decode)      RIP_decode
   set mystack(udpPort,0208,print)       RIP_print
   set mystack(udpPort,0208,input)       RIP_input
   set mystack(udpPort,0208,callbacks)   RIP_input

   set mystack(tcpPort,0015)             enabled
   set mystack(tcpPort,0015,decode)      FTP_decode
   set mystack(tcpPort,0015,print)       FTP_print
   set mystack(tcpPort,0015,input)       FTP_input
   set mystack(tcpPort,0015,callbacks)   FTP_input

   set mystack(tcpPort,0050)             enabled
   set mystack(tcpPort,0050,decode)      HTTP_decode
   set mystack(tcpPort,0050,print)       HTTP_print
   set mystack(tcpPort,0050,input)       HTTP_input
   set mystack(tcpPort,0050,callbacks)   HTTP_input

   set mystack(udpPort,0035)             enabled
   set mystack(udpPort,0035,decode)      DNS_decode
   set mystack(udpPort,0035,print)       DNS_print
   set mystack(udpPort,0035,input)       DNS_input
   set mystack(udpPort,0035,callbacks)   DNS_input

   # -- init default protocols
   ARP_init    mystack
   IPv4_init   mystack

   # -- do not init VRRP by default
   # VRRP_init   mystack

   PPPoE_init  mystack
   ICMP_init   mystack
   UDP_init    mystack
   TCP_init    mystack
   RIP_init    mystack
   DHCP_init   mystack
   SSDP_init   mystack
   DNS_init    mystack

   # -- init PPP related protocols
   PPP_init      mystack
   PPP_LCP_init  mystack
   PPP_NCP_init  mystack
   PPP_PAP_init  mystack
   PPP_CHAP_init mystack

   # - do not automatically initialize OSPF
   # OSPF_init  mystack

   # -- we need to build the right upvar command that will find the
   # right stack instance. The callback is evaluated at global scope

   set stackState "upvar #$level $stack s_$mystack(name)"
   set mystack(getStack) "upvar #$level $stack s_$mystack(name)"

   # -- bind the name to use for future stack calls since name can change
   set mystack(getStackName) "s_$mystack(name)"

   # -- set up a callback that will send the packet, stack, and timestamp
   #
   # %P -> packet
   # %S -> timestamp.seconds
   # %U -> timestamp.microseconds
   #

   # -- skip if vnode since vnodes do not have a real input handler

   if { "$vnode" != "1" } {
      $port read "
         $stackState
         Ethernet_input %P s_$mystack(name) %S %U
      "
   }

   # -- misc config
   set mystack(rfc1541MinimumSize)    on

}

#
# Clean up a stack that is nolonger in use

proc Stack_delete { stack } {

   upvar $stack s

   # -- close the port
   $s(port) close

   # -- unset all the state
   unset s
}

#
# Stack_name - set a name for the stack

proc Stack_name { stack name } {
   upvar $stack mystack
   set mystack(name) $name
}

proc Stack_save_packet { stack protocol variable } {
  upvar $stack s

  if [ catch {
  # -- build the variable name
  set var ${s(_varname)}_${s(_level)}_$protocol

  # -- set the current value for this protocol to the new variable
  set s(current,$protocol) $var

  # -- remember the level at which the stack was created
  set s(current,$protocol,level) $s(_level)

  } reason ] {
    error \
     "Stack_save_packet failed with protocol $protocol, variable $variable, reason $reason"
  }

  # -- return an upvar command that will scope the variable properly  
  return "upvar #$s(_level) $var $variable"


}

#
# Set the mac address for the stack

proc Stack_set_mac { stack mac } {
   upvar $stack mystack

   if { "$mystack(_vnode)" != "1" } {

     pktsrcInfo mystack "Changing MAC address for stack '$mystack(name)' on [ $mystack(port) device ] to [ hextomac $mac ]"
     set mystack(mac) $mac

     # -- reinitialize arp
     ARP_init mystack

   } else {

     error "Stack_set_mac: can not set MAC address on vnode stack"

   }

}

#
# Stack_add_callback
#
# This proc is used to add new callbacks for protocols
#
# The user must supply the protocol and protoInfo. protoInfo is
# protocol dependent. It may be the IP protocol number, udp port
# number, etc.
#

proc Stack_add_callback { stack protocol protoInfo callback } {
   upvar $stack mystack

   set mystack($protocol,$protoInfo) on

   set foundit 0

   # -- only add the callback to the list if its not already there
   # duplicate callbacks are not allowed

   if [ info exists mystack($protocol,$protoInfo,callbacks) ] {
      foreach call $mystack($protocol,$protoInfo,callbacks) {
          if { $call == "$callback" } {
              set foundit 1
              continue
          }
      }
   } 

   if { $foundit == 0 } {
     lappend mystack($protocol,$protoInfo,callbacks) $callback   
   }
}

proc Stack_remove_callback { stack protocol protoInfo callback } {
   upvar $stack mystack

   set newList ""

   # -- I don't think I should unset this
   # unset mystack($protocol,$protoInfo)

   # rebuild the callback list, but skip this callback

   if [ info exists mystack($protocol,$protoInfo,callbacks) ] {
      foreach call $mystack($protocol,$protoInfo,callbacks) {
          if { $call == "$callback" } {
              continue
          }
          lappend newList $call
      }
   } else {
      error "Stack_remove_callback: no callbacks exist for $protocol $protoInfo"
   }
   
   # -- set the new callback list
   set mystack($protocol,$protoInfo,callbacks) $newList   
}

#
# Stack_add_print_callback
#
# This proc is used to add new print callbacks for protocols
#
# The user must supply the protocol and protoInfo. protoInfo is
# protocol dependent. It may be the IP protocol number, udp port
# number, etc. Only one print callback can be specified.
#

proc Stack_add_print_callback { stack protocol protoInfo callback } {
   upvar $stack mystack

   set mystack($protocol,$protoInfo) on

   set foundit 0

   # -- add the print callback
   set mystack($protocol,$protoInfo,print) $callback
}

proc Stack_remove_print_callback { stack protocol protoInfo callback } {
   upvar $stack mystack

   # -- add the print callback
   unset mystack($protocol,$protoInfo,print)
}

proc Stack_add_vnode { stack vnode } {
   upvar $stack s
   upvar $vnode v

   # -- mark this stack with a vnode for the IPv4 address of the vnode
   set s(vnode,$v(ip),stack) $v(getStack)
   set s(vnode,$v(ip),stackname) $v(getStackName)

   Vnode_add_gateway v s
}

proc Stack_remove_vnode { stack vnode } {
   upvar $stack s
   upvar $vnode v

   # -- unmark this stack with a vnode 
   unset s(vnode,$v(ip),stack)
   unset s(vnode,$v(ip),stackname)

}






