/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  http://qacafe.com
 *
 *  $Id: tclPktSrc.c,v 1.36 2002/12/10 12:36:45 joe Exp $
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/errno.h>
#include <stdarg.h>
#include <syslog.h>
#include <signal.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <string.h>

/* For O_NONBLOCK +++joem should we use the tcl lib for this */
#include <fcntl.h>

#include "tclPktSrc.h"

/* For md5 */
#include "mhash_md5.h"

/*
 * pktsrc version
 */
static char *pktsrcVersion = "1.3.1";


/*
 * Forward declarations.  _ANSI_ARGS_ is a macro that checks to
 * see if we are being compiled with an ANSI C compiler before
 * it puts the parameter types into the declaration.
 */

int PktSrcCmd( ClientData cdata, Tcl_Interp *interp, int argc, char *argv[]);
static int PktSrcInstanceCmd _ANSI_ARGS_((ClientData clientData,
			    Tcl_Interp *, int argc, char **argv));

static PktSrc * PktSrcCreateSession();
static int PktSrcClose(PktSrc *pktsrc);
static void PktSrcDeleteSession(PktSrc *pktsrc);
static int PktSrcStartSession(Tcl_Interp *interp, char *config,
   ClientData object, int argc, char **argv);

static void PktSrcDestroyProc(char *memPtr);
static int PktSrcCheckSum( Tcl_Interp *interp, char *data );
static int PktSrcFletcherCheckSum( Tcl_Interp *interp, char *data, char *offset);
static int PktSrcMatchPacket( PktSrc *session, char *packet, char *filter);
static void session_hex_dump(PktSrc *session, char *packet , int length, struct timeval *);
 
#include "pcap.h"

/* From libpcap source, needed for stats */
#include "pcap-int.h"

/* low level pktsrc functions */
int pktsrc_open( pcap_t **pd, char *device, char *errMsg );



/*
 *
 * Pktsrc_Init - This is the init routine required for a package.
 * Create the pktsrc command and provide the package name.
 */
int 
Pktsrc_Init(Tcl_Interp *interp) 
{

  if(Tcl_InitStubs(interp, "8.0", 0 ) == NULL) {
    return TCL_ERROR;
  }

  Tcl_CreateCommand(interp, "pktsrc", PktSrcCmd,
       (ClientData)0, (Tcl_CmdDeleteProc *)NULL);


  Tcl_PkgProvide(interp, "pktsrc", PKTSRC_VERSION );

  return TCL_OK;
}

int
PktSrcCmd(clientData, interp, argc, argv)
    ClientData clientData;	/* Main window associated with
				 * interpreter. */
    Tcl_Interp *interp;		/* Current interpreter. */
    int argc;			/* Number of arguments. */
    char **argv;		/* Argument strings. */
{
    PktSrc *pktsrcPtr;
    int code;

    if (argc < 2) {
       Tcl_AppendResult(interp, "wrong # args:  should be \"",
	    argv[0], " pathName ?options?\"", (char *) NULL);
       return TCL_ERROR;
    }


    if (strcmp(argv[1], "version") == 0) {

        /* return the current pktsrc version */
        Tcl_AppendResult(interp, pktsrcVersion, 0);
        return TCL_OK;

    } else if (strcmp(argv[1], "open") == 0) {
            
        /*
         * Allocate and initialize the widget record.
         */

        pktsrcPtr = PktSrcCreateSession();
        pktsrcPtr->interp = interp;
        code = PktSrcStartSession(interp, 0, pktsrcPtr, argc, argv);

        if (code != TCL_OK) {
           PktSrcDeleteSession(pktsrcPtr);
           return TCL_ERROR;
        }       

       /*
        * Create a Tcl command that operates on the widget.
        */

       Tcl_CreateCommand(interp, pktsrcPtr->name,
   	   PktSrcInstanceCmd, (ClientData) pktsrcPtr, (void (*)()) NULL);

       Tcl_SetResult(interp, pktsrcPtr->name, TCL_STATIC);
       return TCL_OK;

    } else if (strcmp(argv[1], "checksum") == 0) {

      /*
       * Check the argument count
       */

       if ( argc != 3 ) {            
           Tcl_AppendResult(interp, "bad option \"", argv[1],
                     "\": should be pktsrc checksum data",
                     "", (char *) NULL);
 
           return TCL_ERROR;
       } else {
           return PktSrcCheckSum( interp, argv[2] );
       }
    } else if (strcmp(argv[1], "fletcher") == 0) {

      /*
       * Check the argument count
       */

       if ( argc != 4 ) {            
           Tcl_AppendResult(interp, "bad arguments \"", argv[1],
                     "\": should be pktsrc fletcher <data> <offset>",
                     "", (char *) NULL);
 
           return TCL_ERROR;
       } else {
           return PktSrcFletcherCheckSum( interp, argv[2], argv[3] );
       }

    } else if (strcmp(argv[1], "md5") == 0) {

      /*
       * Check the argument count
       */

       if ( argc == 4 ) {   

  	   /* with key */         
           return PktSrcMd5( interp, argv[2], argv[3] );
       } else if ( argc == 3 ) {
	   /* with no key */
           return PktSrcMd5( interp, argv[2], "" );
       } else {            
           Tcl_AppendResult(interp, "bad arguments \"", argv[1],
                     "\": should be pktsrc mkd5 <data> <key>",
                     "", (char *) NULL);
 
           return TCL_ERROR;
       }
 
    } else if (strcmp(argv[1], "text2ascii") == 0) {

      /*
       * Check the argument count
       */

       if ( argc != 3 ) {            
           Tcl_AppendResult(interp, "bad arguments \"", argv[1],
                     "\": should be pktsrc text2ascii <data>",
                     "", (char *) NULL);
 
           return TCL_ERROR;
       } else {
           return PktSrcText2Ascii( interp, argv[2] );
       }

    } else if (strcmp(argv[1], "ascii2text") == 0) {

      /*
       * Check the argument count
       */

       if ( argc != 3 ) {            
           Tcl_AppendResult(interp, "bad arguments \"", argv[1],
                     "\": should be pktsrc ascii2text <data>",
                     "", (char *) NULL);
 
           return TCL_ERROR;
       } else {
           return PktSrcAscii2Text( interp, argv[2] );
       }

    } else if (strcmp(argv[1], "list") == 0) {

      /*
       * Check args
       */

       if ( argc != 2 ) {            
           Tcl_AppendResult(interp, "bad option \"", argv[1],
                     "\": should be pktsrc list",
                     "", (char *) NULL);
 
           return TCL_ERROR;
       } else {
           return pktsrc_show_interfaces(interp);
       }
    } 

    Tcl_AppendResult(interp, "bad option \"", argv[1],
                     "\": should be version, list, open, checksum, fletcher, md5, text2ascii, ascii2text",
                     "", (char *) NULL);
 
    return TCL_ERROR;

}

static int
PktSrcInstanceCmd( ClientData clientData, Tcl_Interp * interp, int argc, char *argv[] )
{
	PktSrc *pktsrcPtr = (PktSrc *)clientData;
	int result = TCL_OK;
	char c;
        char buf[16];
	int length;

	if (argc < 2) {
		Tcl_AppendResult(interp, "wrong # args: should be \"",
			argv[0], " option ?arg arg ...?\"", (char *) NULL);
		return TCL_ERROR;
	}

	c = argv[1][0];
	length = strlen(argv[1]);


	if ( ((c == 'c') && (strncmp(argv[1], "close", length) == 0)) &&
          (length >= 3)) {
      
	  if ( argc == 2 ) {
             result = PktSrcClose(pktsrcPtr);
          } else {
             Tcl_AppendResult(interp, "wrong # args: should be \"",
               argv[0],"\"", (char *) NULL);
             return TCL_ERROR;
          }
	} else if ( ((c == 's') && (strncmp(argv[1], "send", length) == 0)) &&
          (length >= 2)) {
      
            if ( argc == 3 ) {
                result = PktSrcSend( pktsrcPtr, argv[2]);
            } else if ( argc == 4 ) {
                result = PktSrcSend_pad( pktsrcPtr, argv[2], argv[3]);
            } else{
                Tcl_AppendResult(interp, "wrong # args: should be \"",
                    argv[0], " send data\"", (char *) NULL);
                    return TCL_ERROR;
            }
	} 
        else if ( ((c == 's') && (strncmp(argv[1], "send-ip", length) == 0)) &&
          (length >= 2)) {
      
            if ( argc == 3 ) {
                result = PktSrcSendIp( pktsrcPtr, argv[2]);
            } else if ( argc == 4 ) {
                result = PktSrcSendIp_pad( pktsrcPtr, argv[2], argv[3]);
            } else{
                Tcl_AppendResult(interp, "wrong # args: should be \"",
                    argv[0], " send-ip data\"", (char *) NULL);
                    return TCL_ERROR;
            }
        } else if ( ((c == 'f') && (strncmp(argv[1],"flow", length) == 0 )) &&
		    (length >= 2 )) {

	  if ( argc == 5 ) {
	    result = PktSrcSendFlow( pktsrcPtr, argv[2], argv[3], argv[4] );
	  } else {
                Tcl_AppendResult(interp, "wrong # args: should be \"",
                    argv[0], " flow packet size count\"", (char *) NULL);
                    return TCL_ERROR;
          }
        } else if ( ((c == 'r') && (strncmp(argv[1], "read", length) == 0)) &&
          (length >= 2)) {
      
            if ( argc == 2 ) {

                result = PktSrcRead( pktsrcPtr);

            } else if ( argc == 3 ) {
	      /*
               * Use a read callback
               */
	      result = PktSrcReadCallback( pktsrcPtr, argv[2] );

            } else {
                Tcl_AppendResult(interp, "wrong # args: should be \"",
                    argv[0], " read\"", (char *) NULL);
                    return TCL_ERROR;
            }
        } else if ( ((c == 's') && (strncmp(argv[1], "sink", length) == 0)) && (length >= 2)) {
      
            if ( argc == 2 ) {
                result = PktSrcSetSink( pktsrcPtr);
            } else {
                Tcl_AppendResult(interp, "wrong # args: should be \"",
                    argv[0], " sink\"", (char *) NULL);
                    return TCL_ERROR;
            }
	}
 
        else if ( ((c == 's') && (strncmp(argv[1], "send-tcp",length) == 0)) &&
          (length >= 2)) {
      
            if ( argc == 3 ) {
                result = PktSrcSendTcp_Udp( pktsrcPtr, argv[2], 50);
            } else{
                Tcl_AppendResult(interp, "wrong # args: should be \"",
                    argv[0], " send-tcp data\"", (char *) NULL);
                    return TCL_ERROR;
            }
        }

	else if ( ((c == 's') && (strncmp(argv[1], "send-udp",length) == 0)) &&
		  (length >= 2)) {
      
            if ( argc == 3 ) {
                result = PktSrcSendTcp_Udp( pktsrcPtr, argv[2], 40);
            } else{
                Tcl_AppendResult(interp, "wrong # args: should be \"",
                    argv[0], " send-tcp data\"", (char *) NULL);
                    return TCL_ERROR;
            }
        }

	else if ((c == 'd') && (strncmp(argv[1], "device", length) == 0)
	    && (length >= 4)) {
	    Tcl_AppendResult(interp, pktsrcPtr->device, (char *) NULL);
	    return TCL_OK;
        }

	else if ((c == 'h') && (strncmp(argv[1], "hwaddr", length) == 0)
	    && (length >= 4)) {

            return pktsrc_show_hwaddr(interp, pktsrcPtr->device);

        } else if ((c == 'd') && (strncmp(argv[1], "dump", length ) == 0 )
 	    && (length >= 4)) {
	  if ( argc == 3 ) {
            if (strcmp(argv[2],"on") == 0) {
                pktsrcPtr->dump = 1;
            } else if (strcmp(argv[2],"off") == 0) {
                pktsrcPtr->dump = 0;
            }
 
	    /* +++joem I should return at least some Tcl string */
            return TCL_OK;

          } else {
            return TCL_ERROR;
          }

        } else if ((c == 's') && (strncmp(argv[1], "stats", length) == 0)
	    && (length >= 4)) {
	    result = PktSrcGetStats( pktsrcPtr );
	} else if ((c == 'c') && (strncmp(argv[1], "close", length) == 0)
			&& (length >= 4)) {

            /* close the device
            sprintf(buf,"%d", pktsrcPtr->reset );
            */ 
	    Tcl_AppendResult(interp, buf, (char *) NULL);
	    return TCL_OK;

	    /* Filter */
        } else if ((c == 'f') && (strncmp(argv[1], "filter", length) == 0)
	    && (length >= 4)) {

	    /* Set the packet filter */
	  if (argc == 3 ) {
	      result = PktSrcSetFilter( pktsrcPtr, argv[2]);
          } else if ( argc == 2 ) {
              Tcl_AppendResult(interp, pktsrcPtr->filter, (char *) NULL);
              return TCL_OK;
	  } else {
	   Tcl_AppendResult(interp, 
              "Unknown filter command options", (char *) NULL);
              return TCL_ERROR;
          }
	    /* Match a packet */
        } else if ((c == 'm') && (strncmp(argv[1], "match", length) == 0)
	    && (length >= 4)) {

	    /* Try to match this packet */
	  if (argc == 4 ) {
	      result = PktSrcMatchPacket( pktsrcPtr, argv[2], argv[3]);
	  } else {
	   Tcl_AppendResult(interp, 
              "Unknown 'match' command options", (char *) NULL);
              return TCL_ERROR;
          }
        } else {

	  /* 
           * Unknown command
           */

	   Tcl_AppendResult(interp, 
              "Unknown instance command '", argv[1],"'", (char *) NULL);
           result = TCL_ERROR;
        }

	return result;
}



PktSrc *
PktSrcCreateSession()
{
    PktSrc *session;
    static int id = 0;

    session = (PktSrc *) ckalloc(sizeof(PktSrc));
    memset((char *) session, 0, sizeof(PktSrc));

    sprintf(session->name, "pktsrc%d", id++);

    /* mark the connection as down */
    strcpy(session->connState, "down");

    session->packetsOut    = 0;
    session->packetsIn     = 0;
    session->lastOutSize   = 0;

    session->dump = 0;
    return session;
}

void
PktSrcDeleteSession(PktSrc *session)
{
   if ( ! session  ) return;

   /* if session list then clean up */
   PktSrcDestroyProc( (ClientData) session );
}

static void
PktSrcDestroyProc(char *memPtr)
{  
    PktSrc *session = (PktSrc *) memPtr;

    /* free dynamic memory */
    if ( session->device ) ckfree(session->device);
    if ( session->filter ) ckfree(session->filter);

    ckfree((char *) session);
}   

int PktSrcClose( PktSrc *session ) 
{

    /* Delete any callbacks associated with this session */ 
    Tcl_DeleteFileHandler( session->pd->fd );

    /* close the pcap connection */
    pcap_close(session->pd);

    /* now delete this command */
    Tcl_DeleteCommand(session->interp, session->name );

    return TCL_OK;
}

int 
PktSrcStartSession(Tcl_Interp *interp, char *config, 
     ClientData object, int argc, char **argv)
{
    PktSrc *session = (PktSrc *) object;

    if ((argc + 1 ) % 2) {
        Tcl_AppendResult(interp, "wrong # of args: should be \"", argv[0],
          " ", argv[1], " hostname ?option value? ?option value? ...\"",
                         (char *) NULL);
        return TCL_ERROR;
    }

    return PktSrcInitConnection( session, argv[2] );
}

int 
PktSrcInitConnection( PktSrc *session, char *device )
{
    char errMsg[256];

    /* -- check device name */

    if (session->device) {
       ckfree(session->device);
    } 

    /* should be using some ckalloc version of strdup */
    session->device = strdup(device);

    if ( strcmp(session->connState, "up") == 0 ) {
        /* -- I'm already up, is this allowed ? */
        return TCL_OK;
    };
    
    if ( pktsrc_open( &(session->pd), session->device, errMsg ) < 0 ) {
       Tcl_AppendResult(session->interp, "can not open interface \"", 
          session->device, "\" , ", errMsg, (char *) NULL);
       return TCL_ERROR;
    }

    /* +++joem HACK */
    session->fd = session->pd->fd;

    /* mark the connection up */
    strcpy( session->connState, "up" );

    return TCL_OK;
}


int 
PktSrcCheckSum( Tcl_Interp *interp, char *data )
{
   /* convert data from ascii to hex */
   int size = 0;
   char *packet;
   char buf[16];
   unsigned short new_chksum;
   unsigned short chksum;

   size = strlen(data);

   if ( size % 2 != 0 ) {
       sprintf( buf, "%d", size);
       Tcl_AppendResult(interp, 
         "PktSrcCheckSum: ascii packet size is not on an even byte ", (buf), (char *) NULL);
       return TCL_ERROR;
   } else if ( size > 3028 ) {
       sprintf( buf, "%d", size/2);

       Tcl_AppendResult(interp, "data size too big \"", 
          buf, "\"", (char *) NULL);
       return TCL_ERROR;

   } else if ( size < 2 ) {
       sprintf( buf, "%d", size/2);

       Tcl_AppendResult(interp, "data size too small \"", 
          buf, "\"", (char *) NULL);
       return TCL_ERROR;
   } else {

       if ( packet = ascii_to_hex_packet(interp, data, &size) ) {
          new_chksum = htons((unsigned short)ipcksum(packet, size ));
          ckfree( packet);
          sprintf( buf, "%4.4x", new_chksum);
	  Tcl_AppendResult(interp, buf, (char *) NULL);
          return TCL_OK;
       } else {
          return TCL_ERROR;
       }
   } 
}

int 
PktSrcFletcherCheckSum( Tcl_Interp *interp, char *data, char *offset )
{
   /* convert data from ascii to hex */
   int size = 0;
   char *packet;
   char buf[16];
   unsigned short new_chksum;
   unsigned short chksum;
   int off;

   size = strlen(data);

   /* convert offset to an interger */
   off = atoi(offset);

   if ( size % 2 != 0 ) {
       sprintf( buf, "%d", size);
       Tcl_AppendResult(interp, 
         "PktSrcFletcherCheckSum: ascii packet size is not on an even byte ", (buf), (char *) NULL);
       return TCL_ERROR;
   } else if ( size > 3028 ) {
       sprintf( buf, "%d", size/2);

       Tcl_AppendResult(interp, "data size too big \"", 
          buf, "\"", (char *) NULL);
       return TCL_ERROR;

   } else if ( size < 2 ) {
       sprintf( buf, "%d", size/2);

       Tcl_AppendResult(interp, "data size too small \"", 
          buf, "\"", (char *) NULL);
       return TCL_ERROR;
   } else {

       if ( packet = ascii_to_hex_packet(interp, data, &size) ) {

	 /* compute a fletcher cksum */

          new_chksum = fletcherCksum(packet, size, off );
          ckfree( packet);
          sprintf( buf, "%4.4x", new_chksum);
	  Tcl_AppendResult(interp, buf, (char *) NULL);
          return TCL_OK;
       } else {
          return TCL_ERROR;
       }
   } 
}

int 
PktSrcMd5( Tcl_Interp *interp, char *data, char *key )
{
   /* convert data from ascii to hex */
   int size = 0, keylen = 0;
   char *ascii = 0;
   void *context = 0;
   void *return_val = 0;
   char *packet = 0;
   char buf[16];   

   /* get the length of the data and the key */
   size = strlen(data);
   keylen = strlen(key);

   if ( size % 2 != 0 ) {
       sprintf( buf, "%d", size);
       Tcl_AppendResult(interp, 
         "ascii data size is not on an even byte ", (buf), (char *) NULL);
       return TCL_ERROR;
   } else if ( size > 3028 ) {
       sprintf( buf, "%d", size/2);

       Tcl_AppendResult(interp, "data size too big \"", 
          buf, "\"", (char *) NULL);
       return TCL_ERROR;

   } else if ( size < 2 ) {
       sprintf( buf, "%d", size/2);

       Tcl_AppendResult(interp, "data size too small \"", 
          buf, "\"", (char *) NULL);
       return TCL_ERROR;
   } else {

       if ( packet = ascii_to_hex_packet(interp, data, &size) ) {

	  /* allocate and initialize a context */
          context = (void *)malloc(sizeof(MD5_CTX));
          MD5Init((void *) context);

          /* start the key */
          MD5Update((void *) context, key, keylen);

          /* update with data */
          MD5Update((void *) context, packet, size);

          /* finish */
          return_val = MD5Final((void *) context);

          /* convert the digest to an ascii string */
          ascii = PktSrc_makeAscii( (char *)return_val, 16);

          /* return string as tcl string */
          Tcl_AppendResult(interp, ascii, (char *) NULL); 

          /* clean up */
          ckfree(ascii);
          ckfree(return_val);
          ckfree(context);
          ckfree(packet);

          return TCL_OK;

       } else {
          return TCL_ERROR;
       }
    } 
}

int 
PktSrcText2Ascii( Tcl_Interp *interp, char *data )
{
   /* convert data from text to ascii */
   int size = 0;
   char *ascii = 0;

   size = strlen(data);

   ascii = PktSrc_makeAscii( data, size);

   Tcl_AppendResult(interp, ascii, (char *) NULL);

   ckfree(ascii);

   return TCL_OK;
}

int 
PktSrcAscii2Text( Tcl_Interp *interp, char *data )
{
   /* convert ascii to text (hex) */
   int size = 0;
   char *text = 0;

   if (text = ascii_to_hex_packet(interp, data, &size)) {
      Tcl_AppendResult(interp, text, (char *) NULL);
      ckfree(text);
      return TCL_OK;
   } else { 
      return TCL_ERROR;
   }     
}


int 
PktSrcSend( PktSrc *session, char *data )
{
   /* convert data from ascii to hex */
   int size = 0;
   char *packet;
   char buf[16];

   size = strlen(data);

   if ( size % 2 != 0 ) {
       sprintf( buf, "%d", size);
       Tcl_AppendResult(session->interp, 
         "PktSrcSend: ascii packet size is not on an even byte ", (buf), (char *) NULL);
       return TCL_ERROR;
   } else if ( size > 3028 ) {
       sprintf( buf, "%d", size/2);

       Tcl_AppendResult(session->interp, "packet size too big \"", 
          buf, "\"", (char *) NULL);
       return TCL_ERROR;

   } else if ( size < 28 ) {
       sprintf( buf, "%d", size/2);

       Tcl_AppendResult(session->interp, "packet size too small \"", 
          buf, "\"", (char *) NULL);
       return TCL_ERROR;
   } else {
     if ( packet = ascii_to_hex_packet(session->interp, data, &size) ) {
       if (pktsrc_send_packet( session->dump,session->pd, packet, size ) < size ) {
          ckfree( packet);
          session->writeErrors++;
          session->lastOutSize = size;
          Tcl_AppendResult(session->interp, 
             "error sending packet on ",session->device ,": ", strerror(errno), 
             (char *) NULL);
          return TCL_ERROR;
       } else {
          ckfree( packet);
          session->packetsOut++;
          session->lastOutSize = size;
          sprintf( buf, "%d", size);
          Tcl_AppendResult(session->interp, buf, (char *) NULL);
          return TCL_OK;
       }
     } else {
       return TCL_ERROR;
     } 
  }
}

int 
PktSrcSendIp( PktSrc *session, char *data )
{
   /* convert data from ascii to hex */
   int size = 0, offset = 0;
   unsigned char *packet;
   char buf[16];
   short *new_chksum;
   short *new_size;
   short chksum;

   size = strlen(data);

   if ( size % 2 != 0 ) {
       sprintf( buf, "%d", size);
       Tcl_AppendResult(session->interp, 
         "PktSrcSendIp: ascii packet size is not on an even byte ", (buf), (char *) NULL);
       return TCL_ERROR;
   } else if ( size > 3028 ) {
       sprintf( buf, "%d", size/2);

       Tcl_AppendResult(session->interp, "packet size too big \"", 
          buf, "\"", (char *) NULL);
       return TCL_ERROR;

   } else if ( size < 68 ) {
       sprintf( buf, "%d", size/2);

       Tcl_AppendResult(session->interp, "ip packet size too small \"", 
          buf, "\"", (char *) NULL);
       return TCL_ERROR;

   } else {
     if ( packet = ascii_to_hex_packet(session->interp, data, &size) ) {

       /* handle the offset for 802.1q */
       if ( packet[12] == 0x81 && packet[13] == 0 ) {
          offset = 4;
       }
      
       /* set the correct length in the IP header */
       new_size =(unsigned short *)(packet + 16 + offset);

       /* subtract out eth mac for new IP length */
       *new_size = htons( size - 14);

       /* compute new IP checksum */
  
       packet[24 + offset ] = 0;
       packet[25 + offset ] = 0;

       new_chksum = (unsigned short *)(packet + 24 + offset );
       *new_chksum = ipcksum(packet + 14 + offset , 20 );

 
       if (pktsrc_send_packet( session->dump,session->pd, packet, size ) < size ) {
          ckfree( packet);
          session->writeErrors++;
          session->lastOutSize = size;
          Tcl_AppendResult(session->interp, strerror(errno), (char *) NULL);
          return TCL_ERROR;
       } else {
          ckfree( packet);
          session->packetsOut++;
          session->lastOutSize = size;
          sprintf( buf, "%d", size);
          Tcl_AppendResult(session->interp, buf, (char *) NULL);
          return TCL_OK;
       }
     } else {
       return TCL_ERROR;
     } 
  }
}

int 
PktSrcSend_pad( PktSrc *session, char *data, char *length )
{
   /* convert data from ascii to hex */
   int size = 0, len = 0;
   char *packet;
   int ret;

   len = atoi(length);
   size = strlen(data);

   if ( len <= 0 || len > 1518 ) {
       Tcl_AppendResult(session->interp, "invalid pad length ", length ,
         " (1-1518)", (char *) NULL);
       return TCL_ERROR;
   } else {
       packet = (unsigned char *)ckalloc(1518 * 2 + 1);
       memset (packet, 'a', len * 2 + 1 );
       memcpy( packet, data, size);
       packet[len * 2] = 0;
       ret = PktSrcSend( session, packet );
       ckfree( packet);
       return ret;
   }
}

int 
PktSrcSendIp_pad( PktSrc *session, char *data, char *length )
{
   /* convert data from ascii to hex */
   int size = 0, len = 0;
   char *packet;
   int ret;

   len = atoi(length);
   size = strlen(data);

   if ( len <= 0 || len > 1518 ) {
       Tcl_AppendResult(session->interp, "invalid pad length ", length ,
         " (1-1518)", (char *) NULL);
       return TCL_ERROR;
   } else {
       packet = (unsigned char *)ckalloc(1518 * 2 + 1);
       memset (packet, 'a', len * 2 + 1 );
       memcpy( packet, data, size);
       packet[len * 2] = 0;
       ret = PktSrcSendIp( session, packet );
       ckfree( packet);
       return ret;
   }
}


int 
PktSrcSendTcp_Udp( PktSrc *session, char *data, int checksum_offset )
{
   /* convert data from ascii to hex */
   int size = 0, offset = 0;
   unsigned char *packet;
   char buf[16];
   unsigned short *s;
   unsigned short *new_chksum;
   unsigned short *new_size;
   unsigned short chksum;
   unsigned short *tcp_new_chksum;
   unsigned short tcp_chksum;

   size = strlen(data);

   if ( size % 2 != 0 ) {
       sprintf( buf, "%d", size);
       Tcl_AppendResult(session->interp, 
         "PktSrcSendTcp_Udp: ascii packet size is not on an even byte ", (buf), (char *) NULL);
       return TCL_ERROR;
   } else if ( size > 3028 ) {
       sprintf( buf, "%d", size/2);

       Tcl_AppendResult(session->interp, "packet size too big \"", 
          buf, "\"", (char *) NULL);
       return TCL_ERROR;

   } else if ( size < 68 ) {
       sprintf( buf, "%d", size/2);

       Tcl_AppendResult(session->interp, "ip packet size too small \"", 
          buf, "\"", (char *) NULL);
       return TCL_ERROR;
   } else {
     if ( packet = ascii_to_hex_packet(session->interp, data, &size) ) {

       /* handle the offset for 802.1q */
       if ( packet[12] == 0x81 && packet[13] == 0 ) {
          offset = 4;
       }

       /* set the correct length in the IP header */
       new_size =(unsigned short *)(packet + 16 + offset );

       /* subtract out eth mac for new IP length */
       *new_size = htons( size - 14);

       /* compute IP checksum */
       packet[24 + offset ] = 0;
       packet[25 + offset ] = 0;
       new_chksum = (unsigned short *)(packet + 24 + offset );
       *new_chksum = ipcksum(packet + 14 + offset , 20 );

       /* compute TCP/UDP checksum */
       packet[checksum_offset + offset ] = 0;
       packet[++checksum_offset + offset ] = 0;
       new_chksum = (unsigned short *)(packet + 50 + offset);
       *new_chksum = tcp_udp_cksum(packet + offset, (size - 34)); /* just the size of tcp is needed (eth:14/ip:20) */

       if (pktsrc_send_packet( session->dump,session->pd, packet, size ) < size ) {
          ckfree( packet);
          session->writeErrors++;
          session->lastOutSize = size;
          Tcl_AppendResult(session->interp, strerror(errno), (char *) NULL);
          return TCL_ERROR;
       } else {
          ckfree( packet);
          session->packetsOut++;
          session->lastOutSize = size;
          sprintf( buf, "%d", size);
          Tcl_AppendResult(session->interp, buf, (char *) NULL);
          return TCL_OK;
       }
     } else {
       return TCL_ERROR;
     } 
  }
}


int 
PktSrcSendFlow( PktSrc *session, char *data, char *length, char *count )
{
   /* convert data from ascii to hex */
   int size = 0, duration = 0, i = 0, pad = 0;
   char *packet;
   char *newdata;
   char buf[16];
   short *new_chksum;
   short *new_size;
   short chksum;

   duration = atoi(count);
   pad = atoi(length);
   size = strlen(data);


   if ( pad <= 0 || pad > 1518 ) {
       Tcl_AppendResult(session->interp, "invalid pad length ", length ,
         " (1-1518)", (char *) NULL);
       return TCL_ERROR;
   } else if ( size % 2 != 0 ) {
       sprintf( buf, "%d", size);
       Tcl_AppendResult(session->interp, 
         "PktSrcSendFlow: ascii packet size is not on an even byte ", (buf), (char *) NULL);
       return TCL_ERROR;
   } else if ( size > 3036 ) {
       sprintf( buf, "%d", size/2);

       Tcl_AppendResult(session->interp, "packet size too big \"", 
          buf, "\"", (char *) NULL);
       return TCL_ERROR;

   } else if ( size < 68 ) {
       sprintf( buf, "%d", size/2);

       Tcl_AppendResult(session->interp, "ip packet size too small \"", 
          buf, "\"", (char *) NULL);
       return TCL_ERROR;
   } else {

     /* pad data */
     newdata = (unsigned char *)ckalloc(1518 * 2 + 1);
     memset (newdata, 'a', pad * 2 + 1 );
     memcpy( newdata, data, size);
     newdata[pad * 2] = 0;
     packet = ascii_to_hex_packet(session->interp, newdata, &size);

     /* I'm all done with the new data */
     ckfree(newdata);

     if (packet) {

       /* set the correct length in the IP header */
       new_size =(unsigned short *)(packet+16);

       /* subtract out eth mac for new IP length */
       *new_size = htons( size - 14);

       /* compute IP checksum */

       packet[24] = 0;
       packet[25] = 0;

       new_chksum = (unsigned short *)(packet+24);
       *new_chksum = ipcksum(packet + 14, 20 );

 
       for ( i = 0; i < duration; i++ ) {

         if (pktsrc_send_packet( session->dump, session->pd, packet, size ) < size ) {
            ckfree( packet);
            session->writeErrors++;
            session->lastOutSize = size;
            Tcl_AppendResult(session->interp, strerror(errno), (char *) NULL);
            return TCL_ERROR;
          } else {
            /* ckfree( packet); */
            session->packetsOut++;
          }
        }
        ckfree( packet);
        session->lastOutSize = size;
        sprintf( buf, "%d", size);
        Tcl_AppendResult(session->interp, buf, (char *) NULL);
        return TCL_OK;
       
     } else {
       return TCL_ERROR;
     } 
  }
}



int 
PktSrcRead( PktSrc *session )
{
   /* convert data from ascii to hex */
   int length = 0, status = 0;
   char *packet = 0, *ascii = 0;
   struct timeval timestamp;
   char buf[16];


   /* +++joem This will block until a packet that matches the filter
    is rceived , note pktsrc_read blocks anyways */

   while ( status == 0 ) {

     if ((status=pktsrc_read( session->pd, 0, &packet, &length , &timestamp)) < 0 ) {
        Tcl_AppendResult(session->interp, "read error on ", session->device, " ", session->pd->errbuf, (char *) NULL);
        return TCL_ERROR;
     } else if ( status == 0 ) {
        /* packet did not match the filter */
        continue;
     } else {

       /* dump */
       if( session->dump ) {
          session_hex_dump ( session, packet, length, &timestamp );
       }

       /* Got packet and length */ 
       /* +++ bite the bullet and convert to ascii */
       ascii = PktSrc_makeAscii( packet, length);
       Tcl_AppendResult(session->interp, ascii, (char *) NULL);
       ckfree( ascii);

       return TCL_OK;
     }
   }
}



void 
PktSrc_fileevent( ClientData clientData, int mask)
{
   int status = 0, length = 0;
   char *packet = 0, *ascii = 0;
   PktSrc *session = (PktSrc *)clientData;
   struct timeval timestamp;
   char buf[20];
   int	code;
   Tcl_DString tclCmd;
   char *startPtr, *scanPtr, *name;

   if (mask & TCL_READABLE) {


      /* Check for a packet that matches the filter */
      if ((status=pktsrc_read( session->pd, 0, &packet, &length, &timestamp )) < 0 ) {

         /* +++joem is this the correct handling of a background error ? */
         Tcl_AppendResult(session->interp, "read error on ", session->device, " ", session->pd->errbuf, (char *) NULL);
         Tcl_BackgroundError( session->interp );

         return;

      } else if ( status == 0 ) {

         /* packet did not match the filter - noop */

      } else {

         /* Got packet and length */ 
         /* dump */
         if( session->dump ) {
            session_hex_dump ( session, packet, length, &timestamp );
         }

         /* +++ bite the bullet and convert to ascii */
         ascii = PktSrc_makeAscii( packet, length);

         /* subsitutue %P for the packet in the callback */
         Tcl_DStringInit(&tclCmd);

         startPtr = session->read_script;
         for (scanPtr = startPtr; *scanPtr != '\0'; scanPtr++) {
	    if (*scanPtr != '%') {
	       continue;
	    }
	    Tcl_DStringAppend(&tclCmd, startPtr, scanPtr - startPtr);
	    scanPtr++;
	    startPtr = scanPtr + 1;
	    switch (*scanPtr) {
	       case 'P':  
	          Tcl_DStringAppend(&tclCmd, ascii, -1);
	          break;
	       case 'S':  
	          sprintf(buf, "%u", timestamp.tv_sec);
	          Tcl_DStringAppend(&tclCmd, buf, -1);
	          break;
	       case 'U':  
	          sprintf(buf, "%u", timestamp.tv_usec);
	          Tcl_DStringAppend(&tclCmd, buf, -1);
	          break;
               default:
	          sprintf(buf, "%%%c", *scanPtr);
	          Tcl_DStringAppend(&tclCmd, buf, -1);
                  break;
	    }
         }
         Tcl_DStringAppend(&tclCmd, startPtr, scanPtr - startPtr);

         /* done with substitution */

         /* call user callback */
         Tcl_AllowExceptions(session->interp);
         code = Tcl_Eval(session->interp, Tcl_DStringValue(&tclCmd));
         Tcl_DStringFree(&tclCmd);

         if ( code != TCL_OK ) {
            /* handle bgerror */
	    char *errorMsg = strdup(session->interp->result);
	    Tcl_AddErrorInfo(session->interp, "\n    (read callback)");
	    Tcl_BackgroundError(session->interp);
	    Tcl_SetResult(session->interp, errorMsg, TCL_DYNAMIC);
         }
         ckfree( ascii);
      }      
   }

   /* that's it */
}

void PktSrc_sinkevent( ClientData clientData, int mask)
{
   PktSrc *session = (PktSrc *)clientData;
   int length = 0;
   char *packet = 0;
   struct timeval timestamp;

   if (mask & TCL_READABLE) {
      /* call user callback */
     if (pktsrc_read( session->pd, 0, &packet, &length, &timestamp ) < 0 ) {
         Tcl_AppendResult(session->interp, "read error on ", session->device, " ", session->pd->errbuf, (char *) NULL);
         Tcl_BackgroundError( session->interp );   
     } else {
        /* dump */
        if( session->dump ) {
           session_hex_dump ( session, packet, length, &timestamp );
        }
     }
   }
   /* that's it */
}

int 
PktSrcReadCallback( PktSrc *session , char *script )
{
   /* convert data from ascii to hex */
   int size = 0, length = 0;
   char *packet;
   char buf[16];

   /* make sure the socket is non blocking */
   if ( fcntl( session->pd->fd, F_SETFL, O_NONBLOCK) == -1 ) {
      Tcl_AppendResult(session->interp, strerror(errno), (char *) NULL);
      return TCL_ERROR;
   }     

   if (session->read_script) {
     ckfree(session->read_script);
   }

   length = strlen(script) + 1;
   session->read_script = ckalloc(length);
   memcpy( session->read_script, script, length);

   /* Set up a file handler for the connection
    */

   /* Delete old */
   Tcl_DeleteFileHandler( session->pd->fd );

   /* Add new */
   Tcl_CreateFileHandler( session->pd->fd , TCL_READABLE | TCL_EXCEPTION , PktSrc_fileevent, session);

   Tcl_AppendResult(session->interp, "1", (char *) NULL);
   return TCL_OK;
}

int 
PktSrcSetSink( PktSrc *session)
{
   /* convert data from ascii to hex */
   int size = 0, length = 0;
   char buf[16];

   /* make sure the socket is non blocking */
   if ( fcntl( session->pd->fd, F_SETFL, O_NONBLOCK) == -1 ) {
      Tcl_AppendResult(session->interp, strerror(errno), (char *) NULL);
      return TCL_ERROR;
   }     

   /* Set up a file handler for the connection
    */

   /* Delete old */
   Tcl_DeleteFileHandler( session->pd->fd );

   /* Add new */
   Tcl_CreateFileHandler( session->pd->fd , TCL_READABLE | TCL_EXCEPTION , PktSrc_sinkevent, session);

   Tcl_AppendResult(session->interp, "1", (char *) NULL);
   return TCL_OK;
}


char * 
PktSrc_makeAscii( unsigned char *packet, int length)
{
   int i = 0;
   unsigned char *ptr = 0;

   /* +++joem This can't be very fast. Need better story */

   unsigned char *ascii = ptr = ckalloc( length * 2 + 1);
   for ( i = 0; i < length; i++, ptr+=2, packet++ ) {
      sprintf(ptr, "%.2x", *packet);
   }
   *ptr = 0;

   return ascii;
} 


int 
PktSrcSetFilter( PktSrc *session , char *filter )
{

   bpf_u_int32 netmask = 0;
   int Oflag=1;
   struct bpf_program fcode;

   if (pcap_compile(session->pd, &fcode, filter, Oflag, netmask) < 0) {
       Tcl_AppendResult(session->interp, 
        "filter compile error '", filter,"': ", pcap_geterr(session->pd), 0);
       return TCL_ERROR;
   }

   if (pcap_setfilter(session->pd, &fcode) < 0) {
       Tcl_AppendResult(session->interp, 
        "filter set error '", filter,"': ", pcap_geterr(session->pd), 0);
       return TCL_ERROR;
   }

   /* The filter is okay, rember it */
   if (session->filter) {
     ckfree(session->filter);
   }

   session->filter = strdup(filter);

   return TCL_OK;
} 

int PktSrcMatchPacket( PktSrc *session, char *packet, char *filter)
{
   bpf_u_int32 netmask = 0;
   int Oflag=1 , size = 0;
   struct bpf_program fcode;
   char *packethex = 0;
   char buf[16];

   /* try to compile the filter */
   if (pcap_compile(session->pd, &fcode, filter, Oflag, netmask) < 0) {
       Tcl_AppendResult(session->interp, 
        "filter compile error '", filter,"': ", pcap_geterr(session->pd), 0);
       return TCL_ERROR;
   }

   size = strlen(packet);

   if ( size % 2 != 0 ) {
       sprintf( buf, "%d", size);
       Tcl_AppendResult(session->interp, 
         "PktSrcMatchPacket: ascii packet size is not on an even byte ", (buf), (char *) NULL);
       return TCL_ERROR;
   } else if ( size > 3028 ) {
       sprintf( buf, "%d", size/2);

       Tcl_AppendResult(session->interp, "data size too big \"", 
          buf, "\"", (char *) NULL);
       return TCL_ERROR;

   } else if ( size < 2 ) {
       sprintf( buf, "%d", size/2);

       Tcl_AppendResult(session->interp, "data size too small \"", 
          buf, "\"", (char *) NULL);
       return TCL_ERROR;
   } else {

       if ( packethex = ascii_to_hex_packet(session->interp, packet, &size) ) {

         /* call bpf to match the packet */
         if (bpf_filter(fcode.bf_insns, packethex, size, size)) {

           /* found match return 1 */
           Tcl_AppendResult(session->interp, "1", 0);

         } else {

           /* no match return 0 */
           Tcl_AppendResult(session->interp, "0", 0);
         }

          ckfree( packethex );
          return TCL_OK;
       } else {
          return TCL_ERROR;
       }
   } 

   /* shouldn't reach this ... */
   return TCL_ERROR;
}

int 
PktSrcGetStats( PktSrc *session ) 
{
  struct pcap_stat data;
  char temp[16];

  if ( get_stats( session->pd, &data) == -1 ) {
       Tcl_AppendResult(session->interp, 
        "stats error:", pcap_geterr(session->pd), 0);
       return TCL_ERROR;
  }

  /* build a list with receives and drops */

  /* printf("recv = %d\n", data.ps_recv); */
  sprintf(temp, "%u", data.ps_recv );
  Tcl_AppendElement( session->interp, temp);
  sprintf(temp, "%u", data.ps_drop );
  Tcl_AppendElement( session->interp, temp);

  return TCL_OK;

}

char * ascii_to_hex_packet( Tcl_Interp *interp, unsigned char *ascii, int *length ) 
{
   unsigned char *pos;
   int inLen = 0;
   char *ret = 0;
   char buffer[16];
   char buffer2[16];
   char convert[3];
   unsigned char *buf = pos = 0;
   unsigned int x = 0;

   *length = 0;

   inLen = strlen(ascii);
   inLen = inLen/2 + 1;

   /* dynamically allocate the return buffer */
   buf = pos = (unsigned char *)ckalloc(inLen);
   memset( buf, 0, inLen);

   while ( *ascii ) {

      /* -- check for hex digit */
      if (! isxdigit(*ascii)) {
         sprintf(buffer,"%d", x );
         sprintf(buffer2,"%c", *ascii );
         Tcl_AppendResult(interp, "Non hex character '", buffer2,
           "' at position ", buffer, (char *) NULL);
         ckfree(buf);
         return ret;
      }

      /* -- convert in pairs */
      if ( x % 2 ) {
         convert[1] = *ascii;
         convert[2] = 0;
         *pos = (unsigned char)strtol( convert, 0, 16 );
         pos++;
         (*length)++;
      } else {
         convert[0] = *ascii;
      }  

      /* -- next byte */
      x++;
      ascii++; 
   }

   /* -- check alignment */
   if ( x % 2 != 0 ) {
      Tcl_AppendResult(interp, "Data is not at an even byte boundary",
         (char *) NULL);
      ckfree(buf);
      return ret;
   } else {
      ret = buf;
   }

   return ret;
}

void
timestamp_print(const struct timeval *tvp)
{
    int s;

    /* Default */
    s = (tvp->tv_sec) % 86400;
    printf("%02d:%02d:%02d.%06u ",
		    s / 3600, (s % 3600) / 60, s % 60, (u_int32_t)tvp->tv_usec);
}

void
session_hex_dump(PktSrc *session, char *packet , int length, struct timeval * timestamp)
{
    printf("\n<< ");
    timestamp_print(timestamp);
    printf("Received packet on %s, length %d:\n", session->pd->md.device, length);
    hex_dump( packet, length);
    printf("\n");
    fflush(stdout);
}











