#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# 2002 QAcafe http://qacafe.com
#
# $Id: tcp.tcl,v 1.20 2002/12/10 22:29:11 joe Exp $
#
#
# Poor man's TCP
#
#
# ----------------------------------------------------------------------------

# -- this is part of the pktsrc package
package provide pktsrc 1.3

#
# TCP_input
#
# Main input proc for TCP
#
#   Decode packet
#   Handle print requests
#   Save on stack storage
#   Drive callbacks
#   

proc TCP_input { packet stack { options none }} {
   upvar $stack s

   if { $options == "print" } {

      TCP_decode $packet tcp
      TCP_print tcp

   } else {

      # -- remember this packet in the 'tcp' variable. This allows other
      # callbacks to get access to the decodes at any Tcl stack level

      eval [ Stack_save_packet s TCP tcp ]
   
      # -- now decode the packet into the 'tcp' variable
      TCP_decode $packet tcp

      set tcp(computed-checksum) [ TCP_compute_checksum s ]

      if { [ string compare "$tcp(computed-checksum)" "0000" ] } {
         pktsrcWarn s "Received invalid TCP checksum"         
      }
   }



   # -- check for callbacks at the TCP level
   if { [ info exists s(tcpPort,$tcp(dstPort)) ] && $options != "print" } {

      # -- drive any input callbacks
      foreach c $s(tcpPort,$tcp(dstPort),callbacks) {
          $c $tcp(data) s $options
      }

   } elseif { $options != "print" } {

      # -- no tcp session or service at this port

      if { "$s(TCP,reset)" == "on" } {

        # -- need ip info to reset
        set ipState $s(current,IPv4)
        upvar #$s(current,IPv4,level) $ipState ip

        if { "$ip(dst)" == "$s(ip)" } {
           # -- reset this connection
           pktsrcInfo s \
            "Reseting connection to [ hextoip $ip(dst)] (my ip [hextoip $s(ip) ])" 
           TCP_reset s $ip(dst) $tcp(dstPort) $ip(src) $tcp(srcPort) 
        }
     }
   } else {

      # -- drive callbacks using src port to allow packets from services
      # to get decoded

      if { "$options" == "print" } {

         # -- drive any input callbacks
         if [ info exists s(tcpPort,$tcp(srcPort)) ] {
           foreach c $s(tcpPort,$tcp(srcPort),callbacks) {
              $c $tcp(data) s $options
           }          
         } elseif [ info exists s(tcpPort,$tcp(dstPort)) ] {
           foreach c $s(tcpPort,$tcp(dstPort),callbacks) {
              $c $tcp(data) s $options
           } 
         }
      }
   } 
 


   # -- drive any srcPort callbacks, mainly for printing/capture tools
   if [ info exists s(tcpSrcPort,$tcp(srcPort)) ] {
 
        # -- drive any input callbacks
        foreach c $s(tcpSrcPort,$tcp(srcPort),callbacks) {
           $c $tcp(data) s $options
        }
   }

}

proc TCP_session_data { stack { options none }} {
   upvar $stack s

   # -- get the current tcp contents
   set tcpState $s(current,TCP)
   upvar #$s(current,TCP,level) $tcpState tcp

   if [ info exists s(tcpSessionPort,$tcp(dstPort)) ] {

      # -- drive any input callbacks
      foreach c $s(tcpSessionPort,$tcp(dstPort),callbacks) {

          pktsrcInfo s "passing tcp data to '$c'"
          $c $tcp(data) s $options
      }
   }
}

#
# TCP_init 
#
# set up any deaults for TCP
#

proc TCP_init { stack } {
   upvar $stack s

   # -- set the inital src TCP port
   # make it some what unique based on the time so that tools can
   # appear to be using different srcPorts when run back to back

   set s(TCP,nextPort) [expr ([clock seconds ] % 5000) + 10000]

   # -- only support a very small window for now since we don't handle
   # multiple packets in the window

   set s(TCP,window) 1400

   set s(TCP,ignore) off  

   set s(TCP,reset) on

   set s(TCP,client,connections) 0

   set s(TCP,server,connections) 0

   set s(TCP,canned-sequence) ""
}

#
# TCP_decode
#
# Decode a tcp packet into an array
#

proc TCP_decode { packet tcp } {
   upvar $tcp out

   # -- parse out all the fields we find interesting ...

   # --
   #
   # TCP header
   # rfc 793
   #
   # --

   set out(srcPort)     [ string range $packet 0 3 ]
   set out(dstPort)     [ string range $packet 4 7 ]

   set out(sequence)    [ string range $packet 8 15 ]
   set out(ack)         [ string range $packet 16 23 ]

   set out(offset)      [ string range $packet 24 25 ]
   set out(flags)       [ string range $packet 26 27 ]
   set out(window)      [ string range $packet 28 31 ]
   set out(checksum)    [ string range $packet 32 35 ]
   set out(upointer)    [ string range $packet 36 39 ]

   set out(options)     ""

   # -- check for options, data offset
   set dataStart [ expr ((0x$out(offset) & 0xf0 ) >> 2 ) ]
   set optionEnd [ expr $dataStart * 2 - 1 ]
 
   if { $dataStart > 20 } {
      # -- found options
      set out(options) [ string range $packet 40 $optionEnd ]
   }   

   set out(data)        [ string range $packet [ expr $dataStart * 2 ] end ]

}

#
# TCP_print
#
# Display the tcp packet
#

proc TCP_print { tcp } {
   upvar $tcp out

   puts "TCP: ---- TCP Packet ----"
   puts "TCP: "
   puts "TCP: Source Port                = [expr 0x$out(srcPort)] [TCP_port [expr 0x$out(srcPort)]]"
   puts "TCP: Destination Port           = [expr 0x$out(dstPort)] [TCP_port [expr 0x$out(dstPort)]]"

   puts "TCP: Sequence Number            = 0x$out(sequence)"
   puts "TCP: Ack Number                 = 0x$out(ack)"
   puts "TCP: Offset                     = 0x$out(offset)"

   puts "TCP: Flags                      = 0x$out(flags)"
   puts "TCP:   [TCP_get_flags $out(flags)]"
   puts "TCP: Window                     = [expr 0x$out(window)]"

   set cksum ""
   if [ info exists out(computed-checksum) ] {
       if { [ string compare "$out(computed-checksum)" "0000" ] } {
         set cksum "(Invalid Checksum)"
       }
   }
   puts "TCP: Checksum                   = 0x$out(checksum) $cksum"
   puts "TCP: Urgent Pointer             = 0x$out(upointer)"
   puts "TCP:"

   if { "$out(options)" != "" } {
      puts "TCP: Tcp Options                = 0x$out(options)"     
   }
}

#
# TCP_get_flags
#
# Build a tcl string to display TCP flags
#

proc TCP_get_flags { flags } {
 
   
   set fin   [ expr 0x$flags & 0x01 ]
   set syn   [ expr (0x$flags & 0x02) >> 1 ]
   set rst   [ expr (0x$flags & 0x04) >> 2 ]
   set push  [ expr (0x$flags & 0x08) >> 3 ]
   set ack   [ expr (0x$flags & 0x10) >> 4 ]
   set urg   [ expr (0x$flags & 0x20) >> 5 ]

   set flagStr ""
 
   if { $syn } {
       set flagStr "${flagStr}+Syn"
   }
   if { $ack } {
       set flagStr "${flagStr}+Ack"
   }   
   if { $rst } {
       set flagStr "${flagStr}+Rst"
   }
   if { $fin } {
       set flagStr "${flagStr}+Fin"
   }
   if { $push } {
       set flagStr "${flagStr}+Push"
   }
   if { $urg } {
       set flagStr "${flagStr}+Urg"
   }
 
   if { "$flagStr" == "" } {
       set flagStr "flags-0x$flags"
   }

   return $flagStr
}

#
# Display well known TCP ports
#
# port is the decimal value of a well known port

proc TCP_port { port } {
   if { $port == "23" } {
      set str "(Telnet)"
   } elseif { $port == "21" } {
      set str "(Ftp)"
   } elseif { $port == "80" } {
      set str "(Http)"
   } else {
      set str ""
   }
}

#
# TCP_init_session 
#
proc TCP_init_session { stack session } {
   upvar $stack s
   upvar $session sess

   set sess(type)  tcp
   set sess(srcPort) 0000
   set sess(dstPort) 0000

   set sess(dstIp) 00000000
   set sess(srcIp) $s(ip)

   # -- use the default window size for this stack
   set sess(window) $s(TCP,window)   

   # -- pick an initial sequence number
   set sess(sequence) [ TCP_new_sequence s ]
   set sess(last-sent) $sess(sequence)
   set sess(last-ack) $sess(sequence)
   
   set sess(ack) 00000000

   # -- no options sent by default
   set sess(offset) 50

   set sess(flags) 0
   set sess(upointer) 0 
   set sess(state) create
   set sess(ignore) off

   set sess(bytes-in) 0
   set sess(bytes-out) 0

   set sess(data) ""
   

   if { "$s(TCP,ignore)" == "on" } {
     set sess(ignore) on
   }
}

#
# TCP_open (client)
#
# Open a TCP connection (i.e client)
#

proc TCP_open { stack session dstIp dstPort { timeOut 5 } { srcPort -1 } } {
   upvar $stack s
   upvar $session sess

   TCP_init_session s sess

   # -- bump the client connections count
   incr s(TCP,client,connections)

   # -- remember the local port
   if { "$srcPort" == "-1" } {
      set sess(srcPort) [ format "%4.4x" [ TCP_get_client_port s ]]
   } else {
      set sess(srcPort) [ format "%4.4x" $srcPort ]
   }

   set sess(dstPort) [ format "%4.4x" $dstPort ]

   set sess(dstIp) [ iptohex $dstIp ]

   pktsrcInfo s "opening new tcp session [hextoip $s(ip)]:[expr 0x$sess(srcPort)] -> $dstIp:$dstPort"
 
   # -- add a callback for this session
   Stack_add_callback s tcpPort $sess(srcPort) TCP_session_input

   # --- hmmm ....
   if { [ expr 0x$sess(dstPort) ] == 80 } {
       # -- add callback for 80
       Stack_add_callback s tcpPort $sess(srcPort) HTTP_input
   }

   TCP_add_session s sess

   # -- kick off the tcp connection
   TCP_send_state s sess 02 syn-sent

   # -- block until open/closed/timeout
   set run 1
   set elapsed 0
   set timeOut [ expr $timeOut * 10 ]

   while { $run } {
      set state [ TCP_state s sess ]

      if { "$state" == "established" || "$state" == "closed" || $timeOut == "0" } {
         break
      }

      # -- this is wait 1/10 second and process events
      Event_wait 100

      # -- check for a timeout
      incr elapsed 1
      if { $elapsed >= $timeOut } {
        break
      }
   }

   return $state

}

proc TCP_restart { stack session { timeOut 5 } } {
   upvar $stack s
   upvar $session sess

   # -- retransmit the last TCP segment
   TCP_retransmit_last s sess 

   set run 1
   set elapsed 0
   set timeOut [ expr $timeOut * 10 ]

   while { $run } {
      set state [ TCP_state s sess ]

      if { "$state" == "established" || "$state" == "closed" } {
         break
      }

      # -- this is wait 1/10 second and process events
      Event_wait 100

      # -- check for a timeout
      incr elapsed 1
      if { $elapsed >= $timeOut } {
        break
      }
   }

   return $state

}

#
# TCP_accept (server)
#
# Accept a TCP connection 
#

proc TCP_accept { stack session dstIp dstPort srcPort } {
   upvar $stack s
   upvar $session sess

   TCP_init_session s sess

   # -- bump the server connections count
   incr s(TCP,server,connections)

   # -- remember the local port
   set sess(srcPort) $srcPort
   set sess(dstPort) $dstPort
   set sess(dstIp) $dstIp


   TCP_add_session s sess

   # -- kick off the tcp connection
   TCP_set_state s sess listen
}

#
# TCP_server
#
# Listen for TCP connection on port
#

proc TCP_server { stack port { sessionCallback "" }} {
   upvar $stack s

   # -- get the src port in hex
   set srcPort [ format "%4.4x" $port ]

   # -- add a callback for this server
   Stack_add_callback s tcpPort $srcPort TCP_server_input

   if { "$sessionCallback" != "" } {
     Stack_add_callback s tcpSessionPort $srcPort $sessionCallback
   }

   pktsrcInfo s "Accepting tcp connections on port $port"
}

#
# TCP_server_shutdown
#
# Shutdown down tcp server on port
#

proc TCP_server_shutdown { stack port { sessionCallback "" }} {

   upvar $stack s

   # -- get the src port in hex
   set srcPort [ format "%4.4x" $port ]

   # -- add a callback for this server
   Stack_remove_callback s tcpPort $srcPort TCP_server_input

   if { "$sessionCallback" != "" } {
     Stack_remove_callback s tcpSessionPort $srcPort $sessionCallback
   }

   pktsrcInfo s "tcp server on port $port has shutdown"
}

#
# TCP_write
#
# Send data on a TCP connection
# 
# data is hex data
#

proc TCP_write { stack session data } {
   upvar $stack s
   upvar $session sess

   set bytes 0

   set sid [ TCP_get_session s isess $sess(dstIp) $sess(dstPort) $sess(srcPort) ]

   if { "$sid" != "0" } {

      set bytes [ TCP_send_data s isess $data ]   
      TCP_save_session s isess
      return $bytes

   } else {
      error "TCP_write: No such tcp session"
   }

}

#
# TCP_state
#
# return the state of a tcp session
#

proc TCP_state { stack session } {

   upvar $stack s
   upvar $session sess

   set sid [ TCP_get_session s isess $sess(dstIp) $sess(dstPort) $sess(srcPort) ]

   if { "$sid" != "0" } {

      return $isess(state)   

   } else {

      return "no such session"
   }

}

#
# TCP_add_session
#
# Add a TCP connection to the stack
#

proc TCP_add_session { stack session } {
   upvar $stack s
   upvar $session sess

   set key [ TCP_session_key sess ]
   set s(TCP,session,$key,state) create

   # -- I need to save a session using the same trick as save_packet

   TCP_save_session s sess
}

proc TCP_update_session { stack session } {
   upvar $stack s
   upvar $session sess

   set sid [ TCP_get_session s sess $sess(dstIp) $sess(dstPort) $sess(srcPort) ]


}

#
# TCP_session_key
#
# Build a key used to look up TCP sessions
#

proc TCP_session_key { session } {
   upvar $session sess

   return "$sess(dstIp),$sess(srcIp),$sess(dstPort),$sess(srcPort)"
}

#
# TCP_save_session
#
# Save the contents of the session in stack storage
#

proc TCP_save_session { stack session } {
   upvar $stack s
   upvar $session sess

   set key [ TCP_session_key sess ]

   # -- remember all the fields
   set s(TCP,session,$key,fields) "[array names sess]"

   # -- set the key for lookups

   set s(TCP,session,$key,key) $key
   # -- store a copy on the stack
   foreach field [ array names sess ] {

      # -- don't pass the data around
      if { "$field" == "data" } {
          continue
      }
      set s(TCP,session,$key,$field) $sess($field)
   }

}

proc TCP_buffer_data { stack session data } {
   upvar $stack s
   upvar $session sess

   set key [ TCP_session_key sess ]

   set currentData ""

   if [ info exists s(TCP,session,$key,data) ] {
       set currentData $s(TCP,session,$key,data)
   }

   set s(TCP,session,$key,data) "$currentData$data"

}

proc TCP_get_data { stack session } {
   upvar $stack s
   upvar $session sess

   set key [ TCP_session_key sess ]

   set currentData ""

   if [ info exists s(TCP,session,$key,data) ] {
      set currentData $s(TCP,session,$key,data)
      set s(TCP,session,$key,data) ""
   }

   return $currentData
}

#
# TCP_remove_session
#
# Remove a TCP session from the stack storage
#

proc TCP_remove_session { stack session } {
   upvar $stack s
   upvar $session sess

   set key [ TCP_session_key sess ]

   # -- remember all the fields
   set fields $s(TCP,session,$key,fields)
   unset s(TCP,session,$key,fields)

   # -- store a copy on the stack
   foreach field $fields {
      if { "$field" == "data" } {
         continue
      }
      unset s(TCP,session,$key,$field)
   }

   # -- remove callbacks ?

}

#
# TCP_get_session
#
# Find a session, return its contents in the 'session' variable
#

proc TCP_get_session { stack session dstIp dstPort srcPort } {
   upvar $stack s
   upvar $session sess

   set findsess(dstIp) $dstIp
   set findsess(dstPort) $dstPort
   set findsess(srcPort) $srcPort
   set findsess(srcIp) $s(ip)

   set key [ TCP_session_key findsess ]

   # -- check for session
   if [ info exists s(TCP,session,$key,fields) ] {

      foreach field $s(TCP,session,$key,fields) {

         # -- don't pass the data around
         if { "$field" == "data" } {
             continue
         }
         set sess($field) $s(TCP,session,$key,$field)
      }

      return $sess(srcPort)

   } else {
      return 0
   }
}

#
# TCP_set_state
#
# Set the state of a tcp connection
# Save the contents of the session
# If the session is now 'closed', remove it (Timed Wait not implemented)
#

proc TCP_set_state { stack session state } {
   upvar $stack s
   upvar $session sess

   # -- ignore reset connections
   if { "$state" == "reset" } {
       return
   }

   set sess(state) $state
   TCP_save_session s sess

   # -- removed closed sessions, I should be waiting before doing this
   if { "$state" == "closed" } {

     # pktsrcInfo s "removing tcp session (sent $sess(bytes-out) bytes, recv $sess(bytes-in) bytes)"
     # TCP_remove_session s sess
   }
}

#
# TCP_session_input
#
# input process for TCP packets
#
#  - find existing session for this tcp connection
#  - hand off packet and session if found
#  - else, ignore packet

proc TCP_session_input { packet stack {options none}} {
   upvar $stack s

   if { $options == "print" } {
      return
   }

   # -- get the current IP contents
   set ipState $s(current,IPv4)
   upvar #$s(current,IPv4,level) $ipState ip

   # -- get the current tcp contents
   set tcpState $s(current,TCP)
   upvar #$s(current,TCP,level) $tcpState tcp
   
   pktsrcInfo s "[hextoip $ip(dst)]:[expr 0x$tcp(dstPort)] received [ TCP_get_flags $tcp(flags) ] from [hextoip $ip(src)]:[expr 0x$tcp(srcPort)], seq 0x$tcp(sequence), ack 0x$tcp(ack)"

   # -- check that this is a valid session
   set sid [ TCP_get_session s session $ip(src) $tcp(srcPort) $tcp(dstPort) ]

   if { "$sid" != "0" } {

      pktsrcInfo s \
         "found tcp session at port [expr 0x$tcp(dstPort)] state '$session(state)'"

      # -- process the tcp session state machine
      TCP_process_session s session
      TCP_save_session s session

   } else {
      pktsrcInfo s "no existing TCP session"
   }

}


#
# TCP_server_input
#
# input process for TCP server ports
#
#  - find existing session for this tcp connection
#    else if syn, create session
#    else rst connection
#  - hand off packet to session process

proc TCP_server_input { packet stack {options none}} {
   upvar $stack s

   if { $options == "print" } {
      return
   }

   # -- get the current IP contents
   set ipState $s(current,IPv4)
   upvar #$s(current,IPv4,level) $ipState ip

   # -- get the current tcp contents
   set tcpState $s(current,TCP)
   upvar #$s(current,TCP,level) $tcpState tcp
   
   pktsrcInfo s "[hextoip $ip(dst)]:[expr 0x$tcp(dstPort)] received [ TCP_get_flags $tcp(flags) ] from [hextoip $ip(src)]:[expr 0x$tcp(srcPort)], seq 0x$tcp(sequence), ack 0x$tcp(ack)"

   # -- check that this is a valid session
   set sid [ TCP_get_session s session $ip(src) $tcp(srcPort) $tcp(dstPort) ]

   if { "$sid" != "0" } {

      pktsrcInfo s \
         "found tcp session at port [expr 0x$tcp(srcPort)] state '$session(state)'"

      # -- process the tcp session state machine
      TCP_process_session s session

   } else {

      # -- if this is a syn, establish a new session
      if { [ expr 0x$tcp(flags) & 0x02 ] != 0 } {

         pktsrcInfo s "accepting connection from [hextoip $ip(src)]:[expr 0x$tcp(srcPort)]"

         # -- accept this new session
         TCP_accept s session $ip(src) $tcp(srcPort) $tcp(dstPort)

         # -- process this session
         TCP_process_session s session

      } else {

         pktsrcInfo s \
           "reseting connection from [hextoip $ip(src)]:[expr 0x$tcp(srcPort)]"

         # -- send an rst for this session
         TCP_reset s $ip(dst) $tcp(dstPort) $ip(src) $tcp(srcPort)
      } 
   }

}

#
# TCP_process_session
#
# This is the guts of the TCP state machine
#

proc TCP_process_session { stack session } {
   upvar $stack s
   upvar $session sess

   # -- get the current IP contents
   set ipState $s(current,IPv4)
   upvar #$s(current,IPv4,level) $ipState ip

   # -- get the current tcp contents
   set tcpState $s(current,TCP)
   upvar #$s(current,TCP,level) $tcpState tcp


   # -- check ignore mode
   if { "$sess(ignore)" == "on" } {
      pktsrcInfo s "ignoring tcp packet ..."
      return
   }

   # -- check tcp checksum (TBD)

   if { "$sess(state)" == "syn-sent" } {
   
       if { "$tcp(flags)" == "02" } {
           pktsrcWarn s "Received TCP Syn after sending TCP Syn, ignoring ..."
       } elseif { [ expr 0x$tcp(flags) & 0x04 ] != 0 } {
           pktsrcWarn s "Connection was RST by peer"

           # -- close session
           TCP_set_state s sess closed
           TCP_save_session s sess

       } elseif { [ expr 0x$tcp(flags) & 0x12 ] != 0 } {

           # -- I'm expecting a syn-ack
           # -- check ack

           if { [ expr 0x$tcp(ack) - 0x$sess(last-sent) ] != 0 } {
              pktsrcWarn s \
                 "sequence number ack mismatch (seq $sess(sequence) ack $tcp(ack))"
              # ++++ joem handle error
           } else {

              # -- ack this sequence number

              set sess(ack) [ format "%8.8x" [ expr 0x$tcp(sequence) + 1 ] ]
              set sess(sequence) [ format "%8.8x" [ expr 0x$sess(last-sent) ]]

              TCP_send_state s sess 10 established
           }

       }

   } elseif { "$sess(state)" == "listen" } {
   
       # -- I'm expecting a syn-ack
       if { [ expr 0x$tcp(flags) & 0x02 ] != 0 } {

           # -- ack this sequence number

           set sess(ack) [ format "%8.8x" [ expr 0x$tcp(sequence) + 1 ] ]
           TCP_send_state s sess 12 syn-rcvd
       }

   } elseif { "$sess(state)" == "syn-rcvd" } {
   
       # -- I'm expecting a ack
       if { [ expr 0x$tcp(flags) & 0x10 ] != 0 } {

           # -- check ack
           if { [ expr 0x$tcp(ack) - 0x$sess(last-sent) ] != 0 } {
              pktsrcWarn s \
                 "sequence number ack mismatch (seq $sess(sequence) ack $tcp(ack))"
              # ++++ joem handle error

           } else {


              # -- bump my sequence number
              set sess(sequence) [ format "%8.8x" [ expr 0x$sess(last-sent) ]]

              # -- go into established state
              TCP_set_state s sess established
              TCP_save_session s sess

              # -- check for data
              set bytes [ expr [ string length $tcp(data) ] / 2 ]

              pktsrcInfo s "received $bytes bytes in ack"

              if { $bytes > 0 } {

                 # -- if there is new data, acknowledge it
                 set sess(ack) [ format "%8.8x" [ expr 0x$tcp(sequence) + $bytes ] ]
                 incr sess(bytes-in) $bytes
                 TCP_send_state s sess 10 established

                 # -- pass this data to other applications
                 TCP_buffer_data s sess $tcp(data)
              }

              # -- call session data even if there is no data. This indicates
              # that a new session was created for tcp server

              TCP_session_data s

           }
       } elseif { [ expr 0x$tcp(flags) & 0x02 ] != 0 } {

           # -- ack this sequence number, this allows for a syn retransmission
           # when I am in server mode

           TCP_retransmit_last s sess 
       }

   } elseif { "$sess(state)" == "established" } {

       # -- syn 
       if { [ expr 0x$tcp(flags) & 0x02 ] != 0 } {
           pktsrcWarn s "received unexpected syn"
           pktsrcWarn s " [ hextoip $ip(src) ]:[expr 0x$tcp(srcPort)] -> [ hextoip $ip(dst) ]:[expr 0x$tcp(dstPort)]"
           
       } elseif { [ expr 0x$tcp(flags) & 0x10 ] != 0 } {

           # -- check ack
           if { [ expr 0x$tcp(ack) - 0x$sess(last-sent) ] != 0 } {
              pktsrcWarn s \
                 "sequence number ack mismatch (seq $sess(sequence) ack $tcp(ack) last $sess(last-sent))"
              # ++++ joem handle error

           } else {

              # -- bump my sequence number
              set sess(sequence) [ format "%8.8x" [ expr 0x$sess(last-sent) ]]

              # -- I only need to ack if I received additional data
              # set newAck [ format "%8.8x" [ expr 0x$tcp(sequence) + 0 ] ]

              set bytes [ expr [ string length $tcp(data) ] / 2 ]
              pktsrcInfo s "received $bytes bytes in ack"

              set sess(ack) [ format "%8.8x" [ expr 0x$tcp(sequence) + $bytes ] ]

              if { $bytes > 0 } {

                 incr sess(bytes-in) $bytes
                 TCP_send_state s sess 10 established

                 # +++joem combine this ( have session data do this)
                 TCP_buffer_data s sess $tcp(data)
                 TCP_session_data s

              } else {
                 TCP_save_session s sess
              }
           }
       }

       if { [ expr 0x$tcp(flags) & 0x01 ] != 0 } {
           pktsrcInfo s "connection closed by peer"

           # +++joem -- this was causing the problem
           set sess(ack) [ format "%8.8x" [ expr 0x$sess(ack) + 1 ] ]

           TCP_save_session s sess

           # -- close this session down, enter last-ack
           TCP_private_close s sess last-ack

       }
           

   } elseif { "$sess(state)" == "fin-wait-1" } {

       # -- found ack+fin
       if { [ expr (0x$tcp(flags) & 0x11) - 0x11  ] == 0 } {

           # -- check ack
           if { [ expr 0x$tcp(ack) - 0x$sess(last-sent) ] != 0 } {
              pktsrcWarn s \
                 "sequence number ack mismatch (seq $sess(sequence) ack $tcp(ack))"
              # ++++ joem handle error
           }

           set sess(sequence) [ format "%8.8x" [ expr 0x$sess(last-sent) ]]
           set sess(ack) [ format "%8.8x" [ expr 0x$tcp(sequence) + 1 ] ]

           # -- ack the fin
           TCP_send_state s sess 10 fin-wait-2

       } elseif { [ expr (0x$tcp(flags) & 0x10) ] != 0 } {

           # -- check ack
           if { [ expr 0x$tcp(ack) - 0x$sess(last-sent) ] != 0 } {

              # +++ joem leave this message here for now, this is
              # not an error and will happen with the pktsrc client talking
              # to the pktsrc server

              # pktsrcWarn s "unacknowledged data outstanding"
              # pktsrcWarn s "last-sent 0x$sess(last-sent), ack 0x$tcp(ack)"

              set sess(sequence) [ format "%8.8x" [ expr 0x$sess(last-sent) ]]
              set bytes [ expr [ string length $tcp(data) ] / 2 ]
              pktsrcInfo s "received $bytes bytes in ack"

	      if { "$bytes" > "0" } {
                 pktsrcWarn s "fin-wait-1 does not yet handle new data"
              }

           } else {

              # -- out fin has been ack'd wait for the other side's fin

              TCP_set_state s sess fin-wait-2
           }
       }
   } elseif { "$sess(state)" == "fin-wait-2" } {

       if { [ expr (0x$tcp(flags) & 0x01) ] != 0 } {

           # -- check ack
           if { [ expr 0x$tcp(ack) - 0x$sess(last-sent) ] != 0 } {
              pktsrcWarn s \
                 "sequence number ack mismatch (seq $sess(sequence) ack $tcp(ack))"
              # ++++ joem handle error

           } else {

              # -- ack this sequence number

              set sess(sequence) [ format "%8.8x" [ expr 0x$sess(last-sent) ]]
              set sess(ack) [ format "%8.8x" [ expr 0x$tcp(sequence) + 1 ] ]

              # -- I could go to time-wait state, for now close
              TCP_send_state s sess 10 closed
           }
       }
   } elseif { "$sess(state)" == "last-ack" } {

       if { [ expr 0x$tcp(flags) & 0x10 ] != 0 } {

           # -- just close
           TCP_set_state s sess closed

       }
   }

}

proc TCP_private_close { stack session { state fin-wait-1 } } {
   upvar $stack s
   upvar $session sess

   if { "$sess(state)" == "established" || "$sess(state)" == "syn-sent" \
          || "$sess(state)" == "syn-rcvd" } {

      # -- start shutdown
      TCP_send_state s sess 11 $state

   } else {

      # -- shutdown already in progess, do not send fin     

   }
}

#
# TCP_close
#
# Close a TCP connection
#

proc TCP_close { stack session { state fin-wait-1 } } {
   upvar $stack s
   upvar $session sess

   # -- I should verify that we are in the 'established' state

   set sid [ TCP_get_session s isess $sess(dstIp) $sess(dstPort) $sess(srcPort) ]

   if { "$sid" != "0" } {

      if { "$isess(state)" == "established" || "$isess(state)" == "syn-sent" \
          || "$isess(state)" == "syn-rcvd" } {

         # -- start shutdown
         TCP_send_state s isess 11 $state

      } else {

         # -- shutdown already in progess, do not send fin     

      }
   }

   # -- remove any callbacks
   if { [ expr 0x$sess(dstPort) ] == 80 } {
       # -- add callback for 80
       Stack_remove_callback s tcpPort $sess(srcPort) HTTP_input
   }
}

#
# TCP_reset
#
# reset a connection
# no existing session is required
#

proc TCP_reset { stack srcIp srcPort dstIp dstPort } {
   upvar $stack s

   # -- create a pseudo session to use TCP_send_state
   TCP_init_session s session

   set session(srcPort) $srcPort
   set session(dstPort) $dstPort
   set session(dstIp) $dstIp
   set session(srcIp) $srcIp

   # -- set the rst flag
   set flags 04 
     
   # -- reset the connection 
   TCP_send_state s session $flags reset
}

#
# TCP_send_state
#
# Send a TCP control packet with no data
# After sending packet, enter state '$state'
#

proc TCP_send_state { stack session flags state } {
   upvar $stack s
   upvar $session sess

   # -- for a fin we need to move our window ahead. For normal data
   # we only support a window of 1 packet

   if { [ expr 0x$flags & 0x01 ] } {
     set sess(sequence) [ format "%8.8x" [ expr 0x$sess(last-sent) ] ]
   }

   set packet [ format "%s%s%s%s%s%s%4.4x%s" \
       $sess(srcPort) \
       $sess(dstPort) \
       $sess(sequence) \
       $sess(ack) \
       $sess(offset) \
       $flags \
       $sess(window) \
       00000000 ]

   if { "$state" != "$sess(state)" } {
     pktsrcInfo s "sending tcp [ TCP_get_flags $flags], entering state '$state', seq 0x$sess(sequence), ack 0x$sess(ack)"
   } else {
     pktsrcInfo s "sending tcp [ TCP_get_flags $flags], current state '$state', seq 0x$sess(sequence), ack 0x$sess(ack)"
   }



   # -- send the tcp packet
   TCP_send s $packet $sess(dstIp) $sess(dstPort) $sess(srcPort)

   # -- if this is a fin, bump my last-sent number
   if { [ expr 0x$flags & 0x01 ] } {
     set sess(last-sent) [ format "%8.8x" [ expr 0x$sess(last-sent) + 1 ] ]
   }

   # -- this is a syn, bump my last sent

   # +++joem Hmmm, be carefull about retransmissions
   if { [ expr 0x$flags & 0x02 ] } {
     set sess(last-sent) [ format "%8.8x" [ expr 0x$sess(last-sent) + 1 ] ]
   }

   # -- set the session state
   TCP_set_state s sess $state
}

#
# TCP_send_data
#
# Send data on a connection
#
#

proc TCP_send_data { stack session data } {
   upvar $stack s
   upvar $session sess

   set state "established"
   set flags "10"

   set dataLen [ expr [ string length $data ] / 2 ]

   if { $dataLen > 1400 } {
      error "I can't handle long writes > 1400 ... sorry"
   }

   pktsrcInfo s "tcp send data request with length $dataLen"

   # -- increment the last-sent
   # set sess(last-sent) [ format "%8.8x" [ expr 0x$sess(sequence) + $dataLen ] ]


   set packet [ format "%s%s%s%s%s%s%4.4x%s%s" \
       $sess(srcPort) \
       $sess(dstPort) \
       $sess(sequence) \
       $sess(ack) \
       $sess(offset) \
       $flags \
       $sess(window) \
       00000000 $data ]


   if { "$state" != "$sess(state)" } {
     pktsrcInfo s \
       "sending tcp [ TCP_get_flags $flags], entering state '$state', seq 0x$sess(sequence) ack 0x$sess(ack)"
   } else {
     pktsrcInfo s \
       "sending tcp [ TCP_get_flags $flags], current state '$state', seq 0x$sess(sequence) ack 0x$sess(ack)"
   }

   # -- send the tcp packet
   TCP_send s $packet $sess(dstIp) $sess(dstPort) $sess(srcPort)

   # -- increment the last-sent
   set sess(last-sent) [ format "%8.8x" [ expr 0x$sess(sequence) + $dataLen ] ]

   # set sess(sequence) $sess(last-sent)

   # -- incr bytes-out (not acknowledged)
   incr sess(bytes-out) $dataLen

   # -- set the session state
   TCP_set_state s sess $state

   return $dataLen
}


#
# TCP_retransmit_last
#
# resend the last tcp packet
#
#

proc TCP_retransmit_last { stack session } {
   upvar $stack s
   upvar $session sess

   set state $sess(state)
   set flags "10"

   set packet $s(TCP,last-packet)

   set sid [ TCP_get_session s isess $sess(dstIp) $sess(dstPort) $sess(srcPort) ]

   if { "$sid" != "0" } {

      pktsrcInfo s "retransmit last tcp segment"
      pktsrcInfo s "resending tcp seq 0x$isess(sequence) ack 0x$isess(ack)"

      # -- send the tcp packet
      TCP_send s $packet $isess(dstIp) $isess(dstPort) $isess(srcPort)

   } else {

      error "TCP_retransmit_last: No such tcp session"
   }
}


#
# TCP_get_client_port
#
# Get a free client port
#

proc TCP_get_client_port { stack } {
   upvar $stack s

   # -- make sure the port numbers wrap
   if { $s(TCP,nextPort) > 65535 } {
      set s(TCP,nextPort) 5000
   }

   set port $s(TCP,nextPort)     
   incr s(TCP,nextPort)

   return $port
}

#
# TCP_new_sequence
#
# return a sequence number in hex
#

proc TCP_new_sequence { stack } {
   upvar $stack s

   if { "$s(TCP,canned-sequence)" != "" } {
      return $s(TCP,canned-sequence)
   } else {
      return [ format "%8.8x" [ clock clicks ]]
   } 
}

proc TCP_set_sequence_number { stack number } {
   upvar $stack s

   if { "$number" == "random" } {
       set s(TCP,canned-sequence) ""
   } else {
       set s(TCP,canned-sequence) $number
   }
}

#
# TCP_bump_sequence
#
# bumps the TCP sequence number by 'count'
# handles wrapping of the TCP sequence number
#

proc TCP_bump_sequence { seqhex count } {
   return [ format "%8.8x" [ expr 0x$seqhex + $count ]]
} 

#
# TCP_send 
#
# Send a TCP packet to the IP layer
# Compute checksum before send
#

proc TCP_send { stack packet dest destPort srcPort } {
   upvar $stack s

   # -- always remeber the last packet sent
   set s(TCP,last-packet) $packet

   set length [expr [ string length $packet ] / 2 ]

   # -- compute the TCP checksum

   # -- first build the pseudo header
   set pseudo [ format "%s%s0006%4.4x" $s(ip) $dest $length ]
   
   # +++joem need to pad to even size
   set tcp [ format "%s%s" $pseudo $packet ]
   set checksum [ pktsrc checksum $tcp ]

   # -- build the real packet

   # -- replace the checksum in the packet
   set tcpCheck [ string replace $packet 32 35 $checksum ]

   # -- send this to IP using tcp as the protocol
   IPv4_send $tcpCheck s $dest 06
}

#
# TCP_clear_connections
#
# remove any state associated with existing TCP connections
#

proc TCP_clear_connections { stack } {
   upvar $stack s

   # -- reset the connection counters
   set s(TCP,client,connections) 0
   set s(TCP,server,connections) 0

   foreach keylookup [ array names s TCP,session,*,key ] {

      # -- get the key
      set key $s($keylookup)

      set fields $s(TCP,session,${key},fields)

      unset s(TCP,session,${key},fields)
      unset s(TCP,session,${key},key)

      foreach field $fields {
          if { "$field" == "data" } {
              continue
          }
	  unset s(TCP,session,${key},${field}) 
      }
   }
}

#
# TCP_server_show_connections
#
# Display all client tcp connections current connected to a port
#

proc TCP_server_show_connections { stack port } {
   upvar $stack s

   puts "\nTcp client connections at port $port for [hextoip $s(ip)]:"

   puts [ format "  %-21s %-21s %s" "src" "dst" "state" ]
   puts "  ---------------------+---------------------+------------"

   foreach keylookup [ array names s TCP,session,*,key ] {

      # -- get the key
      set key $s($keylookup)

      set srcPort [ expr 0x$s(TCP,session,$key,srcPort) ]
      if { "$srcPort" != "$port" } {
          continue
      }

      set srcIp [ hextoip $s(TCP,session,$key,srcIp) ]
      set dstIp [ hextoip $s(TCP,session,$key,dstIp) ]
      set dstPort [ expr 0x$s(TCP,session,$key,dstPort) ]
      set state $s(TCP,session,$key,state)

      puts [ format \
        "  %-21s %-21s %s" "$srcIp $srcPort" "$dstIp $dstPort" $state ]
   }

   # -- that's it
   puts ""

}

proc TCP_server_lookup { stack ip port { sess "" }} {
   upvar $stack s
   set rport 0

   if { "$sess" != "" } {
      upvar $sess isess
   }

   foreach keylookup [ array names s TCP,session,*,key ] {

      # -- get the key
      set key $s($keylookup)

      # -- just return the first connection
      set srcPort [ expr 0x$s(TCP,session,$key,srcPort) ]
      set dstPort [ expr 0x$s(TCP,session,$key,dstPort) ]
      set dstIp [ hextoip $s(TCP,session,$key,dstIp) ]

      if { "$srcPort" == "$port"  && "$dstIp" == "$ip" } {
          set rport $dstPort

          # -- load in the session info
          set sid [ TCP_get_session s isess $s(TCP,session,$key,dstIp) $s(TCP,session,$key,dstPort) $s(TCP,session,$key,srcPort) ]

          break
      }
   }

   return $rport
}

proc TCP_server_find_first { stack port { sess "" }} {
   upvar $stack s
   set rport 0

   if { "$sess" != "" } {
      upvar $sess isess
   }

   foreach keylookup [ array names s TCP,session,*,key ] {

      # -- get the key
      set key $s($keylookup)

      # -- just return the first connection
      set srcPort [ expr 0x$s(TCP,session,$key,srcPort) ]
      set dstPort [ expr 0x$s(TCP,session,$key,dstPort) ]
      set dstIp [ hextoip $s(TCP,session,$key,dstIp) ]

      if { "$srcPort" == "$port" } {
          set rport $dstPort

          # -- load in the session info
          set sid [ TCP_get_session s isess $s(TCP,session,$key,dstIp) $s(TCP,session,$key,dstPort) $s(TCP,session,$key,srcPort) ]

          break
      }
   }

   return $rport
}

proc TCP_get_stats { stack stat } {
    upvar $stack s

    if { "$stat" == "server-connections" } {
       return $s(TCP,server,connections)
    } elseif { "$stat" == "client-connections" } {
       return $s(TCP,client,connections)
    } else {
       error "TCP_get_stats: unknown stat type '$stat'"
    }
}

#
# -- TCP_ignore
#

proc TCP_ignore { stack session { mode on } } {
   upvar $stack s
   upvar $session sess

   set bytes 0

   set sid [ TCP_get_session s isess $sess(dstIp) $sess(dstPort) $sess(srcPort) ]

   if { "$sid" != "0" } {

      set isess(ignore) $mode 
      TCP_save_session s isess

   } else {
      error "TCP_ignore: No such tcp session"
   }
}

proc TCP_ignore_all { stack { mode on } } {
   upvar $stack s

   set s(TCP,ignore) $mode
}

proc TCP_reset_unknown { stack { mode on } } {
   upvar $stack s

   set s(TCP,reset) $mode
}

#
# TCP_generic_input
#
# generic callback for catching TCP packets
#

proc TCP_generic_input { packet stack { options none }} {
   upvar $stack s

   # -- trigger the packet event
   if { "$options" != "print" } {
      Event_got_packet
   }
}

#
# TCP_get_packet
#
# blocking call for receiving TCP packets
# this call blocks until the packet is received or a timeout occurs
#

proc TCP_get_packet { stack srcPort packet timeout } {
   upvar $stack s

   set ret -1

   # add a callback for TCP packets for this source port
   Stack_add_callback s tcpPort $srcPort TCP_generic_input

   pktsrcInfo s \
     "Waiting up to $timeout seconds for TCP packet at port [expr 0x$srcPort]"

   # -- compute timeout in milli seconds
   set msec [ expr $timeout * 1000 ]

   # -- now wait for the packet or timeout
   set result [ Event_wait_packet $msec  ]

   #
   # result = 2 -> got packet
   #

   if { "$result" == "2" } {

      # -- get the contents of the TCP packet from the stack
      set tcpState $s(current,TCP)
      upvar #$s(current,TCP,level) $tcpState tcp
      upvar $packet p

      # -- copy the packet to pass back to the user
      pktsrc_copy_packet tcp p

      set ret 0


   #
   # result = 1 -> timeout
   #

   } elseif { "$result" == "1" } {

      
      pktsrcInfo s \
         "timeout waiting for TCP packet at port [expr 0x$srcPort]"

      set ret -1

   } else {

      set ret -1
   }

   # -- removing the callback
   Stack_remove_callback s tcpPort $srcPort TCP_generic_input

   return $ret
}

#
# TCP_compute_checksum
#
# calculate the TCP checksum for the current packet on the stack
#

proc TCP_compute_checksum { stack } {
   upvar $stack s

   # -- get the current IP contents
   set ipState $s(current,IPv4)
   upvar #$s(current,IPv4,level) $ipState ip

   set tcpState $s(current,TCP)
   upvar #$s(current,TCP,level) $tcpState tcp

   # -- grab the TCP header and data from IP
   set packet $ip(data)

   # -- we could grab the ip length field, but then we need to
   # check the ip header length as well. Just strlen the data

   set length [expr [ string length $packet ] / 2 ] 

   # -- build the pseudo header
   set pseudo [ format "%s%s0006%4.4x" $ip(src) $ip(dst) $length ]
   
   # -- now compute the checksum
   set checksum [ pktsrc checksum $pseudo$packet ]

   # -- return the checksum
   return $checksum
}




