#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# 2001 QAcafe http://qacafe.com
#
# $Id: udp.tcl,v 1.19 2002/12/10 22:31:33 joe Exp $
#
#
# Poor man's UDP
#
#
# ----------------------------------------------------------------------------

# -- this is part of the pktsrc package
package provide pktsrc 1.3

proc UDP_input { packet stack { options none }} {
   upvar $stack s

   if { $options == "print" } {

      UDP_decode $packet udp
      UDP_print udp

   } else {

      # -- remember this packet in the 'udp' variable. This allows other
      # callbacks to get access to the decodes at any Tcl stack level

      eval [ Stack_save_packet s UDP udp ]
   
      # -- now decode the packet into the 'udp' variable
      UDP_decode $packet udp

   }

   # -- check for callbacks at the UDP level
   if [ info exists s(udpPort,$udp(dstPort)) ] {
      # -- drive any input callbacks
      foreach c $s(udpPort,$udp(dstPort),callbacks) {
          $c $udp(data) s $options
      }

   } elseif { "$options" != "print" } {
 
      set ipState $s(current,IPv4)
      upvar #$s(current,IPv4,level) $ipState ip

      if { "$ip(dst)" != "$s(ip)" } {
        return
      }

      # -- send ICMP Destination Unreachable
      ICMP_port_unreachable s
   }

}

#
#
# UDP_init - set up any deaults for UDP
#

proc UDP_init { stack } {
   upvar $stack s

   # -- turn on UDP checksums
   set s(UDP,checksum)            on


   # -- initial UDP ports are based out of the same space as TCP

   set s(UDP,client,connections) 0
   set s(UDP,server,connections) 0

}

proc UDP_decode { packet udp } {
   upvar $udp out

   # -- parse out all the fields we find interesting ...

   set out(srcPort)     [ string range $packet 0 3 ]
   set out(dstPort)     [ string range $packet 4 7 ]
   set out(length)      [ string range $packet 8 11 ]
   set out(checksum)    [ string range $packet 12 15 ]

   set out(data)        [ string range $packet 16 end ]

}

proc UDP_print { udp } {
   upvar $udp out

   puts "UDP: ---- UDP Packet ----"
   puts "UDP: "
   puts "UDP: Source Port                = [expr 0x$out(srcPort)] [UDP_port [expr 0x$out(srcPort)]]"
   puts "UDP: Destination Port           = [expr 0x$out(dstPort)] [UDP_port [expr 0x$out(dstPort)]]"
   puts "UDP: Length                     = [expr 0x$out(length)]"
   puts "UDP: Checksum                   = 0x$out(checksum)"
   puts "UDP:"
}

#
# Display well known UDP ports
#
# port is the decimal value of a well known port

proc UDP_port { port } {
   if { $port == "520" } {
      set str "(RIP)"
   } elseif { $port == "53" } {
       set str "(DNS)"
   } elseif { $port == "123" } {
       set str "(NTP)"
   } elseif { $port == "137" } {
       set str "(NETBIOS-NS)"
   } elseif { $port == "161" } {
       set str "(SNMP)"
   } elseif { $port == "162" } {
       set str "(SNMP Trap)"
   } elseif { $port == "500" } {
       set str "(IKE)"
   } elseif { $port == "1900" } {
       set str "(SSDP)"
   } else {
      set str ""
   }
}

proc UDP_get_client_port { stack } {
   upvar $stack s

   # -- share the same port space with TCP
   return [ TCP_get_client_port s ]

}

proc UDP_send { stack packet dest destPort srcPort } {
   upvar $stack s

   set length [expr [ string length $packet ] / 2 + 8 ]

   # -- compute the UDP checksum

   # -- first build the pseudo header
   set pseudo [ format "%s%s0011%4.4x" $s(ip) $dest $length ]
   
   # +++joem need to pad to even size
   set udp [ format "%s%4.4x%4.4x%4.4x0000%s" $pseudo $srcPort $destPort $length $packet ]
   set checksum [ pktsrc checksum $udp ]


   # -- check for other checksum options
   if { "$s(UDP,checksum)" == "bad" } {

      # -- use a bad checksum, increment the checksum by 1
      set checksum [ incrhex16 $checksum ]

   } elseif { "$s(UDP,checksum)" == "off" } {
  
      # -- don't set the checksum
      set checksum 0000
   }

   # -- build the real packet
   set udp [ format "%4.4x%4.4x%4.4x%s%s" $srcPort $destPort $length $checksum $packet]

   # -- send this to IP using udp as the protocol
   IPv4_send $udp s $dest 11
}

proc UDP_send_pad { stack packet dest destPort srcPort padSize } {
   upvar $stack s

   set length [expr [ string length $packet ] / 2 + 8 ]

   # -- pad out to $padSize

   set pad [ expr $padSize - $length ]
   set padStr [ string repeat 00 $pad ]
   set packet "$packet$padStr"
   set length $padSize

   # -- compute the UDP checksum

   # -- first build the pseudo header
   set pseudo [ format "%s%s0011%4.4x" $s(ip) $dest $length ]
   
   # +++joem need to pad to even size
   set udp [ format "%s%4.4x%4.4x%4.4x0000%s" $pseudo $srcPort $destPort $length $packet ]
   set checksum [ pktsrc checksum $udp ]


   if { "$s(UDP,checksum)" == "bad" } {

      # -- use a bad checksum, increment the checksum by 1
      set checksum [ incrhex16 $checksum ]

   } elseif { "$s(UDP,checksum)" == "off" } {
  
      # -- don't set the checksum
      set checksum 0000
   }

   # -- build the real packet
   set udp [ format "%4.4x%4.4x%4.4x%s%s" $srcPort $destPort $length $checksum $packet]

   # -- send this to IP using udp as the protocol
   IPv4_send $udp s $dest 11
}

#
# UDP_init_session
#
proc UDP_init_session { stack session } {

   upvar $stack s
   upvar $session sess

   set sess(type)  udp
   set sess(srcPort) 0000
   set sess(dstPort) 0000

   set sess(dstIp) 00000000
   set sess(srcIp) $s(ip)

   set sess(state) create

   set sess(bytes-in) 0
   set sess(bytes-out) 0

   set sess(data) ""
}

#
# UDP_open (client)
#
# Open a UDP connection (i.e client)
#

proc UDP_open { stack session dstIp dstPort } {
   upvar $stack s
   upvar $session sess

   UDP_init_session s sess

   # -- bump the client connections count
   incr s(UDP,client,connections)

   # -- remember the local port
   set port [ UDP_get_client_port s ]

   set sess(srcPort) [ format "%4.4x" $port ]
   set sess(dstPort) [ format "%4.4x" $dstPort ]

   set sess(dstIp) [ iptohex $dstIp ]

   # -- add a callback for this session
   Stack_add_callback s udpPort $sess(srcPort) UDP_session_input

   # --- hmmm ....
   if { [ expr 0x$sess(dstPort) ] == 53 } {
       # -- add callback for 53
       Stack_add_callback s udpPort $sess(srcPort) DNS_input
   }

   UDP_add_session s sess

   return $port
}

#
# UDP_add_session
#
# Add a UDP connection to the stack
#

proc UDP_add_session { stack session } {
   upvar $stack s
   upvar $session sess

   set key [ UDP_session_key sess ]
   set s(UDP,session,$key,state) create

   # -- I need to save a session using the same trick as save_packet

   UDP_save_session s sess
}

#
# UDP_session_key
#
# Build a key used to look up UDP sessions
#

proc UDP_session_key { session } {
   upvar $session sess

   return "$sess(dstIp),$sess(srcIp),$sess(dstPort),$sess(srcPort)"
}

#
# UDP_save_session
#
# Save the contents of the session in stack storage
#

proc UDP_save_session { stack session } {
   upvar $stack s
   upvar $session sess

   set key [ UDP_session_key sess ]

   # -- remember all the fields
   set s(UDP,session,$key,fields) "[array names sess]"

   # -- set the key for lookups

   set s(UDP,session,$key,key) $key
   # -- store a copy on the stack
   foreach field [ array names sess ] {

      # -- don't pass the data around
      if { "$field" == "data" } {
          continue
      }
      set s(UDP,session,$key,$field) $sess($field)
   }

}


#
# UDP_server
#
# Listen for UDP connection on port
#

proc UDP_server { stack port { sessionCallback "" }} {
   upvar $stack s

   # -- get the src port in hex
   set srcPort [ format "%4.4x" $port ]

   # -- add a callback for this server
   Stack_add_callback s udpPort $srcPort UDP_server_input

   if { "$sessionCallback" != "" } {
     Stack_add_callback s udpSessionPort $srcPort $sessionCallback
   }

   pktsrcInfo s "Accepting udp connections on port $port"
}

#
# UDP_server_shutdown
#
# Shutdown down udp server on port
#

proc UDP_server_shutdown { stack port { sessionCallback "" }} {

   upvar $stack s

   # -- get the src port in hex
   set srcPort [ format "%4.4x" $port ]

   # -- add a callback for this server
   Stack_remove_callback s udpPort $srcPort UDP_server_input

   if { "$sessionCallback" != "" } {
     Stack_remove_callback s udpSessionPort $srcPort $sessionCallback
   }

   pktsrcInfo s "udp server on port $port has shutdown"
}

#
# UDP_server_input
#
# input process for UDP packets
#
#  - no data buffer is done here like tcp
#  - call any session callbacks for udp


proc UDP_server_input { packet stack {options none}} {
   upvar $stack s

   if { $options == "print" } {
      return
   }

   # -- get the current IP contents
   set ipState $s(current,IPv4)
   upvar #$s(current,IPv4,level) $ipState ip

   # -- get the current udp contents
   set udpState $s(current,UDP)
   upvar #$s(current,UDP,level) $udpState udp
   
   pktsrcInfo s "[hextoip $ip(dst)]:[expr 0x$udp(dstPort)] received UDP from [hextoip $ip(src)]:[expr 0x$udp(srcPort)]"

   if [ info exists s(udpSessionPort,$udp(dstPort)) ] {

      # -- each new UDP packet to the server side is considered to be a session
      incr s(UDP,server,connections)

      # -- drive any input callbacks
      foreach c $s(udpSessionPort,$udp(dstPort),callbacks) {
          $c $udp(data) s $options
      }
   }
}

#
# UDP_generic_input
#
# generic callback for catching UDP packets
#

proc UDP_generic_input { packet stack { options none }} {
   upvar $stack s

   # -- trigger the packet event
   if { "$options" != "print" } {
      Event_got_packet
   }
}

#
# UDP_get_packet
#
# blocking call for receiving UDP packets
# this call blocks until the packet is received or a timeout occurs
#

proc UDP_get_packet { stack srcPort packet timeout } {
   upvar $stack s

   set ret -1

   # add a callback for UDP packets for this source port
   Stack_add_callback s udpPort $srcPort UDP_generic_input

   pktsrcInfo s \
     "Waiting up to $timeout seconds for UDP packet at port [expr 0x$srcPort]"

   # -- compute timeout in milli seconds
   set msec [ expr $timeout * 1000 ]

   # -- now wait for the packet or timeout
   set result [ Event_wait_packet $msec  ]

   #
   # result = 2 -> got packet
   #

   if { "$result" == "2" } {

      # -- get the contents of the UDP packet from the stack
      set udpState $s(current,UDP)
      upvar #$s(current,UDP,level) $udpState udp
      upvar $packet p

      # -- copy the packet to pass back to the user
      pktsrc_copy_packet udp p

      set ret 0


   #
   # result = 1 -> timeout
   #

   } elseif { "$result" == "1" } {

      
      pktsrcInfo s \
         "timeout waiting for UDP packet at port [expr 0x$srcPort]"

      set ret -1

   } else {

      set ret -1
   }

   # -- removing the callback
   Stack_remove_callback s udpPort $srcPort UDP_generic_input

   return $ret
}

proc UDP_get_data { stack session } {
   upvar $stack s
   upvar $session sess

   set key [ UDP_session_key sess ]

   set currentData ""

   if [ info exists s(UDP,session,$key,data) ] {
      set currentData $s(UDP,session,$key,data)
      set s(UDP,session,$key,data) ""
   }

   return $currentData
}

#
# UDP_remove_session
#
# Remove a UDP session from the stack storage
#

proc UDP_remove_session { stack session } {
   upvar $stack s
   upvar $session sess

   set key [ UDP_session_key sess ]

   # -- remember all the fields
   set fields $s(UDP,session,$key,fields)
   unset s(UDP,session,$key,fields)

   # -- store a copy on the stack
   foreach field $fields {
      if { "$field" == "data" }{
         continue
      }
      unset s(UDP,session,$key,$field)
   }

   # -- remove callbacks ?

}

#
# UDP_get_session
#
# Find a session, return its contents in the 'session' variable
#

proc UDP_get_session { stack session dstIp dstPort srcPort } {
   upvar $stack s
   upvar $session sess

   set findsess(dstIp) $dstIp
   set findsess(dstPort) $dstPort
   set findsess(srcPort) $srcPort
   set findsess(srcIp) $s(ip)

   set key [ UDP_session_key findsess ]

   # -- check for session
   if [ info exists s(UDP,session,$key,fields) ] {

      foreach field $s(UDP,session,$key,fields) {

         # -- don't pass the data around
         if { "$field" == "data" } {
             continue
         }
         set sess($field) $s(UDP,session,$key,$field)
      }

      return $sess(srcPort)

   } else {
      return 0
   }
}

#
# UDP_session_input
#
# input process for UDP packets
#
#  - find existing session for this udp connection
#  - hand off packet and session if found
#  - else, ignore packet

proc UDP_session_input { packet stack {options none}} {
   upvar $stack s

   if { $options == "print" } {
      return
   }

   # -- get the current IP contents
   set ipState $s(current,IPv4)
   upvar #$s(current,IPv4,level) $ipState ip

   # -- get the current udp contents
   set udpState $s(current,UDP)
   upvar #$s(current,UDP,level) $udpState udp
   
   pktsrcInfo s "[hextoip $ip(dst)]:[expr 0x$udp(dstPort)] received UDP from [hextoip $ip(src)]:[expr 0x$udp(srcPort)]"

   # -- check that this is a valid session
   set sid [ UDP_get_session s session $ip(src) $udp(srcPort) $udp(dstPort) ]

   if { "$sid" != "0" } {

      pktsrcInfo s \
         "found udp session at port [expr 0x$udp(dstPort)]"

      # -- process the udp session state machine
      UDP_buffer_data s session $udp(data)
      UDP_save_session s session

   } else {
      pktsrcInfo s "no existing session"
   }

}

proc UDP_buffer_data { stack session data } {
   upvar $stack s
   upvar $session sess

   set key [ UDP_session_key sess ]

   set currentData ""

   if [ info exists s(UDP,session,$key,data) ] {
       set currentData $s(UDP,session,$key,data)
   }

   set s(UDP,session,$key,data) "$currentData$data"

}

#
# UDP_close
#
# Close a UDP connection
#

proc UDP_close { stack session } {
   upvar $stack s
   upvar $session sess

   set sid [ UDP_get_session s isess $sess(dstIp) $sess(dstPort) $sess(srcPort) ]

   if { "$sid" != "0" } {
     # -- remove it ?
   }
}

#
# UDP_clear_connections
#
# remove any state associated with existing UDP connections
#

proc UDP_clear_connections { stack } {
   upvar $stack s

   # -- reset the connection counters
   set s(UDP,client,connections) 0
   set s(UDP,server,connections) 0

   foreach keylookup [ array names s UDP,session,*,key ] {

      # -- get the key
      set key $s($keylookup)

      set fields $s(UDP,session,${key},fields)

      unset s(UDP,session,${key},fields)
      unset s(UDP,session,${key},key)

      foreach field $fields {
          if { "$field" == "data" } {
             continue
          }
	  unset s(UDP,session,${key},${field}) 
      }
   }
}

proc UDP_get_stats { stack stat } {
    upvar $stack s

    if { "$stat" == "server-connections" } {
       return $s(UDP,server,connections)
    } elseif { "$stat" == "client-connections" } {
       return $s(UDP,client,connections)
    } else {
       error "UDP_get_stats: unknown stat type '$stat'"
    }
}


#
# UDP checksum control

proc UDP_checksum_bad { stack } {
   upvar $stack s
   set s(UDP,checksum) bad
}

proc UDP_checksum_on { stack } {
   upvar $stack s
   set s(UDP,checksum) on
}

proc UDP_checksum_off { stack } {
   upvar $stack s
   set s(UDP,checksum) off
}

#
# UDP_compute_checksum
#
# calculate the UDP checksum for the current packet on the stack
#

proc UDP_compute_checksum { stack } {
   upvar $stack s

   # -- get the current IP contents
   set ipState $s(current,IPv4)
   upvar #$s(current,IPv4,level) $ipState ip

   set udpState $s(current,UDP)
   upvar #$s(current,UDP,level) $udpState udp

   if { [ string compare "$udp(checksum)" "0000" ] == 0 } {
       return 0000
   }

   # -- grab the UDP header and data from IP
   set packet $ip(data)

   # -- build the pseudo header
   set pseudo [ format "%s%s0011%s" $ip(src) $ip(dst) $udp(length) ]
   
   # -- now compute the checksum
   set checksum [ pktsrc checksum $pseudo$packet ]

   # -- return the checksum
   return $checksum
}


proc UDP_echo_server { stack port } {
   upvar $stack s


   set srcPort [ format "%4.4x" $port ]

   pktsrcInfo s "Starting up UDP echo server on port $port"

   UDP_server s $port UDP_echo_server_input

}

proc UDP_echo_server_shutdown { stack port } {
   upvar $stack s

   set srcPort [ format "%4.4x" $port ]

   pktsrcInfo s "Shuting down UDP echo server on port $port"

   UDP_server_shutdown s $port UDP_echo_server_input

}


#
# UDP_echo_server_input
#
#

proc UDP_echo_server_input { packet stack stack {options none}} {
   upvar $stack s

   if { "$options" == "print" } {
       return
   }

   # -- get the current IP contents
   set ipState $s(current,IPv4)
   upvar #$s(current,IPv4,level) $ipState ip

   # -- get the current udp contents
   set udpState $s(current,UDP)
   upvar #$s(current,UDP,level) $udpState udp

   # -- just send a dump response back
   pktsrcInfo s "Responding to UDP echo request from [hextoip $ip(src)]:[expr 0x$udp(srcPort)]"

   set response "${udp(data)}"

   UDP_send s $response $ip(src) [ expr 0x$udp(srcPort) ] [ expr 0x$udp(dstPort) ] 

}

proc UDP_echo { stack server dport sport } {
   upvar $stack s

   set srcPort [ format "%4.4x" $sport ]

   pktsrcInfo s "Initiate UDP echo to $server:$dport using src port $sport"

   # -- send request
   set request "000000000000000000000000000000000000000000000000000000000000"

   UDP_send s "$request" [iptohex $server] $dport $sport 

   set result [ UDP_get_packet s $srcPort udp 10 ] 

   if { "$result" != "-1" } {
      pktsrcInfo s "got UDP response from $server"
      return "okay"
   } else {
      pktsrcWarn s "no UDP response from $server after 10 seconds"
      return ""
   }
}