#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# 2002 QAcafe http://qacafe.com
#
# $Id: vrrp.tcl,v 1.2 2002/11/15 13:23:54 joe Exp $
#
#
# pktsrc VRRP
#
# ----------------------------------------------------------------------------

# -- this is part of the pktsrc package
package provide pktsrc 1.3

proc VRRP_input { packet stack { options none }} {
   upvar $stack s

   if { $options == "print" } {

      if [ catch { VRRP_decode $packet vrrp } reason ] {
          pktsrcWarn s "Unable to fully parse VRRP packet"
      } else {
          VRRP_print vrrp
      }
      return

   } else {

      # -- remember this packet in the 'vrrp' variable. This allows other
      # callbacks to get access to the decodes at any Tcl stack level

      eval [ Stack_save_packet s VRRP vrrp ]
   
      # -- now decode the packet into the 'vrrp' variable
      if [ catch { VRRP_decode $packet vrrp } reason ] {
          pktsrcWarn s "Unable to fully parse VRRP packet"
          return
      }

   }

   # -- check for callbacks at the VRRP type level
   # only 1 type is defined - advertisement
 
   if [ info exists s(vrrpType,$vrrp(type)) ] {

      # -- drive any input callbacks
      foreach c $s(vrrpType,$vrrp(type),callbacks) {
          # -- send the whole packet
          $c $packet s $options
      }

   } else {
      # no callback
   }
}

#
#
# VRRP_init - set up any deaults for VRRP
#

proc VRRP_init { stack } {
   upvar $stack s

   # -- default interval is 1 second
   set s(VRRP,interval) 1

   # -- turn on normal checksums
   set s(VRRP,checksum) on

   # -- add mac entry for VRRP multicast address
   ARP_add_entry s [iptohex 224.0.0.18] 01005e000012
}

#
#
# from rfc2338
#
#       0                   1                   2                   3
#       0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
#      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#      |Version| Type  | Virtual Rtr ID|   Priority    | Count IP Addrs|
#      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#      |   Auth Type   |   Adver Int   |          Checksum             |
#      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#      |                         IP Address (1)                        |
#      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#      |                            .                                  |
#      |                            .                                  |
#      |                            .                                  |
#      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#      |                         IP Address (n)                        |
#      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#      |                     Authentication Data (1)                   |
#      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#      |                     Authentication Data (2)                   |
#      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#


proc VRRP_decode { packet VRRP } {
   upvar $VRRP out

   # -- parse out all the fields we find interesting ...

   set out(version)        [ string range $packet 0 0 ]
   set out(type)           [ string range $packet 1 1 ]
   set out(vid)            [ string range $packet 2 3 ]
   set out(priority)       [ string range $packet 4 5 ]
   set out(count)          [ expr 0x[ string range $packet 6 7 ]]
   set out(authType)       [ string range $packet 8 9 ]
   set out(advertInt)      [ string range $packet 10 11 ]
   set out(checksum)       [ string range $packet 12 15 ]


   for { set i 0 } { $i < $out(count) } { incr i } {
      set start [ expr $i * 8 + 16 ]
      set end [ expr $i * 8 + 23 ] 
      set ip [ string range $packet $start $end ]
      if { "$ip" != "" } {
         set index [ expr $i + 1 ]
         set out(ip,$index) $ip
      } else {
         # -- missing data
      }  
   }

   # -- auth data
}

proc VRRP_print { VRRP } {
   upvar $VRRP out

   puts "VRRP: ---- VRRP Packet ----"
   puts "VRRP: "
   puts "VRRP: Version                   = [expr 0x$out(version)]"
   puts "VRRP: Type                      = [VRRP_get_type $out(type)]"
   puts "VRRP: Virtual Router ID         = [expr 0x$out(vid)]"
   puts "VRRP: Priority                  = [expr 0x$out(priority)]"
   puts "VRRP: Count                     = $out(count)"
   puts "VRRP: Authentication Type       = [VRRP_get_auth $out(authType)]"
   puts "VRRP: Advert Interval           = [ expr 0x$out(advertInt) ] second(s)"
   puts "VRRP: Checksum                  = 0x$out(checksum)"

   for { set i 1 } { $i <= $out(count) } { incr i } {
      if [ info exists out(ip,$i) ] {

        puts "VRRP: Address [ format "%-3s" $i ]               = [hextoip $out(ip,$i)]"

      } else {
         break
      }  
   }

   puts "VRRP:"
   puts "VRRP: ---- End VRRP Packet ----"
}

proc VRRP_get_type { type } {
   if { "$type" == "01" } {
       return "ADVERTISEMENT"
   } else {
       return "Unknown Type '$type'"
   }
}

proc VRRP_get_auth { type } {

   if { "$type" == "00" } {
       return "None"

   } elseif { "$type" == "01" } {
       return "Simple Password"

   } elseif { "$type" == "02" } {
       return "IP Auth Header"
   }
}

proc VRRP_send_advert { stack vrid priority args } {

   upvar $stack s

   set vridhex   [ format "%2.2x" $vrid ]
   set prihex    [ format "%2.2x" $priority ]
   set adverthex [ format "%2.2x" $s(VRRP,interval) ]

   set count 0
   set ipStr ""
   foreach ip $args {
     incr count
     set ipStr "$ipStr[iptohex $ip]"
   }
   set counthex [ format "%2.2x" $count ]
   
   set packet [ format "%s%s%s%s%s%s%s%s%s0000000000000000" \
       2 \
       1 \
       $vridhex \
       $prihex \
       $counthex \
       00 \
       $adverthex \
       0000 \
       $ipStr ]

   VRRP_send s $packet
   
}

proc VRRP_send { stack packet } {
   upvar $stack s

   set length [expr [ string length $packet ] / 2 ]

   # -- compute the VRRP checksum
   set checksum [ pktsrc checksum $packet ]

   # -- check for other checksum options
   if { "$s(VRRP,checksum)" == "bad" } {

      # -- use a bad checksum, increment the checksum by 1
      set checksum [ incrhex16 $checksum ]

   } elseif { "$s(VRRP,checksum)" == "off" } {
  
      # -- don't set the checksum
      set checksum 0000
   }

   # -- build the real packet
   set vrrpC [ string replace $packet 12 15 $checksum ]

   # -- send this to IP using vrrp as the protocol
   IPv4_send $vrrpC s [iptohex 224.0.0.18] 70
}







