
#
# $Id: pktsrc-unit-test.tcl,v 1.9 2002/11/07 11:51:41 joe Exp $
#

# -- simple version test
add_test pktsrc_1 "Test out the version command" {

    step 1. Run the pktsrc version command
}

proc pktsrc_1 { i } {

    loginfo "Checking pktsrc version ..."

    set version ""
    set version [ pktsrc version ]

    if { "version" != "" } {
      PASS "Found version $version"
    } else {
      FAIL "Could not find version"
    }
}

# -- ip checksum test
add_test pktsrc_2 "Test out the IP checksum" {

    step 1. Compute an IP checksum over an IP packet
    step 2. Insert the checksum into the packet
    step 3. Re-compute the IP checksum
    step 4. Verify the checksum is 0000
}

proc pktsrc_2 { i } {

    set ip      "4508002ea5ea00003e11"
    set cksum   "0000"

    loginfo "Computing checksum over IP header with 0000 checksum field"
    set checksum [ pktsrc checksum ${ip}00001213141621222324 ]

    loginfo "Checksum is $checksum"
    loginfo "Computing checksum over IP header with checksum inserted"
    set checksum2 [ pktsrc checksum ${ip}${checksum}1213141621222324 ]

    if { "$checksum2" == "0000" } {
      PASS "Checksum is 0000 as expected"
    } else {
      FAIL "Checksum is not 0000 ($checksum2)"
    }
}

# -- md5 test
add_test pktsrc_3 "Test out the MD5 digest on a text string" {

    step 1. Create a text string
    step 2. Compute the MD5 digest on the text string
    step 3. Verify the result compared to the expected result

    NOTE: The 'md5' utility compiled on FreeBSD was used to generated
      the expected md5 result.
}

proc pktsrc_3 { i } {

    set s  "This is a dummy string"

    # -- This was computed on FreeBSD 3.1 running
    # % md5 -s "This is a dummy string"
    #

    set md5expected "e3a4b56bc7b9f2bc30265e0869c20f46"
    loginfo "Computing md of string '$s'"

    # -- call md5 with no key
    set md5 [ pktsrc md5 [ pktsrc text2ascii $s ] "" ]

    loginfo "md5 is '$md5'"

    if { "$md5" == "$md5expected" } {
      PASS "md5 matches expected value"
    } else {
      FAIL "md5 is not $md5expected"
    }

}


# -- pktsrc match packet test
add_test pktsrc_4 "Test out the pktsrc match command" {

    step 1. Create an known acsii hex packet
    step 2. Run the match command with various filters over the packet
    step 3. Verify each match has the expected result
}

proc pktsrc_4 { i } {

    set packet "00010203040500020304050608004500002500000000220600000a0000010a000002"

    set handle [ testvar handle ]

    # -- try several filters
    set fl { 
       { arp                 0 } 
       { ip                  1 } 
       { tcp                 1 }
       { "host 10.0.0.1"     1 }
       { "host 22.2.2.1"     0 }
    }

    foreach pair $fl {
       set filter [ lindex $pair 0 ]
       set expected [ lindex $pair 1 ]

       loginfo "trying filter '$filter' ..."
       set result [ $handle match $packet "$filter" ]

       if { $result == 1 } {
          if { $expected == 1 } {
             PASS "'match' command passed as expected"
          } else {
             FAIL "'match' command passed unepxectedly"
          }
       } else {
          if { $expected == 0 } {
             PASS "'match' command failed as expected"
          } else {
             FAIL "'match' command failed unepxectedly"
          }
      }
   }

}


# -- pktsrc open close test
add_test pktsrc_50 "Open and close a pktsrc handle 10,000 times" {

    step 1. Create a new pktsrc handle using the same device
    step 2. Close the handle
    step 3. Repeat 10,000 times

    This test tries to catch any file handle leaks betweek the pktsrc  
    layer and libpcap.

}

proc pktsrc_50 { i } {

    set ver [ testvar libpcap_ver ]

    if { "$ver" == "0.4" } {
       buddy::FAIL "this test will not be run since libpcap version is 0.4"
       buddy::FAIL "there is a known issue with atexit() usage in pcap 0.4"
       return
    }

    set handle [ testvar handle ]
    set device [ $handle device ]

    loginfo "creating new handles using device $device"

    for { set i 0 } { $i < 10000 } { incr i } {
      set newHandle [ pktsrc open $device ]
      $newHandle close
    }

    buddy::loginfo "last handle created was '$newHandle'"
    PASS "all 10,000 handles opened/closed"

}

# -- pktsrc open close test
add_test pktsrc_60 "Verify hwaddr file handle leak is fixed" {

    step 1. Create 2 new pktsrc handles using the same device
    step 2. Get the mac address using [ pktsrc hwaddr ]
    step 2. Close both handles
    step 3. Repeat 10,000 times

    This test tries to catch a fd leak that was part of hwaddr. This
    was fixed in 1.2.16.

}

proc pktsrc_60 { i } {

    set ver [ testvar libpcap_ver ]

    if { "$ver" == "0.4" } {
       buddy::FAIL "this test will not be run since libpcap version is 0.4"
       buddy::FAIL "there is a known issue with atexit() usage in pcap 0.4"
       return
    }

    # -- call match for 10.0.0.1
    set handle [ testvar handle ]
    set device [ $handle device ]

    loginfo "creating new handles using device $device"

    for { set i 0 } { $i < 10000 } { incr i } {
      set newHandle1 [ pktsrc open $device ]
      set newHandle2 [ pktsrc open $device ]
      $newHandle1 hwaddr
      $newHandle2 hwaddr
      $newHandle1 close
      $newHandle2 close
    }

    buddy::loginfo "last handle created was '$newHandle2'"
    PASS "all 10,000 handles opened/closed"

}


# -- pktsrc URL procs
add_test pktsrc_100 "Verify URL_get_domain and URL_get_page" {

    step 1. Call URL_get_domain with serveral different URLs
    step 2. Verify correct domain is returned
    step 3. Call URL_get_page with serveral differnet URLs
    step 4. Verify correct page is returned

}

proc pktsrc_100 { i } {

    set url(1,url) "http://www.google.com/search?hl=en&ie=ISO-8859-1&q=qacafe"
    set url(1,domain) "www.google.com"
    set url(1,page) "/search?hl=en&ie=ISO-8859-1&q=qacafe"

    set url(2,url) "http://www.google.com"
    set url(2,domain) "www.google.com"
    set url(2,page) "/"

    set url(3,url) "http://www.google.com/"
    set url(3,domain) "www.google.com"
    set url(3,page) "/"

    set url(4,url) "http://google.com/"
    set url(4,domain) "google.com"
    set url(4,page) "/"

    set url(5,url) "google.com/"
    set url(5,domain) "google.com"
    set url(5,page) "/"

    set url(6,url) "google.com"
    set url(6,domain) "google.com"
    set url(6,page) "/"

    set urls 6

    for { set i 1 } { $i <= $urls } { incr i } {
       buddy::loginfo "Testing $url($i,url)"
       set domain [ URL_get_domain $url($i,url) ]
       set page [ URL_get_page $url($i,url) ]

       if { "$domain" != "$url($i,domain)" } {
           buddy::FAIL "found domain '$domain', expected $url($i,domain)"
       } else {
           buddy::PASS "ok, found page '$domain'"
       }

       if { "$page" != "$url($i,page)" } {
           buddy::FAIL "found page '$page', expected $url($i,page)"
       } else {
           buddy::PASS "ok, found page '$page'"
       }
    }

}

add_test pktsrc_net_1 "Verify ip_range_contains" {

    step 1. Call ip_range_contains with various arguments
    step 2. Verify correct result is returned
}

proc pktsrc_net_1 { i } {

    set u(1,start) 10.2.2.1
    set u(1,stop) 10.2.2.255
    set u(1,ip) 10.2.2.59
    set u(1,result) 1

    set u(2,start) 10.2.2.1
    set u(2,stop) 10.2.2.255
    set u(2,ip) 10.2.2.1
    set u(2,result) 1

    set u(3,start) 10.2.2.1
    set u(3,stop) 10.2.2.255
    set u(3,ip) 10.2.2.255
    set u(3,result) 1

    set u(4,start) 10.2.2.1
    set u(4,stop) 10.2.2.255
    set u(4,ip) 10.2.1.59
    set u(4,result) 0

    set number 4

    for { set i 1 } { $i <= $number } { incr i } {
       buddy::loginfo "Testing case $i, $u($i,start) $u($i,stop) $u($i,ip)"
       set result [ ip_range_contains $u($i,start) $u($i,stop) $u($i,ip) ]

       if { "$result" != "$u($i,result)" } {
           buddy::FAIL "result was '$result', expected $u($i,result)"
       } else {
           buddy::PASS "ok, result was '$result'"
       }
    }
}








